;MLC (Multi-platform Language for Calculators) by Epic Programming Studios
;TI-86 interpreter
;Written by Donald Straney (burntfuse) - straney@umbi.umd.edu

;Token parse/exec code

;//////////////////////////////Main loop//////////////////////////////

exec:
exec_reset:
	ld hl,(start_addr)					;get the address of the function MAIN to find where to start
	ld (parse_ptr),hl
	ld (exec_sp),sp						;and save sp in case a command needs to be aborted
exec_loop:
	ld hl,(parse_ptr)
	ld a,(hl)							;look at the next token
	cp eof_token						;if it's the end of the file,
	ret z								;then we're done!
	ld de,exec_loop						;(push the address where all the handlers should return when done)
	push de
	cp int_name_token					;if the line holds an expression,
	jp z,exec_expr_int					;then evaluate it!
	cp array_name_token
	jp z,exec_expr_array
	cp str_name_token
	jp z,exec_expr_str
	cp spr_name_token					;branch if it's changing the parameters of some sprite
	jp z,exec_spr_op
	cp ptr_name_token					;or a pointer
	jp z,exec_ptr_op
	cp map_name_token					;or even a tilemap
	jr z,exec_expr_map
	inc hl
	ld (parse_ptr),hl
exec_command:							;otherwise, it holds a command, so execute it
	add a,a								;find the address of the handler code depending on the command
	ld e,a
	ld d,0
	ld hl,exec_handler_table
	add hl,de
	ld a,(hl)
	inc hl
	ld h,(hl)
	ld l,a
	jp (hl)								;and execute the handler

;//////////////////////////////Expression handlers//////////////////////////////

exec_expr_map:
	call read_var_addr					;of course, get the map's address first
exec_expr_map_:
	ld a,(hl)
	cp index_start_token				;then find if it's a full map definition or just a definition for a single element
	jp nz,exec_expr_map_full
exec_expr_map_element:
	inc hl
	ld (parse_ptr),hl
	call get_var						;(make sure that the map has already been created)
	ld a,h
	or l
	jp z,err_noarray
	call get_expr_val					;get the x position of the element to modify
	ld hl,(expr_sum)
	push hl
	call get_expr_val					;then the y pos
	ld hl,(parse_ptr)
#ifdef FULL_ERR
	ld a,(hl)
	cp index_end_token
	jp nz,err_syntax
#endif
	inc hl
	ld (parse_ptr),hl
	ld bc,(expr_sum)
	pop hl
	ld de,(curr_var)
	call find_map_element				;using those, find the address of the map element
	push hl
	call get_bmp						;get the address of the bitmap that should be stored to the map
	call expect_eol
	pop hl
	ld (hl),c							;and store it
	inc hl
	ld a,b
	and $0f
	ld (hl),a
	ret
exec_expr_map_full:
	push hl
	call get_var						;make sure the map's been created already
	ld a,h
	or l
	jp z,err_noarray
	inc hl
	inc hl
	ld e,(hl)							;and get its size
	inc hl
	ld d,(hl)
	inc hl
	push hl
	ex de,hl
	call _htimesl
	ld b,h
	ld c,l								;bc=length counter
	pop hl								;hl=map data
	pop de								;de=map data to parse
exec_expr_map_full_loop:
	ld a,(de)							;REPLACE W/GET_BMP LATER?
#ifdef FULL_ERR
	cp bmp_name_token
	jp nz,err_bmp
#endif
	cp eol_token
	jr z,exec_expr_map_full_end_loop
	inc de
	ld a,(de)							;store the bitmap in the tilemap
	ld (hl),a
	inc hl
	inc de
	ld a,(de)
	and $0f								;and set it to transparent
	ld (hl),a
	inc hl
	inc de
	dec bc								;decrement the length counter
	jr exec_expr_map_full_loop			;and loop
exec_expr_map_full_end_loop:
	ld a,b								;when done, make sure that the definition lengths match
	or c
	jp nz,err_arraydeflen
	inc de
	ld (parse_ptr),de					;and update the parse pointer
	ret


exec_ptr_op:
	call read_var_addr					;get the pointer's address
	ld a,(hl)							;now...is the pointer being assigned a var?
	cp Lcolon
	jr nz,exec_ptr_op_expr
	inc hl
	ld b,(hl)							;if so, find the var/sprite/bitmap type
	inc hl
	ld e,(hl)
	inc hl
	ld d,(hl)
	inc hl
#ifdef FULL_ERR
	ld a,(hl)
	cp eol_token
	jp nz,err_eol
#endif
	inc hl
	ld (parse_ptr),hl
	ld a,b
	call tok_to_char					;and get the appropriate type char for it
	ld a,b
	ex de,hl
	ld de,(curr_var)
	jp set_ptr							;and set the pointer
exec_ptr_op_expr:
	ld (parse_ptr),hl					;if it's an expression that starts with a pointer,
	call get_ptr						;get the pointer type to determine where to go next!
	ld (curr_var),hl					;(set curr_var so that the var assigned to the pointer will be operated on by the expression handlers)
	ld hl,(parse_ptr)
	or a
	jp z,err_noptr
	cp int_char
	jp z,exec_expr_int_
	cp array_char
	jp z,exec_expr_array_
	cp string_char
	jr z,exec_expr_str_
	cp sprite_char
	jp z,exec_spr_op_
	cp map_char
	jp z,exec_expr_map_
	jp err_syntax



exec_expr_str:							;if it's a string var,
	call read_var_addr					;get the var index and save it for later
exec_expr_str_:
	ld a,(hl)							;see if the user is trying to assign a literal string
	inc hl
	cp int_char							;or a converted number to the string var
	jr z,exec_expr_str_num
#ifdef FULL_ERR
	cp string_char
	jp nz,err_argtype
#endif
exec_expr_str_str:
	call get_str						;evaluate the string expression
	ld de,(curr_var)					;then find the starting address of the string var in memory
	call _strcopy						;copy it to the end of the current string in the string var
	call expect_eol						;it should be at the end of the line now...
	ld hl,(curr_var)					;before returning, make sure that the result of the string operations hasn't overflowed
	ex de,hl
	or a
	sbc hl,de
	ld de,str_max_len+1
	call _cphlde
	jp nc,err_strlen
	ret
exec_expr_str_num:
	call get_expr_val_hl				;get the value of the numerical expression
	ld hl,(expr_sum)
	call format_num						;then format it to a displayable string
	ld de,(curr_var)					;find the storage location for the current var
	jp _strcopy							;and copy the formatted number to there


exec_expr_int:							;go here if it's an integer var
	call read_var_addr
exec_expr_int_:
	ld a,(hl)
	cp LplusSign						;WAIT! is this a var increment or var decrement line???
	jr z,exec_expr_inc					;if so, handle that separately
	cp Ldash
	jr z,exec_expr_dec
	cp Lasterisk
	jr z,exec_expr_selfop
	cp Lslash
	jr z,exec_expr_selfop
	call get_expr_val_hl				;now parse and evaluate the rest of the expression	ld hl,(expr_sum)					;get the sum of the expression
	ld hl,(expr_sum)
	ld de,(curr_var)					;and the index of the var being set
	jp set_var							;so that we can give that var the proper value!
exec_expr_inc:
	inc hl
	ld a,(hl)
	cp eol_token
	jr nz,exec_expr_selfop_alt
	inc hl
	ld (parse_ptr),hl
	push de
	call get_var						;[var]+, so increment var
	inc hl
	pop de
	jp set_var
exec_expr_dec:
	inc hl
	ld a,(hl)
	cp eol_token
	jr nz,exec_expr_selfop_alt
	inc hl
	ld (parse_ptr),hl
	push de
	call get_var
	dec hl								;decrement var
	pop de
	jp set_var
exec_expr_selfop_alt:
	dec hl
exec_expr_selfop:
	ld (parse_ptr),hl
	push de
	call get_var
	ld (expr_sum),hl
	call get_expr_val_loop
	ld hl,(expr_sum)
	pop de
	jp set_var


exec_expr_array:						;if it's an array,
	call read_var_addr
exec_expr_array_:
	ld a,(hl)
	cp index_start_token				;check if it's an entire array definition or just defining a single element
	jr nz,exec_expr_array_whole
exec_expr_array_element:
	inc hl
	call get_expr_val_hl				;get the index of the array element being modified,
	ld de,(curr_var)					;make sure it's not out of range,
	call get_var
	ld a,h
	or l
	jp z,err_noarray
	ld b,(hl)
	ld hl,(expr_sum)
	ld a,h
	or a
	jp nz,err_arrayindex
	ld a,l
	cp b
	jp nc,err_arrayindex
	push hl
	ld hl,(parse_ptr)
#ifdef FULL_ERR
	ld a,(hl)							;make sure all the formatting is correct,
	cp index_end_token
	jp nz,err_int
#endif
	inc hl
	call get_expr_val_hl				;then get the value that the array element is supposed to be set to,
	ld de,(curr_var)					;find the location of the array element to be modified,
	call get_var
	inc hl
	pop de
	ex de,hl
	add hl,hl
	add hl,de
	ld de,(expr_sum)					;and set it!
	ld (hl),e
	inc hl
	ld (hl),d
	ret									;so now we're done
exec_expr_array_whole:
	ld (parse_ptr),hl					;update the parse pointer
	ld de,(curr_var)					;now get the address of the array var data
	call get_var
	ld a,h								;(make sure that it's been defined)
	or l
	jp z,err_noarray
	ld a,(hl)							;(save array length)
	ld (exec_expr_array_length),a
	inc hl
	ld b,0								;init array length counter
exec_expr_array_whole_loop:
	push bc
	push hl
	call get_expr_val					;get the value of this element of the array
	pop hl
	pop bc
	ld de,(expr_sum)					;and store it to the array
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	inc b								;and increment index counter
	jr nc,exec_expr_array_whole_loop	;if it's not the end of the line, then loop again
	ld a,b								;when done, make sure that the literal array and the array var
exec_expr_array_length	equ	$+1
	cp 0								;have the same length!
	jp nz,err_arraydeflen
	ret									;if they do, then go back to execute the next line


exec_spr_op:
	call read_var_addr					;get the index of the sprite we're going to modify
exec_spr_op_:
	ld a,(hl)
	cp LcapB
	jr z,exec_obj_op_bmp
	push af								;and the parameter that gets modified
	inc hl
	call get_expr_val_hl				;now get the value that the parameter should be set to
	jp nc,err_syntax
	pop af								;and set it, depending on which parameter it is
	ld de,(curr_var)
	ld bc,(expr_sum)
	sub LcapD				;DX (D)
	jp z,set_spr_dx
	dec a					;DY (E)
	jp z,set_spr_dy
	sub LcapV-LcapE			;VX (V)
	jp z,set_spr_vx
	dec a					;VY (W)
	jp z,set_spr_vy
	push af
	ld hl,(expr_sum)
	call format_spr_pos
	pop bc
	dec b					;X (X)
	jp z,set_spr_x
#ifdef FULL_ERR
	dec b					;Y (Y)
	jp z,set_spr_y
	jp err_sprprop
#else
	jp set_spr_y
#endif
exec_obj_op_bmp:						;the user's trying to set the sprite's bitmap property...
	inc hl
	call get_bmp_hl
	call expect_eol
	ld de,(curr_var)
	jp set_spr							;and finally, set the original sprite


get_bmp:								;get bitmap at (parse_ptr), return address in bc, parse_ptr updated
	ld hl,(parse_ptr)
get_bmp_hl:
	ld a,(hl)							;is it a bitmap name or another sprite's bitmap?
	inc hl
	ld e,(hl)
	inc hl
	ld d,(hl)
	inc hl
	ld (parse_ptr),hl
	cp bmp_name_token
	jr z,get_bmp_bmp
	cp ptr_name_token
	jr z,get_bmp_ptr
	cp map_name_token
	jr z,get_bmp_map
#ifdef FULL_ERR
	cp spr_name_token
	jp nz,err_bmp
#endif
get_bmp_spr:							;if it's the bitmap of another sprite,
	call get_spr						;using the sprite's address, find its bitmap's address,
	ld hl,(parse_ptr)
#ifdef FULL_ERR
	ld a,(hl)
	cp LcapB
	jp nz,err_bmp
#endif
	inc hl
	ld (parse_ptr),hl
	ld b,d
	ld c,e								;and return it in bc
	ret
get_bmp_ptr:							;if it's a pointer,
	call get_ptr
	cp bitmap_char						;look at the data type
	jr z,get_bmp_ptr_bmp
	cp sprite_char
	jp nz,err_bmp
get_bmp_ptr_spr:
	ld de,(parse_ptr)					;if it's a sprite, make sure that there's a '.BMP' after it
	ld a,(de)
	cp LcapB
	jp nz,err_bmp
	inc de
	ld (parse_ptr),de
	ex de,hl
	call get_spr						;then get the bitmap
	ld b,d								;and return it in bc
	ld c,e
	ret
get_bmp_ptr_bmp:						;if it's a pointer, then get its address
	ld c,l
	ld b,h
	ret
get_bmp_bmp:							;if it's a bitmap name, we can get off easy!
	ld b,d								;just transfer the address to bc
	ld c,e
	ret
get_bmp_map:
	push de
	ld a,(hl)
	cp index_start_token
	jp nz,err_bmp
	inc hl
	call get_expr_val_hl				;get the x pos
	ld hl,(expr_sum)
	push hl
	call get_expr_val					;then the y pos
#ifdef FULL_ERR
	ld a,(hl)
	cp index_end_token
	jp nz,err_syntax
#endif
	inc hl
	ld (parse_ptr),hl
	ld bc,(expr_sum)
	pop hl
	pop de
	call find_map_element
	ld c,(hl)							;get the address of the bitmap from the tilemap data
	inc hl
	ld a,(hl)
	or $a0
	ld b,a
	ret


get_expr_val:							;get value of expression at (parse_ptr), parse_ptr updated, value returned in expr_sum, c set if new line
	ld hl,(parse_ptr)
get_expr_val_hl:
	call get_val						;init the sum of the expression depending on whether it starts out with a varname or number
	ld (expr_sum),hl
get_expr_val_loop:
	ld hl,(parse_ptr)
	ld a,(hl)							;get the next char in the expression
	inc hl
	cp eol_token						;if it's a newline token,
	jr z,get_expr_val_end				;then we're done with this expression!
	sub Lasterisk						;otherwise, check for other operators
	jr z,get_expr_val_m
	dec a
	jr z,get_expr_val_a
	sub Ldash-LplusSign
	jr z,get_expr_val_s
	sub Lslash-Ldash
	jr z,get_expr_val_d
	sub Lbar-Lslash
	jr z,get_expr_val_mod
	jr get_expr_val_end_2
get_expr_val_a:							;	+
	call get_val						;get the value of the next var/num,
	ld de,(expr_sum)
	add hl,de							;add it to the current sum,
	ld (expr_sum),hl
	jr get_expr_val_loop				;and loop
get_expr_val_s:							;	-
	call get_val
	ld de,(expr_sum)
	ex de,hl
	or a
	sbc hl,de
	ld (expr_sum),hl
	jr get_expr_val_loop
get_expr_val_m:							;	*
	call get_val						;ooohh...we have to do a 16-bit multiply here...this'll be hard...
	ld de,(expr_sum)
	call mult_num
	ld (expr_sum),hl
	jr get_expr_val_loop
get_expr_val_d:							;	/
	call get_val						;and a 16-bit divide too...why am I always stuck with the hard ones???
	ld de,(expr_sum)
	ex de,hl
	call div_num
	ld (expr_sum),hl
	jr get_expr_val_loop
get_expr_val_mod:						; x | y
	call get_val
	ld de,(expr_sum)
	ex de,hl							;hl=x, de=y
	push hl
	push de
	call div_num						;hl=x/y
	pop de
	call mult_num						;hl=(x/y)*y
	pop de
	ex de,hl							;hl=x, de=(x/y)*y
	or a
	sbc hl,de							;hl=x-(x/y)*y
	ld (expr_sum),hl
	jr get_expr_val_loop
get_expr_val_end:
	ld (parse_ptr),hl					;point to the beginning of the next line
	scf
	ret
get_expr_val_end_2:
	dec hl
	ld (parse_ptr),hl
	xor a
	ret


get_val:								;parse number or varname at (hl), parse_ptr updated, number returned in hl
	ld a,(hl)
	inc hl
	ld e,(hl)							;first get the address
	inc hl
	ld d,(hl)
	inc hl
	ld (parse_ptr),hl
	cp lit_num_token					;is it a number?
	jp z,get_val_num					;if so, parse that, obviously
	cp spr_name_token					;it's also accepted if it's a sprite's parameter...
	jr z,get_val_spr
	cp ptr_name_token					;or a pointer!!!
	jr z,get_val_ptr
get_val_var:
	cp int_name_token					;is it a var?
	jp z,get_var						;if so, parse that
	cp array_name_token
	jp nz,err_num
get_val_array:							;if it's an element of an array, then get the value of that element
	call get_var
	ld de,(expr_sum)					;FIRST, SAVE THE CURRENT EXPRESSION'S SUM!!!
	push de
	push hl								;save the location of the array's data
	ld hl,(parse_ptr)
	ld a,(hl)							;make sure it's an array element, not a full array
	cp index_start_token
	jp nz,err_var_int					;no no no...bad user!!!
	inc hl
	call get_expr_val_hl				;now get the index of the array element
#ifdef FULL_ERR
	ld a,(hl)
	cp index_end_token
	jp nz,err_int
#endif
	inc hl
	ld (parse_ptr),hl					;and update the parse pointer
	pop de
	ld hl,(expr_sum)					;now find the location of the correct element in the array
	ld a,h								;(make sure it's not out of bounds)
	or a
	jp nz,err_arrayindex
	ld a,(de)
	cp l
	jp c,err_arrayindex
	inc de
	add hl,hl
	add hl,de
	call _ldhlind						;and get the array element's value from that
	pop de
	ld (expr_sum),de					;then restore current expression sum
	ret
get_val_ptr:
	call get_ptr						;hmmm...what type of pointer is it?
	ex de,hl
	ld hl,(parse_ptr)					;(get address of var in de, parse_ptr in hl)
	cp sprite_char
	jr z,get_val_spr
	cp int_char
	jp z,get_var
	cp array_char
	jp z,get_val_array
	or a
	jp z,err_noptr
	jp err_num
get_val_spr:
	ld a,(hl)							;check on the sprite parameter
	inc hl
	ld (parse_ptr),hl					;and update pointer
	sub LcapD
	jr z,get_val_obj_dx
	dec a
	jr z,get_val_obj_dy
	sub LcapV-LcapE
	jr z,get_val_obj_vx
	dec a
	jr z,get_val_obj_vy
	dec a
	jr z,get_val_obj_x
#ifdef FULL_ERR
	dec a
	jp nz,err_sprprop
#endif
get_val_obj_y:							;if it's the y coordinate,
	call get_spr						;get all the sprite's parameters,
	ld h,0
	ld l,c								;but save only the y pos
	ld a,c
	cp 128
	ret c
	dec h
	ret
get_val_obj_x:
	call get_spr
	ld h,0
	ld l,b
	ld a,b
	cp 128
	ret c
	dec h
	ret
get_val_obj_vy:
	call get_spr_vy
	ld l,c
	ld h,b
	ret
get_val_obj_vx:
	call get_spr_vx
	ld l,c
	ld h,b
	ret
get_val_obj_dx:
	call get_spr_dx
	ld l,c
	ld h,b
	ret
get_val_obj_dy:
	call get_spr_dy
	ld l,c
	ld h,b
	ret
get_val_num:
	ex de,hl							;if the expression is a number, get it in hl
	ret


get_str:								;get string at (hl), return pointer in hl, update parse_ptr
	ld de,str_temp_mem
	ld (str_temp_ptr),de
get_str_loop:
	ld a,(hl)							;look at the next token
	sub lit_str_token					;if it's a string,
	jr z,get_str_lit					;then just return a pointer to that
	dec a
	jr z,get_str_var_num				;if it's a number, format it to a string
	sub 3
	jr z,get_str_var_num
	sub 2
	jr z,get_str_ptr					;if it's a pointer, treat it like a var
	dec a
	jr z,get_str_var_str
	dec a
	jr z,get_str_var_num
#ifdef FULL_ERR
	dec a
	jp nz,err_str						;if it's none of these, show error
#endif
get_str_var_num:
	call get_expr_val_hl				;get the value of the expression...
	ld hl,(expr_sum)
	call format_num						;then we have to format it...*sigh*...
	jr get_str_end
get_str_var_str:
	inc hl
	ld e,(hl)
	inc hl
	ld d,(hl)
	inc hl
	ld (parse_ptr),hl
get_str_var_str_:
	ex de,hl
	jr get_str_end
get_str_ptr:
	push hl
	inc hl
	ld e,(hl)
	inc hl
	ld d,(hl)
	inc hl
	ld (parse_ptr),hl
	call get_ptr						;make sure that the pointer's pointing to a string var...
	ex de,hl							;and put the string's address in de...
	pop hl
	or a
	jp z,err_noptr
	cp string_char						;the rest can be handled by the existing code! :-)
	jr z,get_str_var_str_
	cp int_char
	jr z,get_str_var_num
	cp array_char
	jr z,get_str_var_num
	cp sprite_char
	jr z,get_str_var_num
	jp err_str
get_str_lit:
	inc hl								;get past the literal string token, and we have the zero-terminated string right there!!!
	push hl								;but we still have to update the parse pointer...save the starting address of the string
	ld bc,str_max_len+1
	xor a								;search for the end of the string
	cpir
	ld (parse_ptr),hl					;and update the parse pointer
	pop hl								;now restore the pointer to the string
get_str_end:
	ld de,(str_temp_ptr)				;copy the string to a temp area
	call _strcopy
	ld (str_temp_ptr),de
	ld hl,(parse_ptr)
	ld a,(hl)
	cp LplusSign
	jr nz,get_str_ret
	inc hl
	jp get_str_loop
get_str_ret:
	xor a
	ld (de),a
	ld hl,str_temp_mem
	ret


;//////////////////////////////Signed Math Routines//////////////////////////////

mult_num:					;hl=hl*de
	call is_hl_neg
	jp nc,mult_num_hl
	ex de,hl
	call is_hl_neg
	ex de,hl
	jp nc,mult_num_neg_de
	jp mult16
mult_num_hl:
	ex de,hl
	call is_hl_neg
	jp c,mult16
mult_num_neg_de:
	call neg_de
	call mult16
	ex de,hl
	ld hl,0
	or a
	sbc hl,de
	ret

div_num:					;hl=hl/de
	call is_hl_neg
	jp nc,div_num_hl
	ex de,hl
	call is_hl_neg
	ex de,hl
	jp nc,div_num_neg_de
	jp div16
div_num_hl:
	ex de,hl
	call is_hl_neg
	jp nc,div_num_neg_both
div_num_neg_hl:
	call neg_de
	ex de,hl
	jp div_num_neg
div_num_neg_de:
	call neg_de
div_num_neg:
	call div16
	ex de,hl
	ld hl,0
	or a
	sbc hl,de
	ret
div_num_neg_both:
	push hl
	ld hl,0
	or a
	sbc hl,de
	ex de,hl
	pop hl
	ex de,hl
	push hl
	ld hl,0
	or a
	sbc hl,de
	ex de,hl
	pop hl
	jp div16

is_hl_neg:					;nc if hl is neg, c if pos
	ld bc,32768
	push hl
	or a
	sbc hl,bc
	pop hl
	ret

neg_de:						;de= -de
	push hl
	ld hl,0
	or a
	sbc hl,de
	ex de,hl
	pop hl
	ret

cp_num:						;cp hl,de
	call _cphlde
	ret z
	call is_hl_neg
	jp nc,cp_num_hl_neg
	ex de,hl
	call is_hl_neg
	ex de,hl
	jp nc,cp_num_ccf
	jp _cphlde
cp_num_hl_neg:
	ex de,hl
	call is_hl_neg
	ex de,hl
	jp nc,_cphlde
	scf
	ret
cp_num_ccf:
	xor a
	inc a
	ret

;//////////////////////////////Command handlers//////////////////////////////

exec_c_arry:
	call expect_arrayname				;get the index of the array var and save it
	push de
	call get_var						;(make sure it hasn't been defined already)
	ld a,h
	or l
	jp nz,err_twovar_arry
	call expect_num						;now get the size of the array
	ld a,h
	or a
	jp nz,err_arraylen
	or l
	jp z,err_arraylen
	add hl,hl							;and allocate the memory
	inc hl
	ld b,h
	ld c,l
	call alloc_mem
	ld (hl),a
	pop de
	jp set_var


exec_c_size:
	call expect_varname					;get index of array or string var
	push de								;then save the address w/the size
	push af
	call expect_intname					;now get the index of the var where the size will be stored
	ld (exec_c_size_var),de
	pop af								;wait...are we trying to find the size of an array or a string???
	cp array_char
	jr z,exec_c_size_array
	cp string_char
	jp nz,err_vartype					;if it's neither of those, show an error
exec_c_size_str:
	pop hl
	call _strlength						;if it's a string, use the handy API call to find its length
	ld h,b
	ld l,c
	jr exec_c_size_end
exec_c_size_array:						;if it's an array,
	pop de
	call get_var
	ld l,(hl)							;we get the array size from the first byte in the array table!
	ld h,0								;is this easy or WHAT?!?!
exec_c_size_end:
exec_c_size_var	equ	$+1
	ld de,0
	call set_var_warray					;store the size to the var
	jp expect_eol						;so now we're done!


exec_c_clrs:							;clear screen
	call expect_eol
	jp clr_buf


exec_c_text:							;draw small text
	call expect_x_pos					;first get x pos
	ld (_pencol),a						;and set the cursor
	call expect_y_pos					;now get the y pos
	ld (_penrow),a						;and set the cursor pos
	call expect_str						;now get a pointer to the string to be displayed
	push hl								;save that
	call expect_color					;now get the color
	pop hl
	or a								;now branch depending on the color
	jr z,exec_c_text_white
	dec a
	jr z,exec_c_text_lgray
	dec a
	jr z,exec_c_text_dgray
	dec a
	jr z,exec_c_text_black
exec_c_text_inv:
	ld bc,pixel_inv
	ld (pixel_op_patch),bc
	ld a,gray1_buf/256
	call exec_c_text_draw
	ld a,gray2_buf/256
	jr exec_c_text_draw
exec_c_text_white:
	ld bc,pixel_off
	ld (pixel_op_patch),bc
	ld a,gray1_buf/256
	call exec_c_text_draw
	ld a,gray2_buf/256
	jr exec_c_text_draw
exec_c_text_lgray:
	ld bc,pixel_on
	ld (pixel_op_patch),bc
	ld a,gray1_buf/256
	call exec_c_text_draw
	ld bc,pixel_off
	ld (pixel_op_patch),bc
	ld a,gray2_buf/256
	jr exec_c_text_draw
exec_c_text_dgray:
	ld bc,pixel_off
	ld (pixel_op_patch),bc
	ld a,gray1_buf/256
	call exec_c_text_draw
	ld bc,pixel_on
	ld (pixel_op_patch),bc
	ld a,gray2_buf/256
	jr exec_c_text_draw
exec_c_text_black:
	ld bc,pixel_on
	ld (pixel_op_patch),bc
	ld a,gray1_buf/256
	call exec_c_text_draw
	ld a,gray2_buf/256
;	jr exec_c_text_draw
exec_c_text_draw:
	ld (fvputc_vidplane),a
	add a,screen_len/256
	ld (fvputc_vidplane2),a
	push hl
	call fvputs
	pop hl
	ret
exec_c_text_draw2:
	ld (fvputc_vidplane),a
	add a,full_screen_len/256
	ld (fvputc_vidplane2),a
	push hl
	call fvputs128
	pop hl
	ret



exec_c_line:
	call expect_x_pos					;get x1
	push af
	call expect_y_pos					;then y1
	ld e,a
	pop af
	ld d,a
	push de								;now we have the first set of coordinates...
	call expect_x_pos					;so get x2
	push af
	call expect_y_pos					;and y2
	ld l,a
	pop af
	ld h,a								;by now, both coordinates are present and accounted for
	push hl
	call expect_color					;so get the color
	pop hl
	pop de
	or a
	jr z,exec_c_line_white
	dec a
	jr z,exec_c_line_lgray
	dec a
	jr z,exec_c_line_dgray
	dec a
	jr z,exec_c_line_black
exec_c_line_invert:						;and draw depending on the color
	ld bc,pixel_inv
	ld (pixel_op_patch),bc
	ld bc,gray1_buf
	call FLine
	ld bc,gray2_buf
	jp FLine
exec_c_line_white:
	ld bc,pixel_off
	ld (pixel_op_patch),bc
	ld bc,gray1_buf
	call FLine
	ld bc,gray2_buf
	jp FLine
exec_c_line_lgray:
	ld bc,pixel_on
	ld (pixel_op_patch),bc
	ld bc,gray1_buf
	call FLine
	ld bc,pixel_off
	ld (pixel_op_patch),bc
	ld bc,gray2_buf
	jp FLine
exec_c_line_dgray:
	ld bc,pixel_off
	ld (pixel_op_patch),bc
	ld bc,gray1_buf
	call FLine
	ld bc,pixel_on
	ld (pixel_op_patch),bc
	ld bc,gray2_buf
	jp FLine
exec_c_line_black:
	ld bc,pixel_on
	ld (pixel_op_patch),bc
	ld bc,gray1_buf
	call FLine
	ld bc,gray2_buf
	jp FLine


exec_c_pixl:
	call expect_x_pos					;get x pos
	push af
	call expect_y_pos					;then the y pos
	ld c,a
	pop af
	ld b,a
	push bc								;now we have the coordinates, so get the color
	call expect_color
	pop bc
	or a
	jr z,exec_c_pixl_white
	dec a
	jr z,exec_c_pixl_lgray
	dec a
	jr z,exec_c_pixl_dgray
	dec a
	jr z,exec_c_pixl_black
exec_c_pixl_inv:
	call exec_c_pixl_draw1
	call pixel_inv
	call exec_c_pixl_draw2
	jp pixel_inv
exec_c_pixl_white:
	call exec_c_pixl_draw1
	call pixel_off
	call exec_c_pixl_draw2
	jp pixel_off
exec_c_pixl_lgray:
	call exec_c_pixl_draw1
	call pixel_on
	call exec_c_pixl_draw2
	jp pixel_off
exec_c_pixl_dgray:
	call exec_c_pixl_draw1
	call pixel_off
	call exec_c_pixl_draw2
	jp pixel_on
exec_c_pixl_black:
	call exec_c_pixl_draw1
	call pixel_on
	call exec_c_pixl_draw2
	jp pixel_on
exec_c_pixl_draw1:
	call FindPixel
	ld b,a
	ld de,gray1_buf
	add hl,de
	ret
exec_c_pixl_draw2:
	ld a,b
	ld de,gray2_buf-gray1_buf
	add hl,de
	ret


exec_c_pixt:
	call expect_x_pos					;get the x pos of the pixel to test
	push af
	call expect_y_pos					;then its y pos
	ld c,a
	pop af
	ld b,a
	push bc
	call expect_intname					;now get the var where the color will be stored
	ld (exec_c_pixt_var),de
	xor a								;init the temp color var,
	ld (sum),a
	pop bc
	call FindPixel						;then find the pixel in the first screen layer
	push hl
	ld b,a
	ld de,gray1_buf
	add hl,de
	and (hl)							;and test it
	call nz,exec_c_pixt_lgray
	pop hl								;now find the same pixel in the second screen layer
	ld a,b
	ld de,gray2_buf
	add hl,de
	and (hl)
	call nz,exec_c_pixt_dgray
	ld a,(sum)							;get the color we've found
	ld l,a
	ld h,0
exec_c_pixt_var	equ	$+1
	ld de,0								;and store it to the var
	call set_var_warray
	jp expect_eol
exec_c_pixt_lgray:
	ld hl,sum
	inc (hl)
	ret
exec_c_pixt_dgray:
	ld hl,sum
	inc (hl)
	inc (hl)
	ret


exec_c_draw:
	call expect_eol
	jp copy_scrn


exec_c_stop:
	call expect_eol
	ld hl,gray1_buf
	ld de,scr_save1
	ld bc,screen_len*2
	ldir
	ret


exec_c_rclp:
	call expect_eol
	ld hl,scr_save1
	ld de,gray1_buf
	ld bc,screen_len*2
	ldir
	ret


exec_c_rect:
	call expect_x_pos					;get x1
	ld (temp_buf+1),a					;and save it
	call expect_y_pos					;do the same with y1,
	ld (temp_buf),a
	call expect_x_pos					;x2,
	ld (temp_buf+3),a
	call expect_y_pos					;and y2
	ld (temp_buf+2),a
	call expect_color					;now get the line color
	or a
	jr z,exec_c_rect_white				;and branch depending on it
	dec a
	jr z,exec_c_rect_lgray
	dec a
	jr z,exec_c_rect_dgray
	dec a
	jr z,exec_c_rect_black
	dec a
	jr z,exec_c_rect_inv
	jr exec_c_rect_continue
exec_c_rect_inv:
	ld hl,pixel_inv
	ld (pixel_op_patch),hl
	ld bc,gray1_buf
	call exec_c_rect_lines
	ld bc,gray2_buf
	call exec_c_rect_lines
	jr exec_c_rect_continue
exec_c_rect_white:
	ld hl,pixel_off
	ld (pixel_op_patch),hl
	ld bc,gray1_buf
	call exec_c_rect_lines
	ld bc,gray2_buf
	call exec_c_rect_lines
	jr exec_c_rect_continue
exec_c_rect_lgray:
	ld hl,pixel_on
	ld (pixel_op_patch),hl
	ld bc,gray1_buf
	call exec_c_rect_lines
	ld hl,pixel_off
	ld (pixel_op_patch),hl
	ld bc,gray2_buf
	call exec_c_rect_lines
	jr exec_c_rect_continue
exec_c_rect_dgray:
	ld hl,pixel_off
	ld (pixel_op_patch),hl
	ld bc,gray1_buf
	call exec_c_rect_lines
	ld hl,pixel_on
	ld (pixel_op_patch),hl
	ld bc,gray2_buf
	call exec_c_rect_lines
	jr exec_c_rect_continue
exec_c_rect_black:
	ld hl,pixel_on
	ld (pixel_op_patch),hl
	ld bc,gray1_buf
	call exec_c_rect_lines
	ld bc,gray2_buf
	call exec_c_rect_lines
;	jp exec_c_rect_continue
exec_c_rect_continue:
	call expect_color					;now get the fill color
	or a
	jr z,exec_c_rect_iwhite
	dec a
	jr z,exec_c_rect_ilgray
	dec a
	jr z,exec_c_rect_idgray
	dec a
	jr z,exec_c_rect_iblack
	dec a
	jr z,exec_c_rect_iinv
	ret
exec_c_rect_iinv:
	ld hl,pixel_inv
	ld (pixel_op_patch),hl
	call exec_c_rect_layer1
	jr exec_c_rect_layer2
exec_c_rect_iwhite:
	ld hl,pixel_off
	ld (pixel_op_patch),hl
	call exec_c_rect_layer1
	jr exec_c_rect_layer2
exec_c_rect_ilgray:
	ld hl,pixel_on
	ld (pixel_op_patch),hl
	call exec_c_rect_layer1
	ld hl,pixel_off
	ld (pixel_op_patch),hl
	jr exec_c_rect_layer2
exec_c_rect_idgray:
	ld hl,pixel_off
	ld (pixel_op_patch),hl
	call exec_c_rect_layer1
	ld hl,pixel_on
	ld (pixel_op_patch),hl
	jr exec_c_rect_layer2
exec_c_rect_iblack:
	ld hl,pixel_on
	ld (pixel_op_patch),hl
	call exec_c_rect_layer1
	jr exec_c_rect_layer2
exec_c_rect_layer1:
	ld bc,(temp_buf)
	ld de,(temp_buf+2)
	inc b
	inc c
	ld a,gray1_buf/256
	jp box
exec_c_rect_layer2:
	ld bc,(temp_buf)
	ld de,(temp_buf+2)
	inc b
	inc c
	ld a,gray2_buf/256
	jp box
exec_c_rect_lines:
	ld hl,(temp_buf)	;x1,y1 - x1,y2
	ld a,(temp_buf+1)
	ld d,a
	ld a,(temp_buf+2)
	ld e,a
	call FLine
	ld hl,(temp_buf)	;x1,y1 - x2,y1
	ld a,(temp_buf+3)
	ld d,a
	ld a,(temp_buf)
	ld e,a
	call FLine
	ld hl,(temp_buf+2)	;x2,y2 - x1,y2
	ld a,(temp_buf+1)
	ld d,a
	ld a,(temp_buf+2)
	ld e,a
	call FLine
	ld hl,(temp_buf+2)	;x2,y2 - x2,y1
	ld a,(temp_buf+3)
	ld d,a
	ld a,(temp_buf)
	ld e,a
	jp FLine


exec_c_disp:
	ld hl,(parse_ptr)
	ld a,(hl)
	cp lit_num_token
	jr z,exec_c_disp_all
	call expect_sprname					;get the sprite to draw
	call exec_c_disp_set				;update the position from VX and VY
	call get_spr						;now get the sprite's position
	ld a,d
	or e
	jp z,err_nospr
	call draw_bmp						;and draw the sprite
	jp expect_eol
exec_c_disp_all:						;if there isn't a sprite name, it must be a .0
	inc hl								;(get the number just to make sure it is a 0)
	ld b,(hl)
	inc hl
	ld a,(hl)
	inc hl
	ld (parse_ptr),hl
	or b
	jp nz,err_invarg
	ld a,(num_sprites)					;get the number of sprites to draw
	or a
	jp z,find_eol						;if there aren't any, then what's the point of drawing them???
	ld b,a
	ld de,spr_1_data					;init the pointer to the beginning of the sprites...
exec_c_disp_all_loop:
	push de
	push bc
	call exec_c_disp_set				;move the sprite if it has a velocity
	call get_spr
	ld a,d								;then make sure it's been defined,
	or e
	jp z,err_nospr
	call draw_bmp						;and draw it
	pop bc
	pop de
	ld hl,spr_data_len
	add hl,de							;now point to the next sprite,
	ex de,hl
	djnz exec_c_disp_all_loop			;and loop
	jp expect_eol						;when all the sprites have been drawn, we're done
exec_c_disp_set:
	push de
	push de
	call get_spr_vx						;and the amount to add to the coordinates for the velocity...
	push bc
	call get_spr_dx
	pop hl
	add hl,bc							;update the position from that
	ld b,h
	ld c,l
	call set_spr_dx
	pop de
	call get_spr_vy
	push bc
	call get_spr_dy
	pop hl
	add hl,bc
	ld b,h
	ld c,l
	call set_spr_dy
	pop de
	ret


shft_size	equ		temp_buf
shft_mode	equ		temp_buf+2
shft_fill	equ		temp_buf+3
shft_fill_2	equ		temp_buf+3
shft_fill_1	equ		temp_buf+4

exec_c_shft:
	call expect_num						;get the x amount it should be shifted
	ld (exec_c_shft_patchx),hl
	call expect_num						;then the y shift
	ld (exec_c_shft_patchy),hl
	call expect_num						;and whether it should loop or not
;It is illegal to copy and re-use the following code in the handler due to federal standards of bad coding and "spaghetti code" regulations  ;-D
	inc hl
	call format_color
	ld (shft_mode),a
exec_c_shft_patchx	equ	$+1
	ld hl,0
	call is_hl_neg
	jp nc,exec_c_shft_x_neg
	ld a,h
	or a
	jp nz,err_argrange
	ld b,l
	ld a,b
	or a
	jr z,exec_c_shft_y
	ld a,(shft_mode)
	dec a
	call get_color
	ld a,d
	and $01
	ld d,a
	ld a,e
	and $01
	ld e,a
	ld (shft_fill),de
exec_c_shft_x_loop:
	push bc
	ld a,(shft_fill_1)
	ld (rshift_x_patch),a
	ld de,gray1_buf
	call rshift_x
	ld a,(shft_fill_2)
	ld (rshift_x_patch),a
	ld de,gray2_buf
	call rshift_x
	pop bc
	djnz exec_c_shft_x_loop
	jp exec_c_shft_y
exec_c_shft_x_neg:
	ex de,hl
	ld hl,0
	or a
	sbc hl,de
	ld a,h
	or a
	jp nz,err_argrange
	ld b,l
	ld a,b
	or a
	jr z,exec_c_shft_y
	ld a,(shft_mode)
	dec a
	call get_color
	ld a,d
	and $01
	ld d,a
	ld a,e
	and $01
	ld e,a
	ld (shft_fill),de
exec_c_shft_x_neg_loop:
	push bc
	ld a,(shft_fill_1)
	ld (lshift_x_patch),a
	ld de,gray1_buf+screen_len-1
	call lshift_x
	ld a,(shft_fill_2)
	ld (lshift_x_patch),a
	ld de,gray2_buf+screen_len-1
	call lshift_x
	pop bc
	djnz exec_c_shft_x_neg_loop
exec_c_shft_y:
	ld a,(shft_mode)
	dec a
	call get_color
	ld a,d
	ld (exec_c_shft_y_patch_1),a
	ld (exec_c_shft_y_patch_3),a
	ld a,e
	ld (exec_c_shft_y_patch_2),a
	ld (exec_c_shft_y_patch_4),a
exec_c_shft_patchy	equ	$+1
	ld hl,0
	call is_hl_neg
	jp nc,exec_c_shft_y_neg
	ld a,h
	or a
	jp nz,err_argrange
	ld a,l
	or a
	ret z
	cp shft_y_max+1
	jp nc,err_argrange
	add hl,hl				;*2
	add hl,hl				;*4
	ld e,l
	ld d,h
	add hl,hl				;*8
	add hl,de				;*12
	ld (shft_size),hl		;size of block to shift
	ld hl,$52ed	;sbc hl,de
	ld (exec_c_shft_buf_y_patch7),hl
	ld hl,$b8ed	;lddr
	ld (exec_c_shft_buf_y_patch8),hl
	ld hl,$b0ed	;ldir
	ld (exec_c_shft_buf_y_patch9),hl
	ld hl,gray1_buf
	ld de,gray1_buf+screen_len-1
exec_c_shft_y_patch_1	equ	$+1
	ld a,0
	call exec_c_shft_buf_y_pos
	ld hl,gray2_buf
	ld de,gray2_buf+screen_len-1
exec_c_shft_y_patch_2	equ	$+1
	ld a,0
exec_c_shft_buf_y_pos:
	ld (exec_c_shft_buf_y_patch1),hl
	ld (exec_c_shft_buf_y_patch2),hl
	inc hl
	ld (exec_c_shft_buf_y_patch3),hl
	ex de,hl
	ld (exec_c_shft_buf_y_patch5),hl
	ld (exec_c_shft_buf_y_patch6),hl
	inc hl
	ld de,(shft_size)
	or a
	sbc hl,de
	ld (exec_c_shft_buf_y_patch4),hl
exec_c_shft_buf_y:
	ld a,(shft_mode)
	or a
	jp nz,exec_c_shft_skip1
exec_c_shft_buf_y_patch4	equ	$+1
	ld hl,0
	ld de,shft_temp
	ld bc,(shft_size)
	ldir
exec_c_shft_skip1:
	ld hl,screen_len		;length of copy for main shift
	ld de,(shft_size)
	or a
	sbc hl,de
	push hl
exec_c_shft_buf_y_patch5	equ	$+1
	ld hl,gray1_buf+screen_len-1	;starting pos for shift
	ld de,(shft_size)
	or a
exec_c_shft_buf_y_patch7	equ	$
	sbc hl,de
exec_c_shft_buf_y_patch6	equ	$+1
	ld de,gray1_buf+screen_len-1
	pop bc
exec_c_shft_buf_y_patch8	equ	$
	lddr
	ld a,(shft_mode)
	or a
	jp z,exec_c_shft_skip2
exec_c_shft_buf_y_patch1	equ	$+1
	ld hl,gray1_buf
exec_c_shft_buf_y_patch3	equ	$+1
	ld de,gray1_buf+1
exec_c_shft_y_patch	equ	$+1
	ld (hl),0
	ld bc,(shft_size)
	dec bc
exec_c_shft_buf_y_patch9	equ	$
	ldir
	ret
exec_c_shft_skip2:
	ld hl,shft_temp				;copy of buffer back to start
exec_c_shft_buf_y_patch2	equ	$+1
	ld de,gray1_buf
	ld bc,(shft_size)
	ldir
	ret

exec_c_shft_y_neg:
	ex de,hl
	ld hl,0
	or a
	sbc hl,de
	ld a,h
	or a
	jp nz,err_argrange
	ld a,l
	or a
	ret z
	cp shft_y_max+1
	jp nc,err_argrange
	add hl,hl
	add hl,hl
	ld e,l
	ld d,h
	add hl,hl
	add hl,de
	ld (shft_size),hl
	ld hl,$0019	;add hl,de
	ld (exec_c_shft_buf_y_patch7),hl
	ld hl,$b0ed	;ldir
	ld (exec_c_shft_buf_y_patch8),hl
	ld hl,$b8ed	;lddr
	ld (exec_c_shft_buf_y_patch9),hl
	ld hl,gray1_buf
	ld de,gray1_buf+screen_len-1
exec_c_shft_y_patch_3	equ	$+1
	ld a,0
	call exec_c_shft_buf_y_neg
	ld hl,gray2_buf
	ld de,gray2_buf+screen_len-1
exec_c_shft_y_patch_4	equ	$+1
	ld a,0
exec_c_shft_buf_y_neg:
	ld (exec_c_shft_y_patch),a
	ld (exec_c_shft_buf_y_patch4),hl
	ld (exec_c_shft_buf_y_patch6),hl
	ld (exec_c_shft_buf_y_patch5),hl
	ex de,hl
	ld (exec_c_shft_buf_y_patch1),hl
	dec hl
	ld (exec_c_shft_buf_y_patch3),hl
	inc hl
	inc hl
	ld de,(shft_size)
	or a
	sbc hl,de
	ld (exec_c_shft_buf_y_patch2),hl
	jp exec_c_shft_buf_y


#ifdef FAST_KEY
exec_c_gkey:
	call expect_keycode
	push de
	call expect_intname
	ld (exec_c_gkey_patch),de
	pop de
	ld a,d
	call get_key
	and e
	jr nz,exec_c_gkey_no
exec_c_gkey_yes:
	ld hl,1
	jr exec_c_gkey_end
exec_c_gkey_no:
	ld hl,0
exec_c_gkey_end:
exec_c_gkey_patch	equ	$+1
	ld de,0
	call set_var_warray
	jp expect_eol
#else
exec_c_gkey:
	call expect_num						;get keycode to expect
	ld a,l
	push af
	call expect_intname					;and the var where the result will be stored
	ld (exec_c_gkey_patch),de
	call expect_eol						;no more args for us!
	ld hl,key_table+2					;now we actually have to find if the key is being pressed
	pop bc
	ld de,3								;(used to increment pointer)
exec_c_gkey_loop:
	ld a,(hl)							;use the key lookup table backwards to get the direct input values
	or a
	jp z,err_keycode
	cp b
	jr z,exec_c_gkey_end_loop
	add hl,de
	jr exec_c_gkey_loop
exec_c_gkey_end_loop:
	dec hl
	ld a,(hl)
	cpl
	ld d,a
	dec hl
	ld a,(hl)
	call get_key						;now test to see if the key is pressed
	and d
	jr z,exec_c_gkey_yes
exec_c_gkey_no:							;if not, set the var to 0
	ld hl,0
	jr exec_c_gkey_end
exec_c_gkey_yes:						;otherwise, set it to 1
	ld hl,1
exec_c_gkey_end:
exec_c_gkey_patch	equ	$+1
	ld de,0
	jp set_var_warray
#endif


exec_c_wkey:
	call expect_intname					;get the name of the var to store the keypress to
	push de
	call expect_eol
exec_c_wkey_loop:
	call get_keycode
	or a								;loop until a key has been pressed,
	jr z,exec_c_wkey_loop
	ld l,a								;then store the keycode to the var
	ld h,0
	pop de
	jp set_var_warray


exec_c_paus:
	call expect_eol
exec_c_paus_loop:
	ld a,k_g6
	call get_key						;get keypress
	cp k_enter							;if it isn't enter,
	jr nz,exec_c_paus_loop				;then loop again
	ret


exec_c_frun:
	ld de,(func_sp)						;make sure the stack isn't full
	ld hl,end_func_stack
	call _cphlde
	jp z,err_frun
	ld hl,(parse_ptr)					;if it isn't, get address of next command
	push hl
	inc hl
	inc hl
	inc hl
	ex de,hl							;and push it on the stack
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	ld (func_sp),hl
	pop hl								;now get the address of the function we're jumping to
	call _ldhlind
	ld (parse_ptr),hl					;and go to it
	ret


exec_c_fend:
	call expect_eol						;don't need no stinkin' arguments!!!
	ld de,func_stack
	ld hl,(func_sp)
	call _cphlde
	jp z,halt_prog
	dec hl
	ld d,(hl)
	dec hl
	ld e,(hl)
	ld (func_sp),hl
	ld (parse_ptr),de
	ret


exec_c_fgob:
	ld hl,(func_sp)						;first make sure we're not already at the beginning of the stack
	ld de,func_stack
	call _cphlde
	jr z,exec_c_fgob_main
	dec hl								;then get the address of the FRUN that called this function
	ld d,(hl)
	dec hl
	ld e,(hl)
	ex de,hl
	dec hl								;get to the address of this function
	dec hl
	ld d,(hl)
	dec hl
	ld e,(hl)
	ld (parse_ptr),de					;and start executing there
	ret
exec_c_fgob_main:
	ld hl,(start_addr)
	ld (parse_ptr),hl
	ret


exec_c_if:
	ld hl,(parse_ptr)					;branch depending on if strings or numbers are being compared
	ld a,(hl)
	inc hl
	cp int_char
	jr z,exec_c_if_num
	cp bitmap_char
	jr z,exec_c_if_bmp
#ifdef FULL_ERR
	cp string_char
	jp nz,err_invarg
#endif
exec_c_if_str:
	call get_str						;get a pointer to the first string
	ld de,str_temp_mem+str_max_len+1
	call _strcopy
	ld hl,(parse_ptr)					;and look at the operator between the strings
	ld a,(hl)
	inc hl
	push af
	call get_str						;now get a pointer to the second string
	ex de,hl
	call expect_eol						;(make sure that that's the end of the line)
	pop af
	ld de,str_temp_mem+str_max_len+1
	cp Lexclam
	jr z,exec_c_if_str_ne
#ifdef FULL_ERR
	cp LEQ
	jp nz,err_invop
#endif
exec_c_if_str_eq:
	call str_cmp						;compare the strings
#ifdef FULL_IF
	jp z,exec_c_if_yes
	jp exec_c_if_no
#else
	ret z
	jp find_eol
#endif
exec_c_if_str_ne:
	call str_cmp
#ifdef FULL_IF
	jp nz,exec_c_if_yes
	jp exec_c_if_no
#else
	ret nz
	jp find_eol
#endif
exec_c_if_bmp:
	call get_bmp_hl						;get the address of the first bitmap
	push bc
	ld hl,(parse_ptr)
	ld a,(hl)							;save the operator between the bitmaps
	push af
	inc hl
	call get_bmp_hl						;and get the second
	call expect_eol
	ld h,b
	ld l,c								;hl=bmp2
	pop af								;a=operator
	pop de								;de=bmp1
	cp LEQ
	jr z,exec_c_if_bmp_eq
#ifdef FULL_ERR
	cp Lexclam
	jp nz,err_invop
#endif
exec_c_if_bmp_ne:
	call _cphlde						;are the bitmaps the same?
#ifdef FULL_IF
	jp nz,exec_c_if_yes
	jr exec_c_if_no
#else
	ret nz								;if not, return to execute the next line
	jp find_eol							;if they are, skip it instead
#endif
exec_c_if_bmp_eq:
	call _cphlde
#ifdef FULL_IF
	jp z,exec_c_if_yes
	jr exec_c_if_no
#else
	ret z
	jp find_eol
#endif
exec_c_if_num:
	call get_expr_val_hl				;get the value of the first expression
	ld de,(expr_sum)
	push de
	ld a,(hl)							;then the operator
	push af
	inc hl
	call get_expr_val_hl				;then the value of the second expression
	pop bc								;now restore operator and branch depending on what it is
	ld a,b
	pop de
	push af
	ld hl,(expr_sum)					;(put value of first expression in de, second in hl)
	cp LLT
	jr z,exec_c_if_lt
	cp LGT
	jr z,exec_c_if_gt
	cp LEQ
	jr z,exec_c_if_eq
	cp Lexclam
	jr z,exec_c_if_ne
	jp err_invop						;if it isn't a valid operator, then show error
exec_c_if_lt:
	call cp_num
#ifdef FULL_IF
	jr c,exec_c_if_num_no
	jr z,exec_c_if_num_no
#else
	jr c,exec_c_if_no
	jr z,exec_c_if_no					;execute next line only if de (first expression) is smaller
#endif
	jr exec_c_if_continue
exec_c_if_gt:
	call cp_num
#ifdef FULL_IF
	jr nc,exec_c_if_num_no
#else
	jr nc,exec_c_if_no					;this time, execute next line only if de is greater than hl
#endif
	jr exec_c_if_continue
exec_c_if_eq:
	call _cphlde						;are they equal?
	jr z,exec_c_if_continue				;if so, execute the next line normally
#ifdef FULL_IF
	jr exec_c_if_num_no
#else
	jr exec_c_if_no						;otherwise, don't
#endif
exec_c_if_ne:
	call _cphlde						;are they equal?
#ifdef FULL_IF
	jr z,exec_c_if_num_no
#else
	jr z,exec_c_if_no					;if they are, skip the next line - otherwise, run it
#endif
exec_c_if_continue:
	pop af								;is the expression done?
#ifdef FULL_IF
	jr c,exec_c_if_yes
#else
	ret c								;if so, return to execute the next line
#endif
	ld hl,(parse_ptr)
	ld a,(hl)
	cp Lampersand						;otherwise, check for an '&' between expressions
	jp nz,err_invop
	inc hl
	jp exec_c_if_num					;and evaluate the next part of the expression
#ifdef FULL_IF
exec_c_if_no:
	ld hl,(parse_ptr)
	jr exec_c_if_no_continue
exec_c_if_num_no:
	ld hl,(parse_ptr)
	pop af
	jr nc,$+3
	dec hl
	call goto_eol
	inc hl
exec_c_if_no_continue:
	ld a,(hl)
	cp then_token
	jp nz,find_eol_hl					;if there isn't a THEN block, just skip the line
exec_c_if_no_loop:						;otherwise, find the end of the block
	call goto_eol
	inc hl
	ld a,(hl)
	cp ifnd_token
	jr nz,exec_c_if_no_loop
	jp find_eol_hl
exec_c_if_yes:
	ld hl,(parse_ptr)
	ld a,(hl)
	cp then_token						;is it a THEN block following?
	ret nz								;if not, just return
	inc hl
	ld (parse_ptr),hl
	jp expect_eol						;if there is, then skip the THEN token and execute the block
#else
exec_c_if_no:
	ld hl,(parse_ptr)
	pop af
	jr nc,$+3
	dec hl
	call goto_eol
	inc hl
	jp find_eol_hl
#endif


exec_c_rndm:			;rand / (65535 / (upper - lower)) + lower
	call expect_num						;get the lower limit
	ld (exec_c_rndm_patch_dn),hl
	push hl
	call expect_num						;then the upper limit
	inc hl								;(hardly ever hits the upper limit, so we have to increase it)
	ld (exec_c_rndm_patch_up),hl
	push hl								;and save them
	call expect_intname					;get variable where the random number will be stored
	ld (exec_c_rndm_var),de
	call RandomNumber					;then get the random number - first the msb
	push af
	call RandomNumber					;then the lsb
	ld c,a
	pop af
	and $03
	ld b,a
	pop hl
	pop de								;(upper limit in hl, lower limit in de, rand in bc)
	call _cphlde
	jp nc,exec_c_rndm_inrange
	push hl
	ld hl,0
	or a
	sbc hl,de
	ld (exec_c_rndm_patch_ans),hl
	ex de,hl
	pop hl
	add hl,de
	ld de,0
	jr exec_c_rndm_skip0
exec_c_rndm_inrange:
	ld ix,0
	ld (exec_c_rndm_patch_ans),ix
exec_c_rndm_skip0:
	or a
	sbc hl,de							;hl=upper - lower
	push de
	push bc
	ex de,hl							;de=upper-lower
	ld hl,$03ff							;hl=32767
	call div16							;hl=32767/(upper-lower)
	ex de,hl							;de=32767/(upper-lower)
	pop hl								;hl=rand
	call div16							;hl=rand/(32767/(upper-lower))
	pop de
	add hl,de							;hl=rand/(32767/(upper-lower))+lower
exec_c_rndm_patch_ans	equ	$+1
	ld de,0
	or a
	sbc hl,de
exec_c_rndm_patch_up	equ	$+1			;because of the loss in precision (no decimals) with large numbers, results are sometimes out of range
	ld de,0								;so we've got to check for that
	call _cphlde						;is it over the upper limit?
	jp c,exec_c_rndm_skip				;if not, everything's fine
exec_c_rndm_patch_dn	equ	$+1
	ld de,0
	call _cphlde						;is it over the lower limit?
	jp c,exec_c_rndm_skip				;if so, continue
	jp z,exec_c_rndm_skip
	ld h,d								;otherwise, set it to the lower limit
	ld l,e
exec_c_rndm_skip:
exec_c_rndm_var	equ	$+1
	ld de,0								;now get the var index
	call set_var_warray					;and write the number to the var
	jp expect_eol						;and we're done!!!


exec_c_halt:							;end prog
	call expect_eol
halt_prog:
	pop hl								;modify stack to return from executor
	ret


exec_c_wait:
	call expect_num						;get the time to wait
exec_c_wait_loop:						;and go in a pointless loop to waste time!
	dec hl
	ex (sp),hl
	ex (sp),hl
	ex (sp),hl
	ex (sp),hl
	ld a,h
	or l
	jr nz,exec_c_wait_loop
	ret


exec_c_sprt:
	call expect_sprname					;first get the name of the sprite
	push de
	call expect_bmpname					;then the bitmap
	ld b,d
	ld c,e
	pop de
	call set_spr						;set that sprite to reference the bitmap
	jp expect_eol						;that's all for now, folks!


exec_c_dbmp:
	call expect_bmpname					;get the index of the bitmap
	push de
	call expect_num						;then the x pos
	push hl
	call expect_num						;and the y pos
	ld c,l
	pop hl
	ld b,l
	pop de
	jp draw_bmp							;now draw it


exec_c_clsn:
	call expect_sprname					;get the position of the first sprite
	ld (exec_c_clsn_nosprite),de
	call get_spr
	ld a,d
	or e
	jp z,err_nospr
#ifdef CLSN_MASK
	ld (exec_c_clsn_sprite1),de
#endif
	push bc
	call get_bmp_size
	ld (exec_c_clsn_spr1size),a
	ld hl,(parse_ptr)					;(if there's a .0 instead of the second sprite, then test for collision with all others)
	ld a,(hl)
	cp spr_name_token
	jr nz,exec_c_clsn_all
	call expect_sprname					;along with the second one
	call get_spr
	ld a,d
	or e
	jp z,err_nospr
#ifdef CLSN_MASK
	ld (exec_c_clsn_sprite2),de
#endif
	push bc
	call get_bmp_size
	ld (exec_c_clsn_spr2size),a
	call expect_intname					;and the var to store the result in
	ld (exec_c_clsn_var),de
	call expect_eol
	pop hl
	pop de
	call exec_c_clsn_test				;test if there's a collision
	jr nc,exec_c_clsn_no
exec_c_clsn_yes:
	ld hl,1
	jr exec_c_clsn_end
exec_c_clsn_no:
	ld hl,0
exec_c_clsn_end:
exec_c_clsn_var	equ	$+1
	ld de,0								;and set the var accordingly
	jp set_var_warray
exec_c_clsn_all:
	call expect_num						;make sure that it's a .0
	ld a,h
	or l
	jp nz,err_invarg
	call expect_intname					;then store the result var as usual...
	ld (exec_c_clsn_var),de
	call expect_ptrname					;and the result pointer
	ld (exec_c_clsn_ptr),de
	ld a,(num_sprites)					;now get the number of sprites we'll have to test
	ld b,a								;(there can't be 0 since the first sprite has to be one of them)
	pop hl								;restore the position of the first sprite
	ld (exec_c_clsn_all_pos),hl
	ld de,spr_1_data					;and init the pointer at the beginning of the sprite table
exec_c_clsn_all_loop:
	push de
	push bc
exec_c_clsn_nosprite	equ	$+1
	ld hl,0
	call _cphlde						;if it's the base sprite, skip it
	jr z,exec_c_clsn_all_skip
	call get_spr						;otherwise, find the next sprite in the table
	ld (exec_c_clsn_sprite2),de
	call get_bmp_size
	ld (exec_c_clsn_spr2size),a			;get its size
	ld h,b								;and position
	ld l,c
exec_c_clsn_all_pos		equ	$+1
	ld de,0
	call exec_c_clsn_test				;and test it against the base sprite
	jr c,exec_c_clsn_all_yes			;if it's touching, then end the loop and set the var
exec_c_clsn_all_skip:
	pop bc								;if it isn't touching,
	pop de
	ld hl,spr_data_len
	add hl,de							;then move to the next sprite
	ex de,hl
	djnz exec_c_clsn_all_loop			;and loop
	call expect_eol
	jr exec_c_clsn_no					;if it got to here, then there must be no sprites touching it
exec_c_clsn_all_yes:
	pop bc
	pop de
	call expect_eol
	ex de,hl							;set the pointer to point to the sprite that it's touching
exec_c_clsn_ptr	equ	$+1
	ld de,0
	ld a,sprite_char
	call set_ptr
	jr exec_c_clsn_yes
exec_c_clsn_test:						;de=spr1 pos, hl=spr2 pos, c set if collision
exec_c_clsn_spr1size	equ	$+1
	ld c,0
exec_c_clsn_spr2size	equ	$+1
	ld b,0
	ld a,h								;check if the sprites overlap
	add a,b		;x2+8>x1
	cp d
	jp c,exec_c_clsn_test_no
	ld a,d
	add a,c		;x1+8>x2
	cp h
	jp c,exec_c_clsn_test_no
	ld a,l
	add a,b		;y2+8>y1
	cp e
	jp c,exec_c_clsn_test_no
	ld a,e
	add a,c		;y1+8>y2
	cp l
	jp c,exec_c_clsn_test_no
	ld a,b
	cp 16
	jp z,exec_c_clsn_test_yes
	ld a,c
	cp 16
	jp z,exec_c_clsn_test_yes
#ifdef CLSN_MASK
	xor a
	ld (exec_c_clsn_lr),a
	ld a,h								;compare sprite x positions
	cp d
	jr c,exec_c_clsn_endtestx
	ex de,hl
	ld a,$eb ;ex de,hl					;(0 - sprite 1 (hl) is left, sprite 2 (de) is right, 1 - sprite 2 (de) is left, sprite 1 (hl) is right)
	ld (exec_c_clsn_lr),a
exec_c_clsn_endtestx:					;hl=left sprite pos, de=right sprite pos
	ld a,d
	sub h								;a=amount to shift left sprite left
	ld (exec_c_clsn_sl),a
	ld b,0
	ld a,l								;now compare sprite y positions
	cp e
	jr c,exec_c_clsn_endtesty
	ex de,hl
	inc b								;(0 - sprite 1 (hl) is top, sprite 2 (de) is bottom, 1 - sprite 2 (de) is top, sprite 1 (hl) is bottom)
exec_c_clsn_endtesty:					;hl=top sprite pos, de=bottom sprite pos
	ld a,e
	sub l								;a=downwards offset for top sprite
	ld c,a
	ld a,8
	sub c
	ld (exec_c_clsn_rows),a				;calculate the number of rows to test from that
	ld a,c
	djnz exec_c_clsn_setoffset1			;if there's a difference in height, find which sprite to offset
exec_c_clsn_setoffset2:
	ld b,a
	xor a
	jr exec_c_clsn_endoffset
exec_c_clsn_setoffset1:
	ld b,0
exec_c_clsn_endoffset:					;offset the correct sprite...
	ld (exec_c_clsn_offset1),a
	ld a,b
	ld (exec_c_clsn_offset2),a
exec_c_clsn_sprite2	equ	$+1
	ld hl,0
	ld bc,8*2+1
	add hl,bc
exec_c_clsn_offset2	equ	$+1
	ld c,0
	ld b,0
	add hl,bc
	ex de,hl
exec_c_clsn_sprite1	equ	$+1
	ld hl,0
	ld bc,8*2+1
	add hl,bc							;hl=sprite 1, de=sprite 2
exec_c_clsn_offset1	equ	$+1
	ld c,0
	ld b,0
	add hl,bc
exec_c_clsn_lr	equ	$
	nop									;switch them if necessary
exec_c_clsn_rows	equ	$+1
	ld b,8
	ld a,b								;if there aren't any overlapping rows,
	or a
	jp z,exec_c_clsn_no					;then the sprites must not be overlapping!
exec_c_clsn_loop:
	push bc								;otherwise, save the number of rows left to test
	ld a,(hl)
	cpl									;then get the mask byte from the left sprite
	ld c,a
exec_c_clsn_sl	equ	$+1
	ld b,0								;and shift it the correct number of times
	ld a,b
	or a
	jr z,exec_c_clsn_loop_noshift
exec_c_clsn_shiftloop:
	sla c
	djnz exec_c_clsn_shiftloop
exec_c_clsn_loop_noshift:
	ld a,(de)							;then overlay it with the mask byte from the right-hand sprite
	cpl
	and c								;the big test...
	pop bc
	jr nz,exec_c_clsn_test_yes			;if there's an overlap, there's a collision - set the var to 1
	inc hl
	inc de
	djnz exec_c_clsn_loop				;otherwise, move on to the next row
#endif
exec_c_clsn_test_no:					;if all the rows have been tested already, none must be overlapping, so set the var to a 0
	xor a
	ret
exec_c_clsn_test_yes:
	scf
	ret


file_pos	equ	temp_buf
new_size	equ	temp_buf+2
var_type	equ	temp_buf+4

exec_c_wrte:
	call expect_num						;get position in file to write
	ld a,h								;(make sure that it's in range)
	or a
	jp nz,err_argrange
	ld (file_pos),hl
	call expect_varname					;and the data to write to it
	ld (exec_c_wrte_vartype),a
	cp string_char						;check what type of data it is
	jr z,exec_c_wrte_str
	call get_var
	cp array_char
	jr z,exec_c_wrte_array
#ifdef FULL_ERR
	cp int_char
	jp nz,err_vartype
#endif
exec_c_wrte_int:
	ld bc,2								;we know how long an int is!
	ld (wrte_temp),hl
	ld hl,wrte_temp
	jr exec_c_wrte_go
exec_c_wrte_str:
	ex de,hl
	call _strlength						;if it's a string, get the length of the string
	jr exec_c_wrte_go
exec_c_wrte_array:
	ld a,(hl)							;get the length of the array from the size byte
	cp array_write_len+1				;(make sure it's not too big to write)
	jp nc,err_mem_file
	rla
	inc a
	ld c,a
	ld b,0
	push bc								;and copy the array to a temp area
	ld de,wrte_temp
	ldir
	pop bc
	ld hl,wrte_temp
exec_c_wrte_go:		;hl=data start, bc=data len
	di										;disable the interrupts so the table doesn't get screwed by the page swaps
	push hl
	push bc
	ld hl,curr_name
	rst 20h									;get the size of the string var
	rst 10h
	call _data_size_to_de					;de=size of string
	push de
	ld hl,(file_pos)
	inc hl
	ld de,data_seg_len
	call mult16
	pop de									;hl=required size of string, de=size of string
	call _cphlde
	jr c,exec_c_wrte_sizefine				;if the string has enough space, then go ahead and write
	jr z,exec_c_wrte_sizefine
	ld (new_size),hl						;hl=new size of string
	call _enoughmem
	jp c,err_mem_file						;make sure there's enough free memory
	ld hl,temp_name-1
;	rst 20h
;	ld hl,(new_size)
;	call _createstrng						;then create a new temp var
;	call _ex_ahl_bde
;	call _ahl_plus_2
;	call _set_abs_dest_addr
	call exec_c_wrte_create_new_dest

	ld hl,curr_name
;	rst 20h
;	rst 10h
;	call _ex_ahl_bde
;	call _ahl_plus_2						;and copy the contents of the string to it
;	call _set_abs_src_addr
;	ex de,hl
;	call _data_size_to_de
;	ex de,hl
;	xor a
;	call _set_mm_num_bytes
;	call _mm_ldir
	call exec_c_wrte_set_src_copy

	ld hl,curr_name							;now delete the original var
;	rst 20h
;	rst 10h
;	call _delvar
	call exec_c_wrte_del

	ld hl,curr_name							;create a new var with the same name
;	rst 20h
;	ld hl,(new_size)
;	call _createstrng
;	call _ex_ahl_bde
;	call _ahl_plus_2
;	call _set_abs_dest_addr
	call exec_c_wrte_create_new_dest

	ld hl,temp_name-1						;copy the contents of the temp var to it
;	rst 20h
;	rst 10h
;	call _ex_ahl_bde
;	call _ahl_plus_2
;	call _set_abs_src_addr
;	ex de,hl
;	call _data_size_to_de
;	ex de,hl
;	xor a
;	call _set_mm_num_bytes
;	call _mm_ldir
	call exec_c_wrte_set_src_copy

	ld hl,temp_name-1						;and delete the temp var
;	rst 20h
;	rst 10h
;	call _delvar
	call exec_c_wrte_del

exec_c_wrte_sizefine:
	ld hl,curr_name							;find the position in the string to write to
;	rst 20h
;	rst 10h
;	call _ex_ahl_bde
;	call _ahl_plus_2
	call exec_c_wrte_get_var_start

	push af
	push hl
	ld de,(file_pos)
	ld hl,data_seg_len
	call mult16
	pop de
	pop af
	add hl,de
	adc a,0
	push af
	push hl
	call _load_ram_ahl
exec_c_wrte_vartype	equ	$+1
	ld (hl),0								;and write the type byte
	pop hl
	pop af
	call _inc_ptr_ahl
	call _set_abs_dest_addr					;then the data!!!
	pop bc
	ld h,b
	ld l,c
	xor a
	call _set_mm_num_bytes
	pop hl
	call conv_to_abs_addr
	call _set_abs_src_addr
	call _mm_ldir
	call _ram_page_1						;reset the ram page after the copy
	ei										;and enable the interrupts again
	jp expect_eol							;AHHH...WE'RE DONE!!!!
;TEST
exec_c_wrte_create_new_dest:
	rst 20h
	ld hl,(new_size)
	call _createstrng
	call _ex_ahl_bde
	call _ahl_plus_2
	jp _set_abs_dest_addr
exec_c_wrte_set_src_copy:
	call exec_c_wrte_get_var_start
	call _set_abs_src_addr
	ex de,hl
	call _data_size_to_de
	ex de,hl
	xor a
	call _set_mm_num_bytes
	jp _mm_ldir
exec_c_wrte_del:
	rst 20h
	rst 10h
	jp _delvar
exec_c_wrte_get_var_start:
	rst 20h
	rst 10h
	call _ex_ahl_bde
	jp _ahl_plus_2
;END TEST


exec_c_read:
	call expect_num							;get the position to read from
	ld a,h									;make sure it isn't out of range,
	or a
	jp nz,err_argrange
	ld (file_pos),hl						;then save it
	call expect_varname						;get the var which will be holding the result
	ld (curr_var),de						;and save that too
	ld (var_type),a							;along with the vartype
	di										;(again, disable interrupts because of page swaps)
	ld hl,curr_name							;now check if the file position is out of range
	rst 20h
	rst 10h
	push bc
	push de
	call _data_size_to_de
	ex de,hl
	ld de,data_seg_len
	call div16
	ld de,(file_pos)						;hl=max position in file, de=required position
	inc de
	call _cphlde
	jp c,exec_c_read_blank					;if input past end of file, return blank var
	ld hl,(file_pos)						;if not, find the correct position in the file to read from
	ld de,data_seg_len
	call mult16
	pop de
	pop af
	add hl,de
	adc a,0									;ahl=read pos
	call _ahl_plus_2
	push af
	push hl
	call _load_ram_ahl						;get the type of the var saved at that position
	ld a,(var_type)							;make sure it matches the type of the var we're reading to
	cp (hl)
	jp nz,err_vartype						;if it doesn't, show error
	pop hl
	pop af
	call _inc_ptr_ahl						;ahl=start of data
	call _set_abs_src_addr					;set that as source for copy
	ld a,(var_type)							;now branch depending on the vartype
	cp int_char
	jr z,exec_c_read_int
	cp array_char
	jr z,exec_c_read_array
#ifdef FULL_ERR
	cp string_char
	jp nz,err_vartype
#endif
exec_c_read_string:
	call _ram_page_1
	ld hl,(curr_var)						;get the address of the string var where the output is going to be stored
	call conv_to_abs_addr					;convert that to an abs address
	call _set_abs_dest_addr					;to set it as the destination for the copy
	ld hl,str_max_len						;setting the length of the copy is easy - we know the max length for a string var already
	xor a
	call _set_mm_num_bytes
	jr exec_c_read_go
exec_c_read_array:
	call _load_abs_src_addr
	call _load_ram_ahl
	ld l,(hl)								;get the length of the array in the string
	ld a,l
	push af
	ld h,0
	add hl,hl
	inc hl
	xor a
	call _set_mm_num_bytes					;use that to set the copy length
	call _ram_page_1
;	ld a,l
;	push af
;	add hl,hl								;and to create a new array var
;	inc hl
	ld b,h
	ld c,l
	call alloc_mem
	pop af
	ld (hl),a
	inc hl
	call conv_to_abs_addr
	call _set_abs_dest_addr					;and store its address
	ld de,(curr_var)
	call set_var
	jr exec_c_read_go
exec_c_read_int:
	ld hl,(curr_var)
	call conv_to_abs_addr
	call _set_abs_dest_addr
	ld hl,2
	xor a
	call _set_mm_num_bytes
exec_c_read_go:
	call _mm_ldir							;finally, copy the actual data!!!
	call _ram_page_1						;restore settings
	ei
	jp expect_eol							;and end
exec_c_read_blank:
	pop hl
	pop hl
	call _ram_page_1						;if we have to return a blank var,
	call expect_eol
	ei
	ld a,(var_type)							;check what type of var it is
	cp int_char
	jr z,exec_c_read_blank_int
	cp array_char
	jr z,exec_c_read_blank_array
#ifdef FULL_ERR
	cp string_char
	jp nz,err_vartype
#endif
exec_c_read_blank_str:
	ld hl,(curr_var)
	ld (hl),0								;if it's a string, put a terminating char at the beginning
	ret
exec_c_read_blank_array:
	ld de,(curr_var)						;if it's an array, set the length to 0
	ld hl,blank_data
	jp set_var
exec_c_read_blank_int:
	ld hl,0									;if it's an int, just set it to 0
	ld de,(curr_var)
	jp set_var


exec_c_polr:
	call expect_num							;first get the length of the vector
	ld (temp_buf),hl
	call expect_num							;then the angle
	call is_hl_neg							;(compensate for negative angles)
	jr c,exec_c_polr_skip
	ex de,hl
	ld hl,360
	or a
	sbc hl,de
exec_c_polr_skip:
	ld de,361								;(and make sure it's in range)
	call _cphlde
	jp nc,err_argrange
	ld de,trig_precision					;adjust the precision for the lookup table
	call div16
	add hl,hl								;*2
	add hl,hl								;*4 (4 bytes for every entry)
	ld de,trig_table
	add hl,de								;get entry in trig table
	ld e,(hl)								;first get the cos value...
	inc hl
	ld d,(hl)
	inc hl
	push hl
	ld hl,(temp_buf)
	call mult16								;hl=r*(cos [theta]*1000)
	ld de,1000
	call div16								;hl=r*cos [theta]
	push hl
	call expect_intname						;now get the var to store to the x pos
	pop hl
	call set_var_warray						;and store it
	pop hl
	ld e,(hl)								;now get the sin value
	inc hl
	ld d,(hl)
	ld hl,(temp_buf)
	call mult16								;hl=r*(sin [theta]*1000)
	ld de,1000
	call div16								;hl=r*sin [theta]
	push hl
	call expect_intname						;and store that to the y var
	pop hl
	call set_var_warray
	jp expect_eol							;...that wasn't so hard...


exec_c_rset:
	ld hl,func_stack						;first reset the function stack
	ld (func_sp),hl
	call clr_free_mem						;then reset all vars too
	jp exec_reset


exec_c_subs:
	call expect_str							;get the string to separate
	push hl
	call expect_num							;now the starting position
	push hl
	call expect_num							;and the ending position
	pop de
	or a									;(make sure that the ending position is AFTER the starting position)
	sbc hl,de
	jp c,err_argrange
	inc hl
	push hl									;(and that the string isn't too long for a string var)
	ld bc,str_max_len+1
	or a
	sbc hl,bc
	pop hl
	jp nc,err_strlen
	ld b,h
	ld c,l
	pop hl
	add hl,de								;hl=starting address of segment, bc=length of segment
	push hl
	push bc
	call expect_strname						;get the name of the string var where the result will be stored
	pop bc
	pop hl									;hl=start of segment, bc=length of segment, de=destination
	ldir									;copy the string segment
	xor a
	ld (de),a								;and zero-terminate it
	jp expect_eol


exec_c_cint:
	call expect_str							;get the address of the string
	push hl
	call expect_intname						;and the name of the var where the result will be stored
	ld (exec_c_cint_var),de
	pop hl									;now parse the string!
	call str_to_num
exec_c_cint_var	equ	$+1
	ld de,0									;finally, set the var to the parsed number
	call set_var_warray
	jp expect_eol


exec_c_absl:
	call expect_num							;get the number
	call is_hl_neg
	jr c,exec_c_absl_set					;if it isn't negative, then it's fine as it is
	ld de,0
	ex de,hl
	or a
	sbc hl,de								;otherwise, invert it
exec_c_absl_set:
	push hl
	call expect_intname						;get the name of the var to store it in
	pop hl
	call set_var_warray
	jp expect_eol


exec_c_inpt:
	call expect_x_pos						;first get the x pos of the prompt
	ld (_pencol),a
	call expect_y_pos						;then the y pos
	ld (_penrow),a							;and set the cursor
	call expect_str							;then get a pointer to the prompt text
	push hl
	call expect_color						;and find the color it should be drawn in
	pop hl
	or a
	jr z,exec_c_inpt_white
	dec a
	jr z,exec_c_inpt_lgray
	dec a
	jr z,exec_c_inpt_dgray
	dec a
	jr z,exec_c_inpt_black
exec_c_inpt_invert:
	ld bc,pixel_inv
	ld (pixel_op_patch),bc
	ld (Input_Str_Layer1),bc
	ld (Input_Str_Layer2),bc
	call draw_text_all
	jr exec_c_inpt_draw_end
exec_c_inpt_white:
	ld bc,pixel_off
	ld (pixel_op_patch),bc
	ld (Input_Str_Layer1),bc
	ld (Input_Str_Layer2),bc
	call draw_text_all
	jr exec_c_inpt_draw_end
exec_c_inpt_lgray:
	ld bc,pixel_on
	ld (pixel_op_patch),bc
	ld (Input_Str_Layer1),bc
	call draw_text_1
	ld bc,pixel_off
	ld (pixel_op_patch),bc
	ld (Input_Str_Layer2),bc
	call draw_text_2
	jr exec_c_inpt_draw_end
exec_c_inpt_dgray:
	ld bc,pixel_off
	ld (pixel_op_patch),bc
	ld (Input_Str_Layer1),bc
	call draw_text_1
	ld bc,pixel_on
	ld (pixel_op_patch),bc
	ld (Input_Str_Layer2),bc
	call draw_text_2
	jr exec_c_inpt_draw_end
exec_c_inpt_black:
	ld bc,pixel_on
	ld (pixel_op_patch),bc
	ld (Input_Str_Layer1),bc
	ld (Input_Str_Layer2),bc
	call draw_text_all
;	jr exec_c_inpt_draw_end
exec_c_inpt_draw_end:
	ld a,(_pencol)
	add a,b
	ld (_pencol),a
	call expect_varname						;after drawing the prompt, get the var where the input will be stored
	cp array_char
	jr z,exec_c_inpt_num
	cp int_char
	jr z,exec_c_inpt_num
	cp string_char
	jp nz,err_vartype
exec_c_inpt_str:
	ex de,hl								;if it's a string var, just store the string directly to it
	push hl
	call Input_Str
	pop de
	or a									;then make sure that it isn't too long
	sbc hl,de
	ld de,str_max_len+1
	call _cphlde
	jp nc,err_strlen
	jp expect_eol
exec_c_inpt_num:
	push de									;if it's a number,
	ld hl,temp_buf							;get the string in the temp buffer
	call Input_Str
	ld hl,temp_buf							;then convert it
	call str_to_num
	pop de
	call set_var_warray						;and store it to the var
	jp expect_eol


exec_c_nmap:
	call expect_num							;get the x size
	push hl
	call expect_num							;and the y size
	push hl
	call expect_mapname						;then the address of the map to create
	ld (exec_c_nmap_var),de
	call get_var
	ld a,h
	or l
	jp nz,err_twovar_arry					;(make sure it hasn't been created already)
	pop de
	pop hl
	ld a,h									;make sure they're in range
	or a
	jp nz,err_maplen
	or l
	jp z,err_maplen
	ld a,d
	or a
	jp nz,err_maplen
	or e
	jp z,err_maplen
	ld b,l
	ld c,e
	push bc									;then save the sizes
	call mult16								;find the total length
	add hl,hl								;*2
	ld bc,4
	add hl,bc								;*2+4
	ld b,h
	ld c,l
	call alloc_mem							;so that we can allocate the memory for it
	pop bc
	push hl
	inc hl
	inc hl
	ld (hl),b								;set the size of the map from the user's parameters
	inc hl
	ld (hl),c
	pop hl
exec_c_nmap_var	equ	$+1
	ld de,0
	call set_var							;then set the map var to point to the newly-created map in memory
	jp expect_eol


exec_c_tile:
	call expect_mapname						;first get the name of the map where the solid tiles will be set
	call get_var
	ld a,h
	or l
	jp z,err_noarray
	inc hl
	inc hl
	ld c,(hl)
	inc hl
	ld b,(hl)								;then get the x and y sizes of the map
	inc hl
	ld (exec_c_tile_var),hl
	ld h,b									;and multiply them to get the full length of the map data
	ld l,c
	call _htimesl
	ld (exec_c_tile_ctr),hl
exec_c_tile_loop:
	ld hl,(parse_ptr)						;now check on the next argument
	ld a,(hl)
	cp eol_token							;if it's the end of the line,
	jp z,expect_eol							;then we're done
	call expect_bmpname						;otherwise, get the address of the bitmap to set as solid
exec_c_tile_var	equ	$+1
	ld hl,0									;then search for it in the current map
exec_c_tile_ctr	equ	$+1
	ld bc,0
exec_c_tile_search_loop:					;bc=number of tiles to search, hl=start of map data, de=tile to set as solid
	push hl
	call _ldhlind							;and get the address of the tile in the map
	ld a,h
	and $0f
	or $a0
	ld h,a
	call _cphlde							;is it one of the bitmaps that should be solid?
	pop hl
	inc hl
	call z,exec_c_tile_set_solid			;if it is, then go back and set the "solid/transparent" byte to 1
	inc hl									;move on to the next tile
	dec bc									;do this until all the tiles in the map have been tested
	ld a,b
	or c
	jr nz,exec_c_tile_search_loop
	jr exec_c_tile_loop						;when that's done, find the next bitmap that should be solid
exec_c_tile_set_solid:
	ld a,$10
	or (hl)
	ld (hl),a
	ret


exec_c_cmap:
	call expect_mapname						;of course, the name of the MAP TO CHECK is the most important...
	call get_var
	ld a,h
	or l
	jp z,err_noarray
	ld a,(hl)								;get the map position (which is all we'll need for figuring out which element is touching...)
	ld (exec_c_cmap_x),a
	inc hl
	ld a,(hl)
	ld (exec_c_cmap_y),a
	call expect_sprname						;get the address of the sprite to check
	call get_spr							;and find its coordinates
	ld a,d									;(make sure it's been defined)
	or e
	jp z,err_nospr
	push bc
	call expect_arrayname					;then get the addresses of the arrays to store the positions
	push de
	call get_var
	pop de
	ld a,h
	or l
	jr nz,exec_c_cmap_skip_x
	push de
	ld bc,9									;and allocate a 4-element array for each of them
	call alloc_mem
	pop de
	ld (hl),4
	push hl
	call set_var
	pop hl
exec_c_cmap_skip_x:
	push hl
	call expect_arrayname
	push de
	call get_var
	pop de
	ld a,h
	or l
	jr nz,exec_c_cmap_skip_y
	push de
	ld bc,9
	call alloc_mem
	pop de
	ld (hl),4
	call set_var
	ex de,hl
exec_c_cmap_skip_y:
	ex de,hl
	pop hl									;de=address of y array, hl=address of x array
	inc hl
	inc de
	pop bc									;bc=sprite position
	call exec_c_cmap_get_tilenums			;first store the position of the top-left corner
	push bc
	ld a,b
	add a,7
	ld b,a
	call exec_c_cmap_get_tilenums			;then the top-right corner
	pop bc
	ld a,c
	add a,7
	ld c,a
	call exec_c_cmap_get_tilenums			;then the bottom-left
	ld a,b
	add a,7
	ld b,a
	call exec_c_cmap_get_tilenums			;then the bottom-right
	jp expect_eol							;and that's all
exec_c_cmap_get_tilenums:
exec_c_cmap_x	equ	$+1
	ld a,0
	sub b
	neg										;a=spr_x-map_x
	and %11111000
	rra
	rra
	rra										;a=(spr_x-map_x)/8
	ld (hl),a
	inc hl
	ld (hl),0
	inc hl
	ex de,hl
exec_c_cmap_y	equ	$+1
	ld a,0
	sub c
	neg										;a=spr_y-map_y
	and %11111000
	rra
	rra
	rra										;a=(spr_y-map_y)/8
	ld (hl),a
	inc hl
	ld (hl),0
	inc hl
	ex de,hl
	ret


exec_c_tchk:
	call expect_mapname						;obviously, we'll need the address of the map var
	ld (exec_c_tchk_map),de
	call get_var
	ld a,h
	or l
	jp z,err_noarray
	call expect_arrayname					;get address of x array
	call get_var
	ld a,h
	or l
	jp z,err_noarray
	push hl
	call expect_arrayname					;then the y array
	call get_var
	ld a,h
	or l
	jp z,err_noarray
	push hl
	call expect_intname						;then the int where the result will be stored
	ld (exec_c_tchk_var),de
	pop de
	pop hl
	ld a,(de)
	cp (hl)									;make sure the arrays have the same size
	jp nz,err_arrayindex	;??? find better error message
	ld b,a
	inc hl
	inc de									;hl=x array, de=y array, b=loop counter
exec_c_tchk_loop:
	push bc
	push hl
	push de
	ex de,hl
	ld c,(hl)
	inc hl
	ld b,(hl)
	ex de,hl
	call _ldhlind							;get the positions from the arrays
exec_c_tchk_map	equ	$+1
	ld de,0									;and find the address of that element on the current map
	call find_map_element
	inc hl
	ld a,(hl)
	and $f0									;now for the big question - IS THE TILE SOLID???
	jr nz,exec_c_tchk_yes					;if it is, exit the loop and set the var
	pop de
	pop hl
	pop bc
	inc hl									;otherwise, move to the next array element
	inc hl
	inc de
	inc de
	djnz exec_c_tchk_loop
exec_c_tchk_no:								;if all of the positions have been checked and none are solid,
	ld hl,0									;then set the int to 0
	jr exec_c_tchk_end
exec_c_tchk_yes:
	pop de
	pop hl
	pop bc
	ld hl,1									;otherwise, if one of the tiles was solid, set the int to 1
exec_c_tchk_end:
exec_c_tchk_var	equ	$+1
	ld de,0
	call set_var_warray						;finally, set the result int
	jp expect_eol


exec_c_dmap:
	call expect_num							;get the position where the map will be displayed
	ld a,l
	push af
	call expect_num
	pop bc
	ld c,l									;bc=on-screen pos
	ld a,l
	push bc
	call expect_mapname						;now get the address of the map to draw
	pop bc
	call get_var
	ld a,h
	or l
	jp z,err_noarray
	ld (hl),b								;set its position,
	inc hl
	ld (hl),c
	inc hl
	ld a,(hl)								;and find its size
	ld d,a
	ld (exec_c_dmap_width),a
	ld a,b
	ld (exec_c_dmap_x),a
	inc hl
	ld e,(hl)
	inc hl
exec_c_dmap_loop:							;b,c=x,y pos	d,e=x,y size	hl=start of map pos
	push bc
	push de
	push hl
	call _ldhlind							;get the bitmap address
	ld a,h
	or $a0
	ld h,a
	ld a,(hl)
	cp 8									;make sure that it's an 8x8 bitmap
	jp nz,err_map8
	inc hl
	call VeryFastGraySprite8				;then draw it!
	pop hl
	pop de
	pop bc
	inc hl
	inc hl									;point to the next bitmap,
	ld a,b
	add a,8
	ld b,a									;update the position,
	dec d									;and decrement the column counter
	jr nz,exec_c_dmap_loop					;if it's not done with this row, loop again
exec_c_dmap_x	equ	$+1
	ld b,0									;when at the end of the row, re-init the x position
exec_c_dmap_width	equ	$+1
	ld d,0									;and the width (column counter)
	ld a,c									;move to the next line
	add a,8
	ld c,a
	dec e									;and decrement the row counter
	jr nz,exec_c_dmap_loop					;again, if it's not at the end, move on to the next row
;	jr expect_eol

;//////////////////////////////Argument retrieval/processing routines//////////////////////////////

expect_eol:							;if not end of line, show error
	ld hl,(parse_ptr)
	ld a,(hl)							;look at the next token
	cp eol_token
	jp nz,err_eol						;if it isn't the end of the line, show error
	inc hl								;otherwise, move pointer to next line
	ld (parse_ptr),hl
	ret


expect_num:							;get number or var in hl, show error if not number
	call get_expr_val
	ld hl,(expr_sum)
	ret


expect_str:							;get pointer to string in hl, show error if not string
	ld hl,(parse_ptr)
	jp get_str


expect_intname:						;get var address in de, show error if not var, if array element, index returned in expr_sum
	call read_varname
	ld (curr_var_type),a
	cp int_name_token					;return if it's just the name of an integer
	ret z
	cp ptr_name_token
	jr z,expect_intname_ptr
	cp array_name_token
	jp nz,err_var_int
	ld a,(hl)
	cp index_start_token
	jp nz,err_var_int
	push de
	inc hl
	call get_expr_val_hl				;otherwise, get the array index
	ld hl,(expr_sum)
	ld (curr_var_index),hl				;and store it
	ld hl,(parse_ptr)
#ifdef FULL_ERR
	ld a,(hl)
	cp index_end_token
	jp nz,err_syntax
#endif
	inc hl
	ld (parse_ptr),hl
	pop de
	ret
expect_intname_ptr:						;if it's not a varname, it's also allowed to be a pointer
	call get_ptr						;so get the pointer type
	or a
	jp z,err_noptr
	cp int_char
	jp nz,err_var_int
	ld (curr_var_type),a
	ex de,hl							;then get the var's address in de
	ret									;(right where it should be) and return


expect_arrayname:					;get var address in de, show error if not array
	call read_varname
	cp array_name_token
	ret z
expect_arrayname_ptr:
	cp ptr_name_token
	jp nz,err_var_arry
	call get_ptr
	or a
	jp z,err_ptr
	cp array_char
	jp nz,err_var_arry
	ex de,hl
	ret


expect_strname:						;get var address in de, show error if not string
	call read_varname
	cp str_name_token
	ret z
expect_strname_ptr:
	cp ptr_name_token
	jp nz,err_var_str
	call get_ptr
	or a
	jp z,err_noptr
	cp string_char
	jp nz,err_var_str
	ex de,hl
	ret


expect_varname:						;get var address in de, var type (char) in a
	call read_varname
	ld (curr_var_type),a
	call tok_to_char
	ld a,b
	cp pointer_char
	ret nz
expect_varname_ptr:
	call get_ptr
	or a
	jp z,err_noptr
	ex de,hl
	ret


expect_sprname:						;get sprite address in de, show error if not sprite
	call read_varname
	cp spr_name_token
	ret z
expect_sprname_ptr:
	cp ptr_name_token
	jp nz,err_spr
	call get_ptr
	or a
	jp z,err_noptr
	cp sprite_char
	jp nz,err_spr
	ex de,hl
	ret


expect_bmpname:
	call get_bmp
	ld d,b
	ld e,c
	ret
;	call read_varname
;	cp bmp_name_token
;	ret z
;expect_bmpname_ptr:
;	cp ptr_name_token
;	jp nz,err_bmp
;	call get_ptr
;	ex de,hl
;	or a
;	jp z,err_noptr
;	cp bitmap_char
;	ret z
;	cp sprite_char
;	jp nz,err_bmp
;	ld hl,(parse_ptr)
;	ld a,(hl)
;	cp LcapB
;	jp nz,err_bmp
;	inc hl
;	ld (parse_ptr),hl
;	jp get_spr


expect_ptrname:
	call read_varname
	cp ptr_name_token
	ret z
	jp err_ptr


expect_mapname:
	call read_varname
	cp map_name_token
	ret z
expect_mapname_ptr:
	cp ptr_name_token
	jp nz,err_map
	call get_ptr
	or a
	jp z,err_noptr
	cp map_char
	jp nz,err_map
	ex de,hl
	ret


#ifdef FAST_KEY
expect_keycode:
	call read_varname
	cp keycode_token
	ret z
	jp err_num_key
#endif


read_varname:
	ld hl,(parse_ptr)
	ld a,(hl)							;get first the type token
	inc hl
	ld e,(hl)							;then the address
	inc hl
	ld d,(hl)
	inc hl
	ld (parse_ptr),hl					;and update the parse pointer
	ret


expect_x_pos:
	call expect_num
	ld a,h								;make sure that it isn't out of bounds
	or a
	jr nz,find_eol
	ld a,l
	cp 96
	ret c
	jr find_eol

expect_y_pos:
	call expect_num
	ld a,h
	or a
	jr nz,find_eol
	ld a,l
	cp 64
	ret c
	jr find_eol


expect_color:
	call expect_num
format_color:
	ld a,h								;yet again, make sure it isn't out of range...
	or a
	jp nz,err_argrange
	ld a,l
	cp num_colors+1
	jp nc,err_argrange
	ret


;//////////////////////////////Other commonly used routines//////////////////////////////

find_eol:								;skip the current line being executed
	ld hl,(parse_ptr)
find_eol_hl:
	call goto_eol
	inc hl
	ld (parse_ptr),hl
	ld sp,(exec_sp)
	jp exec_loop


read_var_addr:							;read var address from (hl), store in curr_var
	inc hl
	ld e,(hl)
	inc hl
	ld d,(hl)
	ld (curr_var),de
	inc hl
	ret


goto_eol:								;start of current line is hl, return end of line in hl
	ld a,(hl)
	cp eol_token
	ret z
	cp eof_token
	jp z,err_syntax
	cp frun_token
	jp z,goto_eol_3
	cp lit_str_token
	jr z,goto_eol_str
	cp lit_num_token
	jr c,goto_eol_1
goto_eol_3:
	inc hl
	inc hl
goto_eol_1:
	inc hl
	jp goto_eol
goto_eol_str:
	ld bc,str_max_len+1
	xor a
	cpir
	jr goto_eol


find_map_element:						;find map element with x position in hl, y position in bc, var address in de, return address in hl
	push hl
	push bc
	call get_var
	inc hl
	inc hl
	ld b,(hl)							;get the size of the map
	inc b
	inc hl
	ld c,(hl)
	inc c
	inc hl
	ld (find_map_element_addr),hl
	pop hl								;hl=y pos
	ld a,h
	or a
	jp nz,err_arrayindex
	ld a,l
	cp c
	jp nc,err_arrayindex
	pop de								;de=x pos
	ld a,d
	or a
	jp nz,err_arrayindex
	ld a,e
	cp b
	jp nc,err_arrayindex
	push de
	ld h,b
	dec h
	call _htimesl						;hl=y*width
	pop de
	add hl,de							;hl=x+(y*width)
	add hl,hl
find_map_element_addr	equ	$+1
	ld de,0
	add hl,de							;hl=address of map element
	ret


tok_to_char:							;var type token in a, char returned in b
	ld b,map_char
	sub map_name_token
	ret z
	ld b,sprite_char
	dec a
	ret z
	ld b,bitmap_char
	dec a
	ret z
	ld b,pointer_char
	dec a
	ret z
	ld b,string_char
	dec a
	ret z
	ld b,array_char
	dec a
	ret z
	ld b,int_char
	dec a
	ret z
	jp err_invarg


format_spr_pos:							;sprite position in hl, returned in a
	call is_hl_neg
	jr c,format_spr_pos_skip
	ld a,h
	cp $ff
	jp nz,err_argrange
	ld a,l
	cp 128
	jp c,err_argrange
	ld a,l
	ret
format_spr_pos_skip:
	ld a,h
	or a
	jp nz,err_argrange
	ld a,l
	ret

.end