;version 2.6
;3/31/2005 build 1
;(17 files, 9000 lines of code!!!)
;9.616 KB for prog

;MLC (Multi-platform Language for Calculators) by Epic Programming Studios
;TI-86 interpreter
;Written by Donald Straney (burntfuse) - straney@umbi.umd.edu

;Main code

.plugin bin
#include "mlcdata.inc"

;COMPILE OPTIONS

;#define FULL_ERR							;redundant error checking (55 bytes)
#define FULL_IF								;IF-THEN-IFND blocks supported
#define FAST_LINES							;optimized line drawing (78 bytes)
#define FAST_KEY							;pre-processed keypress detection (56 bytes)
#define CLSN_MASK							;CLSN command compares masks of overlapping bitmaps (139 bytes)
#define EMU									;runs fully on VTI 2.5 (1 byte)

;PROGRAM SETTINGS

cmd_len			equ		4					;number of letters in each command
num_max_len		equ		5					;max length of number
str_max_len		equ		35					;max length of string
name_max_len	equ		5					;max length of any name (function, var, sprite, etc.)
num_colors		equ		5					;number of colors
max_vars		equ		110					;max number of vars
max_lbls		equ		125					;max number of labels
max_sprites		equ		15
arg_separator	equ		Lspace				;char used to separate arguments of commands
index_start_char	equ	LlParen				;char used for defining index of an array
index_stop_char	equ		LrParen
command_char	equ		Lpound				;char that precedes commands
string_char		equ		Ldollar				;chars that precede different var/literal types
int_char		equ		Lpercent
num_char		equ		Lperiod
array_char		equ		LatSign
sprite_char		equ		Lcaret
bitmap_char		equ		LlBrack
pointer_char	equ		Lasterisk
map_char		equ		LlBrace
data_token		equ		0					;tokens for data areas
dend_token		equ		1
if_token		equ		22					;conditional tokens
then_token		equ		23
ifnd_token		equ		24
fnct_token		equ		18					;function tokens
frun_token		equ		19
bitm_token		equ		34
gkey_token		equ		14
index_start_token	equ	242					;token put at beginning of array/map index expression
index_end_token	equ		243					;token put at end of array/map index expression
lit_str_token	equ		244					;tokens used to identify arguments
lit_num_token	equ		245
keycode_token	equ		246
map_name_token	equ		247
spr_name_token	equ		248
bmp_name_token	equ		249
ptr_name_token	equ		250
str_name_token	equ		251
array_name_token	equ	252
int_name_token	equ		253
eol_token		equ		254					;token put at end of line
eof_token		equ		255					;token put at end of file
error_str_pos	equ		$0000				;cursor position of error message
error_str_pos_2	equ		$0600
error_str_pos_3	equ		$1200
key_delay		equ		12000				;length of delay between keypresses in INPT command
screen_len		equ		768					;length of screen in bytes
full_screen_len	equ		1024
data_seg_len	equ		52					;length of each var segment in data string
array_write_len	equ		25					;max length of array that can be written to file
trig_precision	equ		15					;precision of trig table in degrees
shft_y_max		equ		20					;maximum distance SHFT can be used for vertical shifts

;MEMORY ALLOCATION

;permanent memory
copy_addr		equ		$8060				;address where the prog to be executed is copied
end_copy_addr	equ		$a5f0
write_addr		equ		$8030				;address where prog is copied to be written back
scrn1_buf		equ		$ca00				;gfx buffers which are actually drawn to screen
scrn2_buf		equ		$fc00
free4_mem		equ		$b000		;300	areas for storing vars, pointers, and sprites
end_free4_mem	equ		$b000+300
free3_mem		equ		$c415		;319
end_free3_mem	equ		$c415+319
free2_mem		equ		$ceef		;180
end_free2_mem	equ		$ceef+180
free1_mem		equ		$c920		;180
end_free1_mem	equ		$c920+180
data_copy_addr	equ		$a9e0		;1556	area where system data is copied
bmp_mem			equ		$a5f0		;1000	area used to store bitmap data
end_bmp_mem		equ		$a5f0+1000
spr_mem			equ		$cfc9		;181
func_stack		equ		$c013		;100	stack used to call functions
end_func_stack	equ		$c013+100			;end of function stack
curr_name		equ		$c386		;10		name of current prog's data string
;execution only
vector_table_addr	equ	$c200		;257	address of vector table for isr
isr_addr		equ		$c8c8				;address of grayscale isr
gray2_buf		equ		$bb00		;768	;gfx buffers
gray1_buf		equ		$b800		;768
scr_save2		equ		$b500		;768	buffers where screen is saved for STOP and RCLP
scr_save1		equ		$b200		;768
shft_temp		equ		$b130		;240	(used during y screen scrolling)
bit_save		equ		$b130		;1		(used during x screen scrolling)
wrte_temp		equ		$b130		;		(used while writing to file)
str_temp_ptr	equ		$b130		;2		(used while parsing string expressions)
str_temp_mem	equ		$b132		;35
;tokenizing only
var_table		equ		$fc00		;991	table used to keep track of vars used in prog
lbl_table		equ		$ca00		;1000	table used to keep track of labels used in prog
lbl_res_ptr		equ		$bb80		;2		(used when resolving labels)
lbl_res_mem		equ		$bb82		;750
spr_temp_mem	equ		$be70		;24		(used when parsing sprite data)
temp_expr_buf	equ		$be70		;30		area for parsing array args during tokenizing
;temporary memory
shell_copy_addr	equ		$ca00		;731	area where GUI string is copied
stext_copy_addr	equ		$9000		;957	area where shell's text is copied
tokenize_copy_addr	equ	$b000		;2926	area where tokenizer code is copied
;file load (pre-decompress) memory
data_load_addr	equ		$8100
stext_load_addr	equ		$8100
tokenize_load_addr	equ	$8100
;system data
vars			equ		$c32e				;start of storage area for sys vars
temp_buf		equ		vars		;7		temp string buffer
parse_ptr		equ		vars+7		;2		pointer to command currently being parsed
write_ptr		equ		vars+9		;2		pointer to current location in token buffer
sum				equ		vars+11		;2		temp accumulator used while parsing numbers
curr_var		equ		vars+13		;2		index of current var
curr_var_type	equ		vars+15		;1		type token of current var
curr_var_index	equ		vars+16		;2		index of current var, if array
expr_sum		equ		vars+18		;2		temp accumulator used for sum of expressions
sp_save			equ		vars+20		;2		sp saved here for error
exec_sp			equ		vars+22		;2		sp saved here for if command in exec section needs to be aborted
input_str_sp	equ		vars+24		;2		sp saved here for input string routine
func_sp			equ		vars+26		;2		stack pointer for function stack
start_addr		equ		vars+28		;2		pointer to start of function MAIN
bmp_mem_ptr		equ		vars+30		;2		pointer to start of free memory for bitmaps
free1_mem_ptr	equ		vars+32		;2		pointers to start of free memory in var/ptr/sprite storage areas
free2_mem_ptr	equ		vars+34		;2
free3_mem_ptr	equ		vars+36		;2
free4_mem_ptr	equ		vars+38		;2
spr_mem_ptr		equ		vars+40		;2


.org _asm_exec_ram

	nop
	jp prog_start

	.dw 0,shell_title

shell_title: .db "MLC interpreter",0


prog_start:
;first the string with the system text gets copied to RAM page 1
	ld hl,data_str_name-1
	ld b,1
	ld de,data_load_addr-$8000
	call load_file
	ret c
	ld hl,data_load_addr
	ld de,data_copy_addr
	call Decompress
prog_loop:
;then the GUI code and data is copied
	ld hl,shell_str_name-1
	ld b,0
	ld de,shell_copy_addr
	call load_file
	ret c
	ld hl,stext_str_name-1
	ld b,$01
	ld de,stext_load_addr-$8000
	call load_file
	ret c
	ld hl,stext_load_addr
	ld de,stext_copy_addr
	call Decompress
;get the name of the prog from the user
	call shell_copy_addr				;get the name of the test prog to be interpreted,
	ld (sp_save),sp						;save the stack pointer for any errors that might occur
	jp c,tok_prog
;find the address of its string, or create it if it doesn't exist,
	push hl
	call get_data_str_name
	ld de,curr_name						;save varname
	call _mov10b
	rst 10h								;does it exist?
	jr nc,skip_create_str				;if not, create it
	ld hl,0								;start out at 0 for size
	call _createstrng
skip_create_str:
	pop hl
;	ld hl,tokenize_str_name-1			;then copy the tokenizer code...
;	ld b,1
;	ld de,tokenize_load_addr-$8000
;	call load_file
;	ret c
;	ld hl,tokenize_load_addr
;	ld de,tokenize_copy_addr
;	call Decompress
;	pop hl
;;load prog to temp area,
;	ld b,$01
;	ld de,copy_addr-$8000
;	call load_file						;load the prog,
;	ld de,copy_addr						;and put 0 at end of file in buffer
;	add hl,de
;	ld (hl),0
;	dec hl								;make sure that there's a newline char at the end of the file
;	ld a,(hl)
;	cp $d6
;	call nz,insert_newline				;BAD BAD BAD!!!! the user DIDN'T put a newline at the end!!!! we'll just have to do it instead...
;	ld de,end_copy_addr-1				;(make sure the prog isn't too big to fit...)
;	call _cphlde
;	ret nc
	call tok_part_1

	ld hl,copy_addr
	ld a,(hl)
	dec a
	jp z,exec_tok_prog

;then tokenize it and resolve the labels
;	ld hl,free1_mem						;init all the storage areas,
;	ld (free1_mem_ptr),hl
;	ld hl,free2_mem
;	ld (free2_mem_ptr),hl
;	ld hl,free3_mem
;	ld (free3_mem_ptr),hl
;	ld hl,free4_mem
;	ld (free4_mem_ptr),hl
;	ld hl,spr_1_data
;	ld (spr_mem_ptr),hl
;	ld hl,bmp_mem
;	ld (bmp_mem_ptr),hl
;	ld hl,lbl_res_mem
;	ld (lbl_res_ptr),hl
;	call tokenize_copy_addr				;tokenize and resolve labels,
	call tok_part_2

;now, finally, execute it!!!
exec_prog:
	call clr_free_mem					;clear the var memory,
	ld hl,str_ti86						;init the model string,
	ld de,free1_mem
	ld bc,str_ti86_len
	ldir
	ld hl,func_stack					;init the function stack,
	ld (func_sp),hl
	call seed_random					;seed the random number generator,
	call clr_buf
	call clr_scrn						;clear the screen,
	call open_gray						;start the grayscale handler,
	call wait_for_no_key				;and wait until no keys are being pressed which could "overflow" into the prog's key handler
	call exec							;then execute!
end_prog:
	call wait_for_no_key
	call close_gray
	call clr_free_mem
	jp prog_loop						;when done with the program, go back to the shell


tok_part_1:
	push hl
	ld hl,tokenize_str_name-1			;then copy the tokenizer code...
	ld b,1
	ld de,tokenize_load_addr-$8000
	call load_file
	jr c,end_tok
	ld hl,tokenize_load_addr
	ld de,tokenize_copy_addr
	call Decompress
	pop hl
;load prog to temp area,
	ld b,$01
	ld de,copy_addr-$8000
	call load_file						;load the prog,
	ld de,copy_addr						;and put 0 at end of file in buffer
	add hl,de
	ld (hl),0
	dec hl								;make sure that there's a newline char at the end of the file
	ld a,(hl)
	cp $d6
	call nz,insert_newline				;BAD BAD BAD!!!! the user DIDN'T put a newline at the end!!!! we'll just have to do it instead...
	ld de,end_copy_addr-1				;(make sure the prog isn't too big to fit...)
	call _cphlde
	ret c
end_tok:
	pop hl
	ret
tok_part_2:
	ld hl,free1_mem						;init all the storage areas,
	ld (free1_mem_ptr),hl
	ld hl,free2_mem
	ld (free2_mem_ptr),hl
	ld hl,free3_mem
	ld (free3_mem_ptr),hl
	ld hl,free4_mem
	ld (free4_mem_ptr),hl
	ld hl,spr_1_data
	ld (spr_mem_ptr),hl
	ld hl,bmp_mem
	ld (bmp_mem_ptr),hl
	ld hl,lbl_res_mem
	ld (lbl_res_ptr),hl
	jp tokenize_copy_addr				;tokenize and resolve labels,


insert_newline:
	inc hl
	ld (hl),$d6
	inc hl
	ld (hl),0
	ret

get_data_str_name:
	rst 20h								;op1=varname
	ld hl,_op1+8
	ld de,_op1+9
	ld bc,7
	lddr
	ld a,Ltheta
	ld (de),a
	ld a,(hl)
	cp 8
	jp z,err_progname
	inc (hl)
	dec hl								;op1=[theta]varname
	ret

clr_free_mem:
	ld hl,free1_mem
	ld bc,end_free1_mem-free1_mem-1
	call clr_free_mem_hl
	ld hl,free2_mem
	ld bc,end_free2_mem-free2_mem-1
	call clr_free_mem_hl
	ld hl,free3_mem
	ld bc,end_free3_mem-free3_mem-1
	call clr_free_mem_hl
	ld hl,free4_mem
	ld bc,end_free4_mem-free4_mem-1
	call clr_free_mem_hl
	ld hl,vector_table_addr				;not needed for program, prevents crash on exit
	ld bc,257
	call clr_free_mem_hl
	ld hl,spr_mem+1
	ld bc,179
clr_free_mem_hl:
	ld d,h
	ld e,l
	inc de
	ld (hl),0
	ldir
	ret

;Other code sections

#include "exec.asm"						;token executor
#include "pretok.asm"					;pre-tokenized program tokenize/exec
#include "var.asm"						;variable management
#include "sprite.asm"					;sprite management
#include "ptr.asm"						;pointer management
#include "err.asm"						;errors
#include "data.asm"						;uhh...data!
#include "FVPUTS.inc"					;Josh Seagoe's text routine
#include "routines.asm"					;other useful routines

.end