;MLC (Multi-platform Language for Calculators) by Epic Programming Studios
;TI-86 interpreter
;Written by Donald Straney (burntfuse) - straney@umbi.umd.edu

;System data (compiled to separate string)

.plugin bin
#include "86.h"
#include "key_equ.asm"

.org $a9e0

str_err: .db "ERROR!",0			;error title
str_line: .db "line ",0

;error messages
str_frun: .db 1,"stack overflow",0
str_argrange: .db 1,"argument out of range",0
str_syntax: .db 1,"syntax error",0
str_eol: .db 1,"end of line expected",0
str_var: .db 1,"variable name expected",0
str_var_int: .db 2,"integer variable",0,"or array element expected",0
str_var_arry: .db 1,"array variable expected",0
str_var_str: .db 1,"string variable expected",0
str_vartype: .db 1,"wrong variable type",0
str_num: .db 1,"number expected",0
str_op: .db 1,"unrecognized operator",0
str_spr_prop: .db 1,"sprite property expected",0
str_bmp: .db 1,"bitmap name expected",0
str_spr: .db 1,"sprite name expected",0
str_str: .db 1,"string expected",0
str_nomain: .db 1,"program must have function 'MAIN'",0
str_invcmd: .db 1,"unrecognized command",0
str_invarg: .db 1,"unrecognized argument type",0
str_invop: .db 1,"unrecognized operator",0
;str_namelen: .db 2,"name is too long",0,"(must be 1-5 characters)",0
str_namelen: .db 1,"name is too long",0
;str_numlen: .db 2,"number is too long",0,"(must be from -32768 to +32768)",0
str_numlen: .db 1,"number is too long",0
str_mem_var: .db 1,"too many variables",0
str_mem_lbl: .db 1,"too many functions",0
str_nolbl: .db 1,"function doesn't exist",0
str_keycode: .db 1,"invalid keycode",0
str_num_key: .db 1,"keycode expected",0
str_mem: .db 1,"out of memory",0
str_mem_file: .db 2,"not enough memory",0,"to expand file",0
str_arraylen: .db 1,"arrays must have 1 to 255 elements",0
str_maplen: .db 2,"tilemaps must have heights",0,"and widths from 1 to 255",0
;str_arrayindex: .db 1,"array or map index out of bounds",0
str_arrayindex: .db 1,"index out of bounds",0
str_noarray: .db 1,"array or map hasn't been defined",0
;str_strlen: .db 2,"string is too long",0,"(must be <36 characters)",0
str_strlen: .db 1,"string is too long",0
;str_mem_bmp: .db 2,"not enough memory",0,"to create bitmap",0
str_mem_bmp: .db 1,"too many bitmaps",0
str_sprformat: .db 1,"incorrect bitmap format",0
str_nobmp: .db 1,"bitmap doesn't exist",0
str_mem_spr: .db 1,"too many sprites",0
str_nospr: .db 1,"sprite hasn't been defined",0
;str_progname: .db 2,"program name is too long",0,"(must be <8 characters)",0
str_progname: .db 1,"program name is too long",0
str_twovar_arry: .db 2,"array or map has already",0,"been created",0
str_twovar_spr: .db 1,"bitmap has already been defined",0
;str_deflen: .db 2,"definition must have",0,"same length as variable",0
str_deflen: .db 2,"array or map definition has",0,"wrong length",0
str_int: .db 1,"internal error",0
;str_cint: .db 2,"string cannot be converted to number:",0,"invalid character",0
str_cint: .db 1,"invalid character in string",0
;str_noptr: .db 2,"pointer has not been assigned",0,"a variable, sprite, or bitmap",0
str_noptr: .db 1,"pointer has not been defined",0
str_ptr: .db 1,"pointer name expected",0
str_map: .db 1,"tilemap name expected",0
str_map8: .db 1,"only 8x8 sprites allowed in maps",0


key_table:			;table w/direct input values vs. scancode
.db k_g7,k_left,24
.db k_g7,k_up,25
.db k_g7,k_right,26
.db k_g7,k_down,34
.db k_g1,k_2nd,21
.db k_g2,k_alpha,31
.db k_g1,k_more,23
.db k_g1,k_exit,22

.db k_g1,k_f1,11
.db k_g1,k_f2,12
.db k_g1,k_f3,13
.db k_g1,k_f4,14
.db k_g1,k_f5,15

.db k_g3,k_xvar,32
.db k_g4,k_del,33

.db k_g2,k_log,41
.db k_g3,k_sin,42
.db k_g4,k_cos,43
.db k_g5,k_tan,44
.db k_g6,k_exp,45

.db k_g2,k_ln,51
.db k_g3,k_ee,52
.db k_g4,k_lparen,53
.db k_g5,k_rparen,54
.db k_g6,k_div,55

.db k_g2,k_x2,61
.db k_g3,k_7,62
.db k_g4,k_8,63
.db k_g5,k_9,64
.db k_g6,k_mult,65

.db k_g2,k_comma,71
.db k_g3,k_4,72
.db k_g4,k_5,73
.db k_g5,k_6,74
.db k_g6,k_sub,75

.db k_g2,k_sto,81
.db k_g3,k_1,82
.db k_g4,k_2,83
.db k_g5,k_3,84
.db k_g6,k_add,85

.db k_g3,k_0,92
.db k_g4,k_period,93
.db k_g5,k_neg,94
.db k_g6,k_enter,95

.db 0,0,0

trig_table:
;1000*cos x, 1000*sin x
.dw 1000,0		;0
.dw 966,259		;15
.dw 866,500		;30
.dw 707,707		;45
.dw 500,866		;60
.dw 259,966		;75
.dw 0,1000		;90
.dw -259,966	;105
.dw -500,866	;120
.dw -707,707	;135
.dw -866,500	;150
.dw -966,259	;165
.dw -1000,0		;180
.dw -966,-259	;195
.dw -866,-500	;210
.dw -707,-707	;225
.dw -500,-866	;240
.dw -259,-966	;255
.dw 0,-1000		;270
.dw 259,-966	;285
.dw 500,-866	;300
.dw 707,-707	;315
.dw 866,-500	;330
.dw 966,-259	;345
.dw 1000,0		;360

input_str_keys_alpha:		;keycode-to-char conversion for INPT command (alpha mode)
.db 41,LcapA
.db 42,LcapB
.db 43,LcapC
.db 44,LcapD
.db 45,LcapE
.db 51,LcapF
.db 52,LcapG
.db 53,LcapH
.db 54,LcapI
.db 55,LcapJ
.db 61,LcapK
.db 62,LcapL
.db 63,LcapM
.db 64,LcapN
.db 65,LcapO
.db 71,LcapP
.db 72,LcapQ
.db 73,LcapR
.db 74,LcapS
.db 75,LcapT
.db 82,LcapU
.db 83,LcapV
.db 84,LcapW
.db 85,LcapX
.db 92,LcapY
.db 93,LcapZ
.db 94,Lspace
.db 0

input_str_keys_num:				;same (number mode)
.db 92,L0
.db 82,L1
.db 83,L2
.db 84,L3
.db 72,L4
.db 73,L5
.db 74,L6
.db 62,L7
.db 63,L8
.db 64,L9
.db 71,Lcomma
.db 55,Lslash
.db 65,Lasterisk
.db 75,Ldash
.db 85,LplusSign
.db 0

.end