;MLC (Multi-platform Language for Calculators) by Epic Programming Studios
;TI-86 interpreter
;Written by Donald Straney (burntfuse) - straney@umbi.umd.edu

;User interface

.plugin string86
#include "mlc.inc"
#include "mlcstext.inc"

.org shell_copy_addr

prog_rows	equ		7
prog_row	equ		9
prog_end_row	equ	prog_row+((prog_rows-1)*6)
prog_name_width	equ	34	;30
cur_width		equ	6
prog_draw_pos	equ	prog_row*256
cur_start_pos	equ	prog_row*256+prog_name_width
status_row		equ	56
status_pos		equ	status_row*256+2
status_pos_2	equ	status_row*256+72
credits_pos		equ	$0001


get_prog_name:
	call _ram_page_1					;make sure the right page is loaded,
	call _flushallmenus					;kill any open menus,
	call _runindicoff					;turn off the annoying run indicator,
	call _clrlcd						;clear screen,
	res alt_int,(iy+exceptionflg)		;turn off any user interrupts,
	ld hl,0
	ld (_pencol),hl						;and draw the title bar
	ld hl,str_title_bar
	set textInverse,(iy+textflags)
	set textEraseBelow,(iy+textflags)
	call _vputs
	ld hl,prog_draw_pos					;now set the cursor,
	ld (_pencol),hl
	res textInverse,(iy+textflags)
	res textEraseBelow,(iy+textflags)
	call refresh_proglist				;draw the list of MLC progs,
	call _ram_page_1
	xor a
	ld (_vnewline_patch),a
	ld hl,$fc00+(16*(status_row-1))		;and draw the status bar
	call draw_bar
;	ld hl,$fc00+(16*(status_row+6))
	ld hl,$fff0
	call draw_bar
	ld hl,$fc00+(16*status_row)
	call draw_bar_sides
	ld hl,status_pos_2
	ld (_pencol),hl
	ld hl,str_status
	call _vputs
	ld hl,prog_table					;now init the prog data pointer
	ld (prog_ptr),hl
	ld hl,cur_start_pos					;and init prog cursor position
	ld (cur_pos),hl
	ld (old_cur_pos),hl
	ld a,(prog_table)					;are there any MLC progs?
	or a
	jp z,get_prog_name_noprogs			;if not, show the user
	call draw_cursor					;otherwise, draw the cursor and status bar
	call refresh_status
get_prog_name_loop:
	call _getky							;wait for a keypress
	cp K_UP								;if it's the up key, go to the previous prog
	jr z,get_prog_name_up
	cp K_DOWN							;if it's the down key, go to the next prog
	jp z,get_prog_name_down
	cp K_ENTER							;if it's enter or 2nd, run the current prog
	jp z,get_prog_name_run
	cp K_SECOND
	jp z,get_prog_name_run
	cp K_ALPHA							;if it's alpha, turn off calc
	jr z,get_prog_name_off
	cp K_F1								;if it's F1, show the help
	jp z,get_prog_name_help
	cp K_F2								;if F2, show credits
	jp z,get_prog_name_credits
	cp K_DEL							;if del, delete the program
	jr z,get_prog_name_delete
	cp K_PLUS							;if one of the contrast-changing keys has been pressed,
	jr z,get_prog_name_contrup			;then change it
	cp K_MINUS
	jr z,get_prog_name_contrdn
	cp K_PRGM							;if it's PRGM, signal to the main program to tokenize the program
	jp z,get_prog_name_tok
	cp K_EXIT							;if it's exit, then exit the shell
	jr nz,get_prog_name_loop
	pop hl
	jp _clrlcd
get_prog_name_off:
	ld a,$01							;i couldn't find any author name on the routine i got this from
	out (3),a
	halt
	ld a,$0b
	out (3),a
	res onInterrupt,(iy+onflags)
	jr get_prog_name_loop
get_prog_name_contrup:
	ld hl,_contrast
	ld a,(hl)
	cp 32
	jr z,get_prog_name_loop
	inc a
	ld (hl),a
	out (2),a
	jr get_prog_name_loop
get_prog_name_contrdn:
	ld hl,_contrast
	ld a,(hl)
	or a
	jr z,get_prog_name_loop
	dec a
	ld (hl),a
	out (2),a
	jr get_prog_name_loop
get_prog_name_delete:
	ld hl,(prog_ptr)					;get the name of the program,
	dec hl
	rst 20h								;find it,
	rst 10h
	call _delvar						;and delete it
	ld hl,(prog_ptr)
	dec hl
	call get_data_str_name
	rst 10h
	call nc,_delvar						;along with its string
	call _ram_page_1
	jp get_prog_name
get_prog_name_up:
	ld hl,(prog_ptr)					;when moving up, first make sure that it
	ld de,prog_table					;isn't already at the beginning
	call _cphlde
	jp z,get_prog_name_loop
	ld bc,9								;if not, move the pointer to the prev. prog in the table
	or a
	sbc hl,bc
	ld (prog_ptr),hl
	ld hl,(cur_pos)						;check on the cursor position
	ld (old_cur_pos),hl
	ld a,h
	cp prog_row							;is it already at the top of this column?
	jr nz,get_prog_name_up_skip
	ld h,prog_end_row					;if so, move to bottom of previous column
	ld a,l
	sub prog_name_width
	ld l,a
	jr get_prog_name_end_move
get_prog_name_up_skip:
	ld a,h								;if it's in the middle of a column,
	sub 6								;just move it 1 line up
	ld h,a
	jr get_prog_name_end_move
get_prog_name_down:
	ld hl,(prog_ptr)					;when moving down, first make sure that it
	ld bc,9
	add hl,bc
	ld a,(hl)							;isn't already at the end
	or a
	jp z,get_prog_name_loop
	ld (prog_ptr),hl					;if not, move the pointer to the next prog in the table
	ld hl,(cur_pos)						;check on the cursor position
	ld (old_cur_pos),hl
	ld a,h
	cp prog_end_row						;is it already at the bottom of this column?
	jr nz,get_prog_name_down_skip
	ld h,prog_row						;if so, move to top of next column
	ld a,l
	add a,prog_name_width
	ld l,a
	jr get_prog_name_end_move
get_prog_name_down_skip:
	ld a,h								;if it's in the middle of a column,
	add a,6								;just move it 1 line down
	ld h,a
get_prog_name_end_move:
	ld (cur_pos),hl
	call draw_cursor					;then redraw the cursor
	call refresh_status					;and the status bar
	jp get_prog_name_loop				;and go back to the key loop
get_prog_name_run:
	ld hl,(prog_ptr)					;get the address of the current prog in the table
	dec hl
	xor a
	ret									;and pass that to the interpreter
get_prog_name_tok:
	ld hl,(prog_ptr)					;for tokenizing, do the same but set the carry flag
	dec hl								;to show that the prog needs to be tokenized
	push hl
	rst 20h
	rst 10h
	ld a,b
	ex de,hl
	call _ahl_plus_2
	call _load_ram_ahl
	ld a,(hl)
	pop hl
	call _ram_page_1
	or a
	jp nz,get_prog_name_loop
	scf
	ret
get_prog_name_credits:
	ld hl,credits_text_1
	ld b,num_credit_lines_1
	call disp_text
	ld hl,credits_text_2
	ld b,num_credit_lines_2
	call disp_text
	ld hl,credits_text_3
	ld b,num_credit_lines_3
	call disp_text
	res textInverse,(iy+textflags)
	jp get_prog_name
get_prog_name_help:
	ld hl,help_text
	ld b,num_help_lines
	call disp_text
	res textInverse,(iy+textflags)
	jp get_prog_name
get_prog_name_noprogs:
	ld hl,prog_draw_pos					;draw the message
	ld (_pencol),hl
	ld hl,str_noprogs
	call _vputs
get_prog_name_noprogs_loop:				;then wait for exit to be pressed
	call _getky
	cp K_EXIT
	jr nz,get_prog_name_noprogs_loop
	pop hl
	jp _clrlcd


refresh_status:
	ld hl,status_pos					;position the cursor
	ld (_pencol),hl
	ld hl,str_size						;draw first the "size:" message
	call _vputs
	ld hl,(prog_ptr)					;then get the size of the current program
	dec hl
	rst 20h
	rst 10h
	call _data_size_to_de
	ex de,hl
	call DisplayHL						;and draw that too
	call _ram_page_1
	ld hl,str_status_erase
	jp _vputs


draw_bar:
	ld b,16
draw_bar_loop:
	ld (hl),$ff
	inc hl
	djnz draw_bar_loop
	ret

draw_bar_sides:
	ld b,7
	ld de,15
draw_bar_sides_loop:
	ld (hl),$80
	add hl,de
	ld (hl),$01
	inc hl
	djnz draw_bar_sides_loop
	ret


disp_text:								;text at hl, number of lines in b
	push hl
	push bc
	ld hl,$fc00							;first make the screen all black
	ld de,$fc01
	ld (hl),$ff
	ld bc,1023
	ldir
	ld de,credits_pos					;then reset the cursor position
	ld (_pencol),de
	ld a,e
	ld (_vnewline_patch),a
	set textInverse,(iy+textflags)
	pop bc								;and display the text
	pop hl
disp_credits_loop_1:
	call _vputs
	call _vnewline
	djnz disp_credits_loop_1
disp_credits_loop_2:					;wait for 2nd or enter to be pressed
	call _getky
	cp K_SECOND
	ret z
	cp K_ENTER
	ret z
	jr disp_credits_loop_2


draw_cursor:
	ld hl,(old_cur_pos)					;first erase old cursor
	ld (_pencol),hl
	ld hl,str_erase
	call _vputs
	ld hl,(cur_pos)						;then draw the new one
	ld (_pencol),hl
	ld hl,str_cur
	jp _vputs


refresh_proglist:
	call _ram_page_7				;load ram page with vat,
	ld hl,prog_table				;init prog table pointer,
	ld (prog_ptr),hl
	ld hl,$bfff						;and start at beginning of vat
refresh_proglist_again:
	ld b,prog_rows					;set number of vars to draw,
refresh_proglist_loop:
	ld de,($d298)					;get addr of end of VAT
	call _cphlde					;are we there yet?
	jp z,refresh_proglist_done		;if so, stop
	push bc
	ld a,(hl)						;get var type
	cp $12							;is the var a program?
	jr nz,refresh_proglist_skip		;if not, don't display it
	dec hl
	ld e,(hl)						;and the abs address in bde
	dec hl
	ld d,(hl)
	dec hl
	ld a,(hl)
	dec hl
	dec hl
	push hl
	ld hl,3
	add hl,de
	adc a,0							;(ahl=start of data)
	call _set_abs_src_addr
	ld hl,header_test				;copy the first 5 bytes of data
	xor a
	call _set_abs_dest_addr
	ld hl,5
	call _mm_ldir_set_size
	xor a
	ld (header_test+5),a
	ld hl,header_test
	ld de,header_str
	call str_cmp					;make sure they're "MLC:[\n]"
	pop hl
	call _ram_page_7
	jr nz,refresh_proglist_skip_2
	ld a,(hl)						;if so, get the name length
	dec hl
	ld de,(prog_ptr)
	ld (de),a
	inc de
	ld b,a
refresh_proglist_drawname:			;draw var name
	ld a,(hl)
	ld (de),a						;and put it in prog table
	push de
	call _vputmap
	pop de
	dec hl
	inc de
	djnz refresh_proglist_drawname
	jr refresh_proglist_end_loop
refresh_proglist_skip:
	dec hl
	dec hl
	dec hl
	dec hl
	dec hl
refresh_proglist_skip_2:
	ld c,(hl)
	dec hl
	ld b,0
	or a
	sbc hl,bc
	pop bc
	jr refresh_proglist_loop
refresh_proglist_end_loop:
	ld de,(prog_ptr)				;update the prog table pointer
	ld bc,9
	ex de,hl
	add hl,bc
	ld (prog_ptr),hl
	ex de,hl
	call _vnewline					;move the cursor to the next line
	pop bc
	djnz refresh_proglist_loop		;and loop
	ld de,_pencol					;when we're done with that column, move to the next
	ld a,(de)
	add a,prog_name_width+cur_width
	ld (_vnewline_patch),a
	ld (de),a
	inc de
	ld a,prog_row
	ld (de),a
	jp refresh_proglist_again
refresh_proglist_done:
	ld hl,(prog_ptr)				;when done, zero-terminate prog table
	ld (hl),0
	ret


header_test:

shell_data	equ		header_test+6
prog_ptr	equ		shell_data
old_cur_pos	equ		shell_data+2
cur_pos		equ		shell_data+4
prog_table	equ		shell_data+6


.end