;MLC (Multi-platform Language for Calculators) by Epic Programming Studios
;TI-86 interpreter
;Written by Donald Straney (burntfuse) - straney@umbi.umd.edu

;Tokenizer code (compiled to separate string)

.plugin bin
#include "mlc.inc"

;#define FULL_ERR							;redundant error checking (55 bytes)
#define FAST_KEY							;pre-processed keypress detection (56 bytes)

.org tokenize_copy_addr

;//////////////////////////////Main loop//////////////////////////////

tokenize:								;tokenize uncompiled BASIC prog at copy_addr, write to write_addr
	xor a								;first init the label and var tables
	ld (num_lbls),a
	ld (num_vars),a
	ld (num_sprites),a
	ld hl,str_mdl						;then create the string var '$MDL' with the calc's model
	ld b,str_mdl_len
	call add_var
	ld hl,copy_addr+5					;start pointer at beginning of prog
	ld (parse_ptr),hl
	ld hl,write_addr
	ld (write_ptr),hl
tokenize_loop:
	ld hl,(parse_ptr)					;first check if we're at the end of the file
	ld a,(hl)
	or a
	jr z,tokenize_done
	cp $d6								;is this line blank?
	jr z,tokenize_newline
	ld de,tokenize_loop					;(allow alt expression handlers to use ret)
	push de
	cp command_char						;does it have a command?
	jp nz,tokenize_expression			;if not, it has an expression - we have to evaluate it
	inc hl								;otherwise, point to beginning of command,
	ld (parse_ptr),hl
	ld a,cmd_len						;init command length constant,
	ld (temp_buf),a
	call get_cmd_token					;then get the token for the current command,
	cp data_token						;(if it's the start of a data area, handle that separately)
	jp z,tokenize_data
	ld hl,(write_ptr)					;write it to token buffer,
	ld (hl),a
	inc hl
	ld (write_ptr),hl					;and update token buffer pointer
	cp frun_token						;(if it's a function token, handle this differently)
	jp z,tokenize_frun
	cp fnct_token
	jp z,tokenize_fnct
	cp if_token							;(or if it's an _IF_ token)
	jp z,tokenize_if
	pop de
#ifdef FAST_KEY
	cp gkey_token
	call z,tokenize_gkey
#endif
	ld hl,(parse_ptr)
	ld a,(hl)
tokenize_arg_loop:						;now tokenize arguments of command
	inc hl								;is it an argument or the end of the line?
	cp $d6								;(if it's the end of the line, end the loop)
	jr z,tokenize_arg_loop_end
	cp arg_separator					;(if it's an argument, skip the separator and tokenize the argument)
	jp nz,err_syntax
tokenize_arg_loop_alt:
	call tokenize_next_arg				;get token for argument
	ld hl,(parse_ptr)
	ld a,(hl)
	call check_valid_op
	jr nz,tokenize_arg_loop				;if there isn't an operator after it, loop
	inc hl
	ld de,(write_ptr)
	ld (de),a
	inc de
	ld (write_ptr),de
	jr tokenize_arg_loop_alt
tokenize_arg_loop_end:
	ld (parse_ptr),hl					;update the parse pointer
	ld hl,(write_ptr)					;write end-of-line token
	ld (hl),eol_token
	inc hl
	ld (write_ptr),hl
	jr tokenize_loop					;and tokenize next line
tokenize_newline:
	inc hl								;if the user has chosen to leave a line blank, why should we care?
	ld (parse_ptr),hl
	jp tokenize_loop
tokenize_done:
	ld hl,(write_ptr)					;when done tokenizing,
	ld (hl),eof_token					;put eof token at end of token buffer
	call res_labels						;then resolve the labels
	call res_bmps						;and bitmap names
	ld hl,str_main						;save the address of 'fnct main'
	ld b,str_main_len
	call get_label
	jp c,err_nomain
	ld (start_addr),hl
	ret									;and return

;//////////////////////////////Label/bitmap resolve loop//////////////////////////////

res_labels:
	ld hl,write_addr					;init the pointer
	ld (parse_ptr),hl
res_labels_loop:
	ld a,(hl)							;check on the next char
	cp eof_token						;are we at the end of the file?
	ret z								;if so, we're done!!!
	cp frun_token						;does it have an FRUN token?
	jp nz,res_labels_find_end			;if not, we can skip it
	inc hl								;otherwise, get the address of label name,
	push hl
	call _ldhlind
	call parse_label					;and resolve the label,
	call get_label
	pop de								;then write the address after the FRUN token
	ld (parse_ptr),de
	jp c,err_nolbl						;(if it hasn't been defined, show error)
	ex de,hl
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	inc hl
	jr res_labels_loop
res_labels_find_end:
	call goto_eol
	inc hl
	ld (parse_ptr),hl
	jr res_labels_loop


res_bmps:
	ld hl,write_addr					;init pointer to current position in token buffer
	ld (parse_ptr),hl
res_bmps_loop:
	ld a,(hl)							;get the next token
	cp eof_token						;if we're at the end of the file, return
	ret z
	cp frun_token
	jr z,res_bmps_loop_3
	cp int_name_token					;if it's anything but a bitmap name token, skip it
	jr z,res_bmps_loop_3
	cp array_name_token
	jr z,res_bmps_loop_3
	cp str_name_token
	jr z,res_bmps_loop_3
	cp spr_name_token
	jr z,res_bmps_loop_3
	cp ptr_name_token
	jr z,res_bmps_loop_3
	cp lit_str_token
	jr z,res_bmps_loop_str
	cp lit_num_token
	jr z,res_bmps_loop_3
	cp keycode_token
	jr z,res_bmps_loop_3
	cp map_name_token
	jr z,res_bmps_loop_3
	cp bmp_name_token
	jr nz,res_bmps_loop_1
res_bmps_loop_bmp:						;if it is a bitmap name token,
	inc hl
	ld e,(hl)							;get the address of the bitmap table entry,
	inc hl
	ld d,(hl)
	dec hl
	push hl
	call get_bmp_addr					;translate that to the address of the bitmap data,
	ld a,d								;(first making sure that it exists)
	or e
	jp z,err_nobmp
	pop hl
	ld (hl),e							;then replace the address with the new address
	inc hl
	ld (hl),d
	inc hl
	jr res_bmps_loop_end
res_bmps_loop_3:
	inc hl
	inc hl
res_bmps_loop_1:
	inc hl
res_bmps_loop_end:
	ld (parse_ptr),hl
	jr res_bmps_loop
res_bmps_loop_str:
	ld bc,str_max_len+1
	xor a
	cpir
	jr res_bmps_loop_end

;//////////////////////////////Keycode tokenizer//////////////////////////////

#ifdef FAST_KEY
tokenize_gkey:
	ld hl,(parse_ptr)
	inc hl								;skip the space after the command
	ld a,(hl)
	cp num_char							;and make sure that there's a number for the keycode
	jp nz,err_num_key
	call parse_num
	ld a,h								;now check if it's out of range
	or a
	jp nz,err_keycode
	ld b,l
	ld hl,key_table+2					;and search for it in the key table
	ld de,3
tokenize_gkey_loop:
	ld a,(hl)
	or a
	jp z,err_keycode
	cp b
	jr z,tokenize_gkey_end_loop
	add hl,de
	jr tokenize_gkey_loop
tokenize_gkey_end_loop:
	dec hl								;to get the direct input values
	ld a,(hl)
	cpl
	ld e,a
	dec hl
	ld d,(hl)
	ld b,keycode_token
	jp write_bde						;and write those to the token buffer
#endif

;//////////////////////////////Sprite tokenizer//////////////////////////////

tokenize_data:
	ld hl,(parse_ptr)					;first make sure that's the end of the line
	ld a,(hl)
	cp $d6
	jp nz,err_eol
	inc hl
	ld (parse_ptr),hl
tokenize_data_loop:						;then find and create the sprites in this data area
	ld hl,(parse_ptr)					;look at the first char in the current line
	ld a,(hl)
	cp $d6
	jp z,tokenize_data_newline			;is it a blank line?
	cp command_char
	jp nz,err_syntax
	inc hl
	ld (parse_ptr),hl
	ld a,cmd_len
	ld (temp_buf),a
	call get_cmd_token					;if none of those, get the token for it
	cp dend_token						;is it the end of the data area?
	ret z								;if so, we're done parsing the data, so go back to main tokenizer
	cp bitm_token						;is it a bitmap definition?
	jp nz,err_invcmd					;if not, show an error - there shouldn't be any commands here!
tokenize_data_sprite:					;otherwise, add the bitmap to the table
	ld hl,(parse_ptr)
	inc hl								;if it's a bitmap,
	ld a,(hl)
	cp bitmap_char
	jp nz,err_bmp
	call parse_varname_nocheck			;parse the name
	push bc								;(and save the parameters)
	push hl
	ld hl,(parse_ptr)
	ld a,(hl)
	cp Lspace
	jp nz,err_bmpformat
	call parse_num
	ld de,8								;make sure that the size is supported...
	call _cphlde
	jr z,tokenize_data_sprite_8
	ld de,16
	call _cphlde
	jp nz,err_bmpformat
tokenize_data_sprite_16:
	ld a,32
	jr $+4
tokenize_data_sprite_8:
	ld a,8
	ld (tokenize_data_size_1),a			;(patch the code for the different sizes)
	ld (tokenize_data_size_2),a
	ld hl,(parse_ptr)
	ld a,(hl)
	cp $d6
	jp nz,err_bmpformat
	inc hl
	ld de,spr_temp_mem					;now we've got to parse the data
	ld b,3
tokenize_data_sprite_loop_outer:
	push bc
tokenize_data_size_1	equ	$+1
	ld b,0								;(b=number of bytes in each layer of sprite)
tokenize_data_sprite_loop_inner:
	call conv_hex
	jp c,err_bmpformat
	ld (de),a
	inc de
	djnz tokenize_data_sprite_loop_inner
	pop bc
	ld a,(hl)
	cp $d6
	jp nz,err_bmpformat
	inc hl
	djnz tokenize_data_sprite_loop_outer
	ld (parse_ptr),hl					;when done with this sprite, update the parse pointer
	pop hl								;and add the bitmap to the table
	pop bc
	call check_var_index				;does it exist already?
	jr c,tokenize_data_sprite_create	;if it does, just assign the bitmap to the existing entry in the table
	push de
	call get_bmp_addr					;(first make sure it hasn't already been defined)
	ld a,d
	or e
	jp nz,err_twovar_bmp
	pop de
	jr $+5
tokenize_data_sprite_create:
	call create_var						;otherwise, create a new bitmap
	ld hl,spr_temp_mem					;now allocate the memory for the bitmap! (set the data source)
tokenize_data_size_2	equ	$+1
	ld a,0								;(and the size)
	call alloc_bmp
	jp tokenize_data_loop				;then go back to search for more sprites
tokenize_data_newline:					;extra blank lines should be allowed...
	inc hl
	ld (parse_ptr),hl
	jp tokenize_data_loop

;//////////////////////////////Label tokenizer//////////////////////////////

tokenize_fnct:
	ld hl,(parse_ptr)
	inc hl
	call parse_label					;get the label name
	ld de,(write_ptr)
	push de
	inc de
	call add_label						;define the label,
	pop hl
	ld (hl),eol_token					;put an end-of-line token,
	inc hl
	ld (write_ptr),hl
	ret									;and go back to parse the next line

tokenize_frun:
	ld hl,(lbl_res_ptr)
	ld de,(write_ptr)					;write the address of the label string to the token buffer
	ld a,l
	ld (de),a
	inc de
	ld a,h
	ld (de),a
	inc de
	ld (write_ptr),de
	ld de,(parse_ptr)					;now copy the label name to the buffer
	inc de
	ex de,hl
	ld bc,name_max_len+1
	ldir
	ld (lbl_res_ptr),de					;and update the label buffer pointer
	ld hl,(parse_ptr)
tokenize_frun_loop:						;then find the end of the line
	ld a,(hl)
	cp $d6
	inc hl
	jr nz,tokenize_frun_loop
	ld (parse_ptr),hl					;and update the parse pointer
	ld hl,(write_ptr)
	ld (hl),eol_token
	inc hl
	ld (write_ptr),hl
	ret

;//////////////////////////////Expression tokenizer//////////////////////////////

tokenize_expression:
	ld hl,(parse_ptr)					;first, there has to be a var name
	ld a,(hl)							;get the vartype
	cp int_char							;and look at it
	jp z,tokenize_expression_int
	cp array_char
	jp z,tokenize_expression_array
	cp string_char
	jp z,tokenize_expression_str
	cp pointer_char
	jp z,tokenize_expression_ptr
	cp sprite_char
	jr z,tokenize_expression_spr
	cp map_char
	jp nz,err_invarg
tokenize_expression_map:
	call parse_varname					;get the map's index
	call c,create_var
	ld b,map_name_token
tokenize_expression_map_:
	call write_bde						;and write its index to the token buffer
	ld hl,(parse_ptr)
	ld a,(hl)							;test for the parentheses
	cp index_start_char
	jr z,tokenize_expression_map_elem	;if parentheses, it's just be a definition of a single element
tokenize_expression_map_whole:
	inc hl
	ld (parse_ptr),hl
	cp LEQ								;otherwise, it must be a definition of the whole map
	jp nz,err_syntax
tokenize_expression_map_loop:
	call tokenize_bmp					;tokenize the current element
	ld hl,(parse_ptr)
	ld a,(hl)
	cp $d6
	jp z,tokenize_eol					;if it hit the end of the line, then we're done with this tilemap
	cp arg_separator					;otherwise, make sure that there's a space in between elements,
	jp nz,err_syntax
	inc hl
	ld (parse_ptr),hl					;then point the parse pointer to the beginning of the next element,
	jr tokenize_expression_map_loop		;and loop
tokenize_expression_map_elem:
	call tokenize_map_index				;get the index and write that to the token buffer
	ld hl,(parse_ptr)
	ld a,(hl)
	cp LEQ
	jp nz,err_syntax
	inc hl
	call tokenize_bmp_hl				;now get the value that will be stored to that element and tokenize it
	jp tokenize_eol


tokenize_expression_spr:
	call parse_varname					;get the name of the sprite and its index
	call c,create_var					;if it doesn't exist, create it
	ld b,spr_name_token
tokenize_expression_spr_:
	call write_bde						;now write it to the token buffer
	ld hl,(parse_ptr)					;make sure there's a '.' after it
	ld a,(hl)
	cp Lperiod
	jp nz,err_syntax
	inc hl
	call get_spr_prop					;now check on the property of the sprite that the user wants to change
	ld de,(write_ptr)
	ld (de),a
	inc de
	ld (write_ptr),de
	cp LcapB
	jr z,tokenize_expression_spr_setbmp
	ld a,(hl)
	cp LEQ								;(first make sure that there's an '=' after it)
	jp nz,err_syntax
	inc hl
	call tokenize_expr_hl				;then tokenize the expression with the property
	jp nc,err_syntax
	ret
tokenize_expression_spr_setbmp:
	ld a,(hl)
	cp LEQ								;if so, expect an '=' after it
	jp nz,err_syntax
	inc hl
	call tokenize_bmp_hl
	jp tokenize_eol


tokenize_expression_ptr:
	call parse_varname					;yeah, yeah, parse the pointer name, if it doesn't exist, create it and all that crap...
	call c,create_var
	ld (curr_var),de
	ld hl,(parse_ptr)
	ld a,(hl)							;ah, the big question...is the user trying to assign a var to the pointer or trying to
	cp Lcolon							;perform some operation on the pointed-to var?
	jp nz,tokenize_expression_ptr_val
	inc hl
	ld a,(hl)
	push af
	call parse_varname
	call c,create_var
	pop af								;if it's a var or sprite name, determine the type
	call char_to_tok					;and return its associated type token
	ld hl,(write_ptr)
	ld (hl),ptr_name_token				;write the pointer's address to the token buffer
	inc hl
	ld a,(curr_var)
	ld (hl),a
	inc hl
	ld a,(curr_var+1)
	ld (hl),a
	inc hl
	ld (hl),Lcolon						;then a ':' to show that it's being assigned a var
	inc hl
	call write_bde_hl					;then the type token for the var and the address of the var itself
	jp tokenize_eol						;make sure that's the end of the line, and that's it
tokenize_expression_ptr_val:
	ld b,ptr_name_token
	ld a,(hl)							;jump to correct part of tokenize_expression depending on what follows pointer name
	cp index_start_char
	jr z,tokenize_expression_ptr_other
	cp Lperiod
	jp z,tokenize_expression_spr_
	cp LEQ
	jp nz,err_syntax
	inc hl
	ld a,(hl)
	cp num_char
	jp z,tokenize_expression_int_
	cp int_char
	jp z,tokenize_expression_int_
	cp Lquote
	jr z,tokenize_expression_str_
	cp array_char
	jp z,tokenize_expression_int_
	cp string_char
	jr z,tokenize_expression_str_
	cp bitmap_char
	jp z,tokenize_expression_map_
	jp err_syntax
tokenize_expression_ptr_other:					;array or tilemap
	push hl
	ld a,(hl)
	inc hl
	cp LEQ
	jr nz,tokenize_expression_ptr_other+1
	ld a,(hl)
	pop hl
	cp int_char
	jp z,tokenize_expression_array_
	cp array_char
	jp z,tokenize_expression_array_
	cp num_char
	jp z,tokenize_expression_array_
	cp pointer_char
	jp z,tokenize_expression_array_
	cp sprite_char
	jp z,tokenize_expression_array_
	cp bitmap_char
	jp z,tokenize_expression_map_
	jp err_syntax


tokenize_expression_str:
	call parse_varname					;parse the varname
	call c,create_var					;if it doesn't exist, create it
	ld b,str_name_token
tokenize_expression_str_:
	call write_bde						;now write the var index to the token buffer
	ld hl,(parse_ptr)					;make sure there's an '=' after it
	ld a,(hl)
	cp LEQ
	jp nz,err_syntax
	inc hl								;is the expression numerical or a string?
	ld a,(hl)
	cp Lquote
	jr z,tokenize_expression_str_str
	cp string_char
	jr z,tokenize_expression_str_str
	cp pointer_char
	jr z,tokenize_expression_str_str
tokenize_expression_str_num:			;if it's a numerical expression,
	ld de,(write_ptr)					;write an int token,
	ld a,int_char
	ld (de),a
	inc de
	ld (write_ptr),de
	call tokenize_expr_hl				;tokenize it normally
	jp nc,err_syntax
	ret
tokenize_expression_str_str:			;otherwise, if it's a string...
	ld de,(write_ptr)					;write a string token
	ld a,string_char
	ld (de),a
	inc de
	ld (write_ptr),de
tokenize_expression_str_loop:
	call parse_str						;tokenize the string var or literal
	ld hl,(parse_ptr)					;is it at the end of the line yet?
	ld a,(hl)
	cp $d6
	jp z,tokenize_eol
	inc hl
	ld de,(write_ptr)					;if not, then we can assume that there's an operator
	ld (de),a
	cp LplusSign						;(only '+' is allowed!)
	jp nz,err_invop
	inc de
	ld (write_ptr),de
	jr tokenize_expression_str_loop		;then parse the next string in the expression


tokenize_expression_int:				;if it's an integer var,
	call parse_varname					;then parse the varname
	call c,create_var					;check if the var exists already - if not, we have to create it
	ld b,int_name_token
tokenize_expression_int_:
	call write_bde						;write the var name token and the var index to the token buffer
	ld hl,(parse_ptr)
	ld a,(hl)
	cp LEQ								;check if there's an equals after the varname
	jr nz,tokenize_expression_int_op	;there has to be!!! if not, show an error
	inc hl
	call tokenize_expr_hl				;parse the rest of the expression
	jp nc,err_syntax
	ret									;and return to parse the next line
tokenize_expression_int_op:
	cp LplusSign						;it's a "[var]+" or a "[var]-" line
	jr z,tokenize_expression_int_op_cont
	cp Ldash
	jr z,tokenize_expression_int_op_cont
	cp Lasterisk
	jr z,tokenize_expression_int_selfop
	cp Lslash
	jp nz,err_invop						;make sure it's a valid operator...
tokenize_expression_int_selfop:
	inc hl								;if it's an '*=','/=','+=', or '-=',
	ld c,a
	ld a,(hl)
tokenize_expression_int_selfop_:
	cp LEQ
	jp nz,err_syntax
	inc hl
	ld (parse_ptr),hl
	call write_bdec						;write the var address and all that crap, then the operator
	call tokenize_expr					;and parse the rest of the expression
	jp nc,err_syntax
	ret
tokenize_expression_int_op_cont:
	inc hl
	ld c,a
	ld a,(hl)
	cp $d6								;and that the line ends after it...
	jr nz,tokenize_expression_int_selfop_
	ld (parse_ptr),hl
	ld hl,(write_ptr)					;then write the operator to the token buffer
	ld (hl),c
	inc hl
	ld (write_ptr),hl
	jp tokenize_eol


tokenize_expression_array:				;if it's an array,
	call parse_varname					;as before, we first parse the varname
	call c,create_var
	ld b,array_name_token
tokenize_expression_array_:
	call write_bde						;and write its index to the token buffer
	ld hl,(parse_ptr)
	ld a,(hl)							;test for the parentheses
	cp index_start_char
	jr z,tokenize_expression_array_elem	;if parentheses, it's just be a definition of a single element
tokenize_expression_array_whole:
	cp LEQ								;otherwise, it must be a definition of the whole array!
	jp nz,err_syntax
	inc hl
tokenize_expression_array_loop:
	call tokenize_expr_hl				;tokenize the first element of the array
	ret c								;if it hit the end of the line, then we're done with this array
	ld a,(hl)							;otherwise, make sure that there's a space in between elements,
	cp arg_separator
	jp nz,err_syntax
	inc hl
	ld (parse_ptr),hl					;then point the parse pointer to the beginning of the next element,
	jr tokenize_expression_array_loop	;and loop
tokenize_expression_array_elem:
	call tokenize_array_index			;tokenize the index
	ld hl,(parse_ptr)
	ld a,(hl)
	cp LEQ
	jp nz,err_syntax
	inc hl
	call tokenize_expr_hl				;now get the value that will be stored to that array element and tokenize it
	jp nc,err_syntax
	ret									;return to the main tokenizer to do the next line


tokenize_bmp:							;tokenize bitmap expression at (parse_ptr)
	ld hl,(parse_ptr)
tokenize_bmp_hl:
	ld a,(hl)
	cp bitmap_char						;look at the argument type (a bitmap, the bitmap of another sprite, etc.)
	jr z,tokenize_bmp_bmp
	cp sprite_char
	jr z,tokenize_bmp_spr
	cp map_char
	jr z,tokenize_bmp_map
	cp pointer_char
	jp nz,err_bmp
tokenize_bmp_ptr:
	call parse_varname					;if it's a pointer name,
	call c,create_var
	ld b,ptr_name_token					;create the pointer
	call write_bde
	ld hl,(parse_ptr)					;then look after the name to see if it's a sprite or bitmap
	ld a,(hl)
	cp Lperiod							;if there's no period after the name, it must be a bitmap
	ret nz
	jr tokenize_bmp_spr_				;otherwise, it's the bitmap of another sprite, so use the existing code for that
tokenize_bmp_spr:
	call parse_varname					;if it's the bitmap of another sprite, get its address
	call c,create_var
	ld b,spr_name_token
	call write_bde						;and write that to the token buffer, along with an end-of-line token
	ld hl,(parse_ptr)
	ld a,(hl)
	cp Lperiod							;wait! but first, make sure it's the other sprite's bitmap that's being transferred...
	jp nz,err_bmp
tokenize_bmp_spr_:
	inc hl
	ld a,(hl)
	cp LcapB
	jp nz,err_bmp
	inc hl
	ld a,(hl)
	cp LcapM
	jp nz,err_bmp
	inc hl
	ld a,(hl)
	cp LcapP
	jp nz,err_bmp
	inc hl
	ld (parse_ptr),hl
	ld hl,(write_ptr)					;then write the '.BMP' to the token buffer
	ld (hl),LcapB
	inc hl
	ld (write_ptr),hl
	ret									;and it's done!!!!!
tokenize_bmp_bmp:						;if it's a bitmap name,
	call parse_varname					;just get the address
	call c,create_var
	ld b,bmp_name_token
	jp write_bde						;and write that with the type token
tokenize_bmp_map:
	call parse_varname					;if it's a tilemap, it must be just a single element
	call c,create_var
	ld b,map_name_token					;so write the basic stuff to the token buffer
	call write_bde
	jp tokenize_map_index				;and the index


tokenize_if:							;basically just a modified version of tokenize_expression
	ld hl,(parse_ptr)					;skip the space after the #_IF_
	inc hl
	ld (parse_ptr),hl
	ld a,(hl)							;WAIT...what kind of expression is this???
	cp string_char						;if it's comparing two strings, we need to do this differently!
	jr z,tokenize_if_str
	cp Lquote
	jr z,tokenize_if_str
	cp int_char
	jr z,tokenize_if_num
	cp num_char
	jr z,tokenize_if_num
	cp array_char
	jr z,tokenize_if_num
	cp bitmap_char
	jr z,tokenize_if_bmp
	cp map_char
	jr z,tokenize_if_bmp
	cp pointer_char
	jr z,tokenize_if_num
	cp sprite_char
	jp nz,err_invarg
	push hl
	call parse_varname
	ld hl,(parse_ptr)
	ld a,(hl)
	cp Lperiod
	jp nz,err_invarg
	inc hl
	ld a,(hl)
	cp LcapB
	jr z,tokenize_if_sprite_bmp
tokenize_if_sprite_num:
	pop hl
	ld (parse_ptr),hl
	jr tokenize_if_num
tokenize_if_sprite_bmp:
	pop hl
	ld (parse_ptr),hl
tokenize_if_bmp:
	ld hl,(write_ptr)
	ld (hl),bitmap_char
	inc hl
	ld (write_ptr),hl
	call tokenize_bmp					;if it's a bitmap comparison, tokenize the first bitmap expression,
	ld hl,(parse_ptr)
	ld a,(hl)
	cp LEQ								;expect an '=' or '!' between the bitmaps,
	jr z,$+7
	cp Lexclam
	jp nz,err_syntax
	ld de,(write_ptr)
	ld (de),a
	inc de
	ld (write_ptr),de
	inc hl
	call tokenize_bmp_hl				;then do the second one
	jp tokenize_eol						;and make sure that's the end of the line
tokenize_if_num:						;otherwise, it's just a normal numerical comparison,
	ld hl,(write_ptr)
	ld (hl),int_char
	inc hl
	ld (write_ptr),hl
	call tokenize_expr					;so tokenize the expression
	jp nc,err_eol
	ret									;and go back to the main tokenizer
tokenize_if_str:
	ex de,hl
	ld hl,(write_ptr)
	ld (hl),string_char
	inc hl
	ld (write_ptr),hl
	ex de,hl
	call parse_str						;if it's a string expression, parse the first string
	ld hl,(parse_ptr)
	ld a,(hl)
	cp LEQ								;then copy the operator
	jr z,$+7
	cp Lexclam
	jp nz,err_invop
	inc hl
	ld de,(write_ptr)
	ld (de),a
	inc de
	ld (write_ptr),de
	call parse_str						;then the second string
	jp tokenize_eol						;make sure that this ends the expression


tokenize_expr:							;tokenize expression at (parse_ptr), parse_ptr updated, data written to (write_ptr), write_ptr updated, c set if new line
	ld hl,(parse_ptr)
tokenize_expr_hl:
	push hl
	ld de,(write_ptr)					;(make sure the buffer doesn't overflow)
	ld hl,temp_expr_buf+30
	call _cphlde
	jp c,err_mem
	pop hl
	ld a,(hl)							;get the next char
	cp num_char							;if it's a number,
	jp z,tokenize_expr_num				;then tokenize that
	cp int_char							;if it's a var,
	jp z,tokenize_expr_int				;then tokenize that instead
	cp array_char						;handle array vars too
	jp z,tokenize_expr_array
	cp sprite_char						;sprite positions are also allowed
	jp z,tokenize_expr_sprite
	cp pointer_char						;and pointers
	jp nz,err_num						;otherwise, it's unrecognized
tokenize_expr_ptr:
	call tokenize_arg_ptr
	jr tokenize_expr_checkop
tokenize_expr_sprite:
	call tokenize_arg_spr
	jr tokenize_expr_checkop			;IT'S DONE!!!
tokenize_expr_int:
	call tokenize_arg_intvar
	jr tokenize_expr_checkop			;and loop again
tokenize_expr_array:
	call tokenize_arg_arrayvar
	jr tokenize_expr_checkop
tokenize_expr_num:
	call tokenize_arg_num
tokenize_expr_checkop:
	ld hl,(parse_ptr)
	ld a,(hl)
	cp $d6								;if it's a newline,
	jp z,tokenize_expr_end				;then we're done with the expression
	cp index_stop_char
	jp z,tokenize_expr_end_2
	cp arg_separator
	jp z,tokenize_expr_end_2
	call check_valid_op
	jp nz,err_invop
	inc hl								;if it's an operator,
	ld (parse_ptr),hl					;update the parse pointer,
	ld hl,(write_ptr)					;write the operator to the token buffer,
	ld (hl),a
	inc hl
	ld (write_ptr),hl
	jp tokenize_expr					;and loop
tokenize_expr_end:
	inc hl
	ld (parse_ptr),hl					;update the parse pointer to the beginning of the next line
	ld hl,(write_ptr)					;and write a token in the token buffer to show that it's the end of the line
	ld (hl),eol_token
	inc hl
	ld (write_ptr),hl
	scf									;set carry flag to show that we *did* go to the next line...
	ret
tokenize_expr_end_2:
	xor a								;reset c
	ld (parse_ptr),hl					;and update the pointer
	ret


;//////////////////////////////Command/argument tokenizer//////////////////////////////

get_cmd_token:							;parse command at (parse_ptr), return token in a
	ld hl,(parse_ptr)
	ld de,temp_buf+1					;copy current command to temp buffer to compare
	ld bc,cmd_len
	ldir
	ld (parse_ptr),hl					;(update parse pointer to point to args list)
	ld de,temp_buf						;now compare command with command list
	ld hl,cmd_table
get_cmd_token_compare_loop:
	ld a,(hl)							;are we at the end of the table?
	or a
	jp z,err_invcmd						;if so, then this command doesn't exist - show error
	call _pstrcmp						;compare command with current one in command list
	jr z,get_cmd_token_end				;if we've found it, then execute it
	ld bc,cmd_len+2						;otherwise, look at next command in list
	add hl,bc
	jr get_cmd_token_compare_loop		;and loop
get_cmd_token_end:
	ld bc,cmd_len+1						;now get equivalent token for command
	add hl,bc
	ld a,(hl)
	ret


tokenize_next_arg:						;parse argument at (hl), write to (write_ptr), parse_ptr and write_ptr updated
	ld a,(hl)							;and get first char of arg
	cp Lquote							;is it a literal string?
	jp z,tokenize_arg_str
	cp array_char						;is it an array or map name?
	jp z,tokenize_arg_arrayvar
	cp map_char
	jp z,tokenize_arg_mapvar
	cp sprite_char						;is it a sprite name?
	jp z,tokenize_arg_spr
	cp num_char							;is it a literal number?
	jr z,tokenize_arg_num
	cp int_char							;or a varname?
	jr z,tokenize_arg_intvar
	cp string_char
	jr z,tokenize_arg_strvar
	cp bitmap_char
	jr z,tokenize_arg_bitmap
	cp pointer_char
	jp nz,err_invarg
tokenize_arg_ptr:
	call parse_varname					;i've written those comments about parsing the name
	call c,create_var					;and creating a new one if it doesn't exist TOO MANY FREAKIN' TIMES!!!
	ld hl,(parse_ptr)
	ld a,(hl)
	ld c,ptr_name_token
	cp index_start_char					;anyways, look at the next char (since it could be an array or sprite)
	jp z,tokenize_arg_arrayvar_			;if it is one of those, let the appropriate handler take care of it! why write the same code a second time???
	cp Lperiod
	jr z,tokenize_arg_spr_
	ld b,ptr_name_token					;if it's just a normal, everyday, working-class pointer, then just write its address
	jp write_bde

tokenize_arg_bitmap:					;a bitmap name...
	call parse_varname
	call c,create_var
	ld b,bmp_name_token
	jp write_bde

tokenize_arg_num:
	call parse_num						;parse the number
	ex de,hl
	ld b,lit_num_token					;when done, put in buffer
	jp write_bde

tokenize_arg_intvar:					;if not any of those, it's a varname
	call parse_varname					;so parse it
	call c,create_var					;if it doesn't exist, create it
	ld b,int_name_token
	jp write_bde

tokenize_arg_strvar:
	call parse_varname
	call c,create_var
	ld b,str_name_token
	jp write_bde

tokenize_arg_spr:
	call parse_varname					;parse the sprite name, get the address
	call c,create_var
	ld hl,(parse_ptr)					;wait...is it a sprite name or sprite position?
	ld a,(hl)
	cp Lperiod
	jr nz,tokenize_arg_spr_var
	ld c,spr_name_token
tokenize_arg_spr_:
	inc hl
	ld a,(hl)
	call get_spr_prop					;if it's a sprite position, get the property name
	ld (parse_ptr),hl
	ld b,c
	ld c,a
	jp write_bdec						;then write the sprite name and property
tokenize_arg_spr_var:
	ld b,spr_name_token					;set the token to use before it
	jp write_bde

tokenize_arg_arrayvar:
	call parse_varname					;get the var's index
	call c,create_var					;if it doesn't exist, create it
	ld hl,(parse_ptr)					;make sure that the user is showing which array element to use
	ld a,(hl)
	cp index_start_char					;if there's an left parenthesis, then an array index follows, otherwise, it's just a varname
	jr nz,tokenize_arg_arrayvar_var
	ld c,array_name_token
tokenize_arg_arrayvar_:
	ld b,c
	call write_bde
	jp tokenize_array_index				;then the rest of the array element index
tokenize_arg_arrayvar_var:
	ld b,array_name_token				;if it's an array varname, we only need to write the var index
	jp write_bde

tokenize_arg_mapvar:
	call parse_varname					;get the var's index
	call c,create_var					;if it doesn't exist, create it
	ld hl,(parse_ptr)					;see if the user is showing which element to use
	ld a,(hl)
	cp index_start_char					;if there's an left parenthesis, then an array index follows, otherwise, it's just a varname
	jr nz,tokenize_arg_mapvar_var
	ld b,map_name_token
	call write_bde
	jp tokenize_map_index				;then the map element index
tokenize_arg_mapvar_var:
	ld b,map_name_token					;if it's an array varname, we only need to write the var index
	jp write_bde

tokenize_arg_str:
	inc hl								;skip the quote at the beginning,
	ld de,(write_ptr)					;write a "literal string" token to the token buffer,
	ld a,lit_str_token
	ld (de),a
	inc de
	ld b,0								;and init the length counter
tokenize_arg_str_loop:
	ld a,(hl)							;get the next char in the string
	cp Lquote							;if it's the end of the string, end the loop
	jr z,tokenize_arg_str_end_loop
	ld (de),a							;otherwise, copy it to the token buffer
	inc hl
	inc de
	inc b								;and increment the length counter
	ld a,b
	cp str_max_len+1					;is the string too long?
	jp z,err_strlen						;if it is, show an error
	jr tokenize_arg_str_loop			;if not, loop to process the next char
tokenize_arg_str_end_loop:
	inc hl
	ld (parse_ptr),hl					;when the string has ended, update the parse pointer
	ex de,hl
	ld (hl),0							;then zero-terminate the string in the token buffer
	inc hl
	ld (write_ptr),hl					;and update the write pointer
	ret

;//////////////////////////////General tokenizers//////////////////////////////

parse_str:								;parse string at (hl), parse_ptr updated
	ld a,(hl)
	cp string_char
	jp z,parse_str_var
	cp pointer_char
	jp z,parse_str_ptr
	cp Lquote
	jp z,parse_str_lit
	jp err_str
parse_str_lit	equ	tokenize_arg_str
parse_str_var	equ	tokenize_arg_strvar
parse_str_ptr	equ	tokenize_arg_ptr


parse_label:							;parse label at (hl), parse_ptr updated, parameters set for add_label or get_label
	push hl
	ld b,0
parse_label_loop:
	ld a,(hl)
	cp $d6
	jr z,parse_label_end_loop
	ld a,b
	cp name_max_len+1
	jp z,err_namelen
	inc b
	inc hl
	jr parse_label_loop
parse_label_end_loop:
	ld (parse_ptr),hl
	pop hl
	ret


parse_varname:							;parse varname at (hl), parse_ptr updated, c set if not found, otherwise address returned in de
	call parse_varname_nocheck			;first actually parse the name
	jp check_var_index					;then see if this is a valid variable and return

parse_varname_nocheck:
	push hl								;save its starting address
	ld b,1								;and init counter
#ifdef FULL_ERR
	ld a,(hl)
#endif
	inc hl								;skip the starting type char
#ifdef FULL_ERR
	cp array_char
	jr z,parse_varname_loop
	cp int_char
	jr z,parse_varname_loop
	cp string_char
	jr z,parse_varname_loop
	cp map_char
	jr z,parse_varname_loop
	cp bitmap_char
	jr z,parse_varname_loop
	cp sprite_char
	jr z,parse_varname_loop
	cp pointer_char
	jp nz,err_invarg
#endif
parse_varname_loop:						;first, we need to find its length
	ld a,(hl)							;get the next char from the varname
	cp L0								;if it's not a letter, number, or var identifier char, it must be the end of the varname
	jr c,parse_varname_end_loop
	cp L9+1
	jr c,parse_varname_loop_continue
	cp LcapA
	jr c,parse_varname_end_loop
	cp LcapZ+1
	jr nc,parse_varname_end_loop
parse_varname_loop_continue:
	ld a,b								;if not, check on the length of the varname
	cp name_max_len+1					;is it above the max length?
	jp z,err_namelen					;if it is, show error
	inc b								;otherwise, increment varname length counter
	inc hl								;and pointer
	jr parse_varname_loop				;and loop again
parse_varname_end_loop:
	ld (parse_ptr),hl					;when done, update parse pointer
	pop hl								;and restore the pointer to the beginning of the varname
	ret


parse_num:								;parse number at (hl), parse_ptr updated, hex value returned in hl
	inc hl								;(skip '.' at beginning)
	ld de,temp_buf
	ld b,0								;init length counter
	ld a,(hl)
	cp Ldash
	jr nz,parse_num_loop
	ld (de),a
	inc hl
	inc de
	inc b
parse_num_loop:							;copy number to temp_buf
	ld a,(hl)							;get the current char
	cp L0								;is it a number?
	jr c,parse_num_end_loop
	cp L9+1
	jr nc,parse_num_end_loop			;if not, then it's the end of the number, so convert it
	ld (de),a							;otherwise, copy that char to the temp buffer
	inc hl								;and increment the pointers
	inc de
	inc b								;along with the length counter
	ld a,b
	cp num_max_len+1					;is the number too long?
	jp z,err_numlen						;if it is, show an error
	jr parse_num_loop					;otherwise, loop again
parse_num_end_loop:
	xor a
	ld (de),a							;when the whole number has been copied, zero-terminate it in the buffer,
	ld (parse_ptr),hl					;update the parse pointer,
	ld hl,temp_buf
	jp str_to_num						;then parse the string and convert it to a number


;//////////////////////////////Commonly used routines//////////////////////////////

write_bdec:								;write b, de, c to token buffer
	call write_bde
	ld (hl),c
	inc hl
	ld (write_ptr),hl
	ret

write_bde:								;write b, de to token buffer
	ld hl,(write_ptr)
write_bde_hl:
	ld (hl),b
	inc hl
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	ld (write_ptr),hl
	ret

tokenize_eol:							;check for eol in prog buffer, write eol to token buffer
	ld hl,(parse_ptr)
	ld a,(hl)
	cp $d6
	jp nz,err_eol
	ld hl,(write_ptr)
	ld (hl),eol_token
	inc hl
	ld (write_ptr),hl
	ret

tokenize_array_index:					;tokenize an array index (no inputs)
	ld hl,(write_ptr)					;show that it's the start of the array index
	ld (hl),index_start_token
	inc hl
	ld (write_ptr),hl
	ld hl,(parse_ptr)
	ld a,(hl)
	cp index_start_char
	jp nz,err_int
	inc hl
	call tokenize_expr_hl				;get the array index and write that to the token buffer
	ld a,(hl)							;then make sure that there's a ')' after the array name and index
	cp index_stop_char
	jp nz,err_syntax
	inc hl
	ld (parse_ptr),hl
	ld hl,(write_ptr)					;and write an index-end token
	ld (hl),index_end_token
	inc hl
	ld (write_ptr),hl
	ret

tokenize_map_index:						;tokenize a map index (no inputs)
	ld hl,(write_ptr)
	ld (hl),index_start_token
	inc hl
	ld (write_ptr),hl
	ld hl,(parse_ptr)					;make sure that it's a map element, not a full tilemap
	ld a,(hl)
	cp index_start_char
	jp nz,err_bmp
	inc hl
	call tokenize_expr_hl				;tokenize the x index
	ld a,(hl)
	cp arg_separator
	jp nz,err_syntax
	inc hl
	call tokenize_expr_hl				;then the y index
	ld a,(hl)
	cp index_stop_char
	jp nz,err_syntax
	inc hl
	ld (parse_ptr),hl
	ld hl,(write_ptr)
	ld (hl),index_end_token
	inc hl
	ld (write_ptr),hl
	ret

get_spr_prop:							;get sprite property at (hl), return char in a, hl updated
	ld a,(hl)
	inc hl
	cp LcapX
	ret z
	cp LcapY
	ret z
	cp LcapB
	jr z,get_spr_prop_bmp
	cp LcapV
	jr z,get_spr_prop_v
	cp LcapD
	jp nz,err_spr_prop
get_spr_prop_d:
	ld b,LcapD
	jr $+4
get_spr_prop_v:
	ld b,LcapV
	ld a,(hl)
	cp LcapX
	jr z,get_spr_prop_dv_end
	inc b
	cp LcapY
	jp nz,err_spr_prop
get_spr_prop_dv_end:
	ld a,b
	inc hl
	ret
get_spr_prop_bmp:
	ld a,(hl)
	cp LcapM
	jp nz,err_spr_prop
	inc hl
	ld a,(hl)
	cp LcapP
	jp nz,err_spr_prop
	inc hl
	ld a,LcapB
	ret

create_var:
	ld hl,temp_buf						;if we have to create the var, the name will still be in temp_buf
	ld b,(hl)							;from running check_var_index in parse_varname, so we can
	inc hl								;just use it there!
	jp add_var

conv_hex:								;convert two ASCII numbers to hex in a, c set if error
	ld a,(hl)
	call conv_digit
	call _shlacc
	ld c,a
	inc hl
	ld a,(hl)
	inc hl
	call conv_digit
	or c
	ret
conv_digit:
	cp L0
	jr c,conv_digit_error
	cp L9+1
	jr c,conv_digit_num
	cp LcapA
	jr c,conv_digit_error
	cp LcapF+1
	jr c,conv_digit_letter
conv_digit_error:
	pop hl
	scf
	ret
conv_digit_letter:
	sub LcapA-$0a
	ret
conv_digit_num:
	sub L0
	ret


char_to_tok:							;convert var type char in a to var type token in b
	ld b,map_name_token
	cp map_char
	ret z
	inc b
	cp sprite_char
	ret z
	inc b
	cp bitmap_char
	ret z
	inc b
	inc b
	cp string_char
	ret z
	inc b
	cp array_char
	ret z
	inc b
	cp int_char
	ret z
	jp err_syntax


check_valid_op:						;z if operator in a is valid
	cp LplusSign
	ret z
	cp Ldash
	ret z
	cp Lslash
	ret z
	cp Lasterisk
	ret z
	cp LLT
	ret z
	cp LGT
	ret z
	cp LEQ
	ret z
	cp Lexclam
	ret z
	cp Lbar
	ret z
	cp Lampersand
	ret

;//////////////////////////////Data/data management//////////////////////////////

#include "tokvar.asm"					;vars
#include "toklbl.asm"					;labels


cmd_table:				;holds token for each command

;data
.db 4,"DATA",0
.db 4,"DEND",1
.db 4,"ARRY",2
.db 4,"SIZE",3

;graphics commands
.db 4,"CLRS",4
.db 4,"TEXT",5
.db 4,"LINE",6
.db 4,"PIXL",7
.db 4,"PIXT",8
.db 4,"RECT",9
.db 4,"SHFT",10
.db 4,"DRAW",11
.db 4,"STOP",12
.db 4,"RCLP",13

;keyboard commands
.db 4,"GKEY",14
.db 4,"WKEY",15
.db 4,"PAUS",16
.db 4,"INPT",17

;function commands
.db 4,"FNCT",18
.db 4,"FRUN",19
.db 4,"FEND",20
.db 4,"FGOB",21

;flow control commands
.db 4,"IIFF",22
.db 4,"THEN",23
.db 4,"IFND",24
.db 4,"HALT",25
.db 4,"WAIT",26
.db 4,"RSET",27

;number commands
.db 4,"RNDM",28
.db 4,"POLR",29
.db 4,"ABSL",30

;string commands
.db 4,"SUBS",31
.db 4,"CINT",32

;sprite commands
.db 4,"DISP",33
.db 4,"BITM",34
.db 4,"SPRT",35
.db 4,"DBMP",36
.db 4,"CLSN",37

;map commands
.db 4,"NMAP",38
.db 4,"TILE",39
.db 4,"CMAP",40
.db 4,"TCHK",41
.db 4,"DMAP",42

;file commands
.db 4,"WRTE",43
.db 4,"READ",44

.db 0

.end