;MLC (Multi-platform Language for Calculators) by Epic Programming Studios
;TI-86 interpreter
;Written by Donald Straney (burntfuse) - straney@umbi.umd.edu

;Routines


;input: none
;output: _penXXX cursor moved to beginning of next line

_vnewline:
	push hl
	ld hl,_penrow
	ld a,(hl)
	add a,6
	ld (hl),a
	dec hl
_vnewline_patch	equ	$+1
	ld (hl),0
	pop hl
	ret


;input: hl is address (RAM page 1 loaded)
;output: ahl is abs address

conv_to_abs_addr:
	ld de,$c000							;is it on page 0?
	call _cphlde
	jr c,conv_to_abs_addr_1
conv_to_abs_addr_0:
	xor a								;if so, no modifications needed!
	ret
conv_to_abs_addr_1:
	ld de,$8000							;otherwise, subtract $8000 from the address
	or a
	sbc hl,de
	ld a,1								;and set the RAM page
	ret


;input: bc is amount of memory (in bytes) to allocate
;output: hl is start of memory

alloc_mem:
	push de
	ld hl,(free1_mem_ptr)				;test the first memory area
	add hl,bc
	ld de,end_free1_mem+1
	call _cphlde
	ld de,free1_mem_ptr
	jr c,alloc_mem_end					;if it can hold it, then use that one
	ld hl,(free2_mem_ptr)				;otherwise, try the next one
	add hl,bc
	ld de,end_free2_mem+1
	call _cphlde
	ld de,free2_mem_ptr
	jr c,alloc_mem_end
	ld hl,(free3_mem_ptr)
	add hl,bc
	ld de,end_free3_mem+1
	call _cphlde
	ld de,free3_mem_ptr
	jr c,alloc_mem_end
	ld hl,(free4_mem_ptr)
	add hl,bc
	ld de,end_free4_mem+1
	call _cphlde
	ld de,free4_mem_ptr
	jp nc,err_mem						;if it can't fit in any of the areas, show error
alloc_mem_end:
	ex de,hl
	ld c,(hl)							;otherwise, update the appropriate free memory pointer
	ld (hl),e
	inc hl
	ld b,(hl)
	ld (hl),d
	ld h,b								;and return the address of the allocated memory
	ld l,c
	pop de
	ret


;input: hl is zero-terminated string to convert to number
;output: hl is converted number

str_to_num:
	ld de,0									;first clear accumulator
	ld (sum),de
	ld de,temp_buf							;then copy number to temp buffer
	ld a,(hl)								;is the first char a '-'?
	cp Ldash
	jr nz,str_to_num_nominus				;if not, continue normally
	inc hl
	xor a									;otherwise, set to invert number
	jr $+4
str_to_num_nominus:
	ld a,$c9
	ld (str_to_num_patch),a
str_to_num_loop_1:
	ld a,(hl)
	or a
	jr z,str_to_num_end_loop_1
	sub L0
	cp 9+1									;are we at the end of the number yet?
	jp nc,err_cint
	ld (de),a								;if not, copy this digit to the buffer (as number, not char)
	inc de
	inc hl
	jr str_to_num_loop_1					;and loop
str_to_num_end_loop_1:
	ld hl,temp_buf							;now convert number to hex
	ex de,hl
	or a
	sbc hl,de								;first get length
	ld b,l
	ld a,b									;check if it's too long
	cp num_max_len+1
	jp nc,err_numlen
	ld a,h
	or a
	jp nz,err_numlen
str_to_num_loop_2:
	ld a,(de)
	ld l,a
	ld h,0
	push bc
	dec b
	jr z,str_to_num_end_loop_3
str_to_num_loop_3:
	push bc
	call _hltimes10							;multiply digit by 10 correct number of times
	pop bc
	djnz str_to_num_loop_3
str_to_num_end_loop_3:
	ld bc,(sum)								;then add to current sum
	add hl,bc
	ld (sum),hl
	pop bc
	inc de									;and point to next digit
	djnz str_to_num_loop_2					;loop until all digits have been parsed
str_to_num_patch	equ	$
	nop
	dec hl
	ld a,h									;(if number is supposed to be negative, adjust it)
	cpl
	ld h,a
	ld a,l
	cpl
	ld l,a
	ret


;input: hl is area to store string
;output: zero-terminated string stored to hl

Input_Str:
	xor a
	ld (Input_Str_Table),a
	ld (Input_Str_Start),hl
	ld (input_str_sp),sp
Input_Str_Loop:
	call Input_Str_Wait
	push hl
	ld hl,str_cursor
	ld bc,pixel_inv
	ld (pixel_op_patch),bc
	call draw_text_all
Input_Str_Key_Loop:
	call get_keycode
	or a
	jr z,Input_Str_Key_Loop
	ld c,a
	pop hl
	cp 95	;enter
	jr z,Input_Str_Enter
	cp 33	;delete
	jr z,Input_Str_Del
	cp 31	;alpha
	jp z,Input_Str_Alpha
	ld a,(_pencol)
	cp 112-5
	jr c,Input_Str_Find
	push hl
	jr Input_Str_Key_Loop
Input_Str_Find:
	ld de,input_str_keys_num
Input_Str_Table	equ	$+1
	ld b,0						;0=numbers, 1=letters
	djnz Input_Str_Find_Loop
	ld de,input_str_keys_alpha
Input_Str_Find_Loop:
	ld a,(de)
	or a
	jr z,Input_Str_Loop
	cp c
	jr z,Input_Str_Found
	inc de
	jr Input_Str_Find_Loop
Input_Str_Found:
	inc de
	ld a,(de)
	ld (hl),a
	inc hl
	ld a,(_pencol)
	push af
	push hl
	ld a,(de)
	ld hl,temp_buf
	ld (hl),a
	inc hl
	ld (hl),0
	ld hl,str_cursor
	ld bc,pixel_inv
	ld (pixel_op_patch),bc
	call draw_text_all
	ld hl,temp_buf
Input_Str_Layer1	equ	$+1
	ld bc,0
	ld (pixel_op_patch),bc
	call draw_text_1
Input_Str_Layer2	equ	$+1
	ld bc,0
	ld (pixel_op_patch),bc
	call draw_text_2
	ld a,(_pencol)
	add a,b
	ld (_pencol),a
	pop hl
	jp Input_Str_Loop
Input_Str_Enter:
	ld (hl),0
	ld sp,(input_str_sp)
	ret
Input_Str_Del:
Input_Str_Start	equ	$+1
	ld de,0
	call _cphlde
	jp z,Input_Str_Loop
	push hl
	ld bc,pixel_inv
	ld (pixel_op_patch),bc
	ld hl,str_cursor
	call draw_text_all
	pop hl
	pop af
	ld (_pencol),a
	dec hl
	push hl
	ld a,(hl)
	ld hl,temp_buf
	ld (hl),a
	inc hl
	ld (hl),0
	dec hl
	ld bc,pixel_inv
	ld (pixel_op_patch),bc
	call draw_text_all
	pop hl
	jp Input_Str_Loop
Input_Str_Alpha:
	push hl
	ld a,(Input_Str_Table)
	xor 1
	ld (Input_Str_Table),a
	jp Input_Str_Key_Loop

Input_Str_Wait:
	ld de,key_delay
Input_Str_Wait_Loop:
	dec de
	ex (sp),hl
	ex (sp),hl
	ld a,d
	or e
	jr nz,Input_Str_Wait_Loop
	ret


;input: hl is start of zero-terminated string
;output: string drawn in all gfx buffers

draw_text_all:
	call draw_text_1
;	jr draw_text_2
draw_text_2:
	ld a,gray2_buf/256
	call exec_c_text_draw
	ld a,scrn2_buf/256
	jp exec_c_text_draw2
draw_text_1:
	ld a,gray1_buf/256
	call exec_c_text_draw
	ld a,scrn1_buf/256
	jp exec_c_text_draw2


;input: hl is number
;output: hl is pointer to start of converted string

format_num:
	call is_hl_neg
	jp c,FormatHL
	ld de,0
	ex de,hl							;if it's negative,
	or a
	sbc hl,de
	call FormatHL
	ld de,_op1+1
	ld bc,6
	ldir
	ld hl,_op1
	ld (hl),Ldash						;put a '-' in front of it
	ret


;input: de is start of gfx buffer to shift
;output: buffer shifted right by 1 pixel

rshift_x:
	ld b,64
rshift_x_loop:
	push bc
	ld hl,bit_save
rshift_x_patch	equ	$+1
	ld (hl),0
	ld a,(temp_buf+2)
	or a
	jp nz,rshift_x_noloop
	ld a,e
	add a,11
	ld l,a
	ld a,d
	adc a,0
	ld h,a
	ld a,(hl)
	and $01
	ld hl,bit_save
	ld (hl),a
rshift_x_noloop:
	ld b,12
rshift_x_inner_loop:
	ld c,(hl)
	ld (hl),0
	ld a,(de)
	rrca
	jp nc,rshift_x_inner_loop2
	ld (hl),1
rshift_x_inner_loop2:
	dec c
	jp nz,rshift_x_inner_loop3_0
rshift_x_inner_loop3_1:
	set 7,a
	jp rshift_x_inner_loop4
rshift_x_inner_loop3_0:
	res 7,a
rshift_x_inner_loop4:
	ld (de),a
	inc de
	djnz rshift_x_inner_loop
	pop bc
	djnz rshift_x_loop
	ret


;input: de is end of gfx buffer to shift
;output: buffer shifted left by 1 pixel

lshift_x:
	ld b,64
lshift_x_loop:
	push bc
	ld hl,bit_save
lshift_x_patch	equ	$+1
	ld (hl),0
	ld a,(temp_buf+2)
	or a
	jp nz,lshift_x_noloop
	ld a,e
	sub 11
	ld l,a
	ld a,d
	sbc a,0
	ld h,a
	ld a,(hl)
	and $80
	rlca
	ld hl,bit_save
	ld (hl),a
lshift_x_noloop:
	ld b,12
lshift_x_inner_loop:
	ld c,(hl)
	ld (hl),0
	ld a,(de)
	rlca
	jp nc,lshift_x_inner_loop2
	ld (hl),1
lshift_x_inner_loop2:
	dec c
	jp nz,lshift_x_inner_loop3_0
lshift_x_inner_loop3_1:
	set 0,a
	jp lshift_x_inner_loop4
lshift_x_inner_loop3_0:
	res 0,a
lshift_x_inner_loop4:
	ld (de),a
	dec de
	djnz lshift_x_inner_loop
	pop bc
	djnz lshift_x_loop
	ret


;input: a is standard color (0-3)
;output: d is layer 1 color, e is layer 2 color

get_color:
	inc a
	ret z
	dec a
	or a
	jp z,get_color_0
	dec a
	jp z,get_color_1
	dec a
	jp z,get_color_2
	dec a
	jp z,get_color_3
	jp err_argrange
get_color_0:
	ld de,0
	ret
get_color_1:
	ld de,$ff00
	ret
get_color_2:
	ld de,$00ff
	ret
get_color_3:
	ld de,$ffff
	ret


;input: prog name at (hl), destination abs address in bde
;output: file copied to dest address on RAM page 1, hl has prog length, c set if error

load_file:
	ld a,b
	ld (load_file_patch1),a
	ld (load_file_patch2),de
	rst 20h
	rst 10h								;find its address,
	ret c
	call _ex_ahl_bde
	call _ahl_plus_2_pg3				;get past the length bytes,
	call _set_abs_src_addr				;and set that as source address for copy
	ex de,hl
	call _data_size_to_de				;get the length of the file
	ex de,hl
	xor a
	call _set_mm_num_bytes				;and set that as length of copy
	push hl
load_file_patch1	equ	$+1
	ld a,0								;and set destination address for copy
load_file_patch2	equ	$+1
	ld hl,0
	call _set_abs_dest_addr
	call _mm_ldir						;now copy file
	pop hl
	xor a
	ret


;input: none
;output: returns when no keys are being pressed

wait_for_no_key:
	call get_keycode
	or a
	jr nz,wait_for_no_key
	ret


;input: none
;output: keycode in a

get_keycode:
	ld a,%10111111
	ld b,7
get_keycode_loop:
	ld e,a
	res 7,a
	ld d,b
	call get_key
	cp $ff
	jr nz,get_keycode_search
	ld a,e
	rra
	set 7,a
	ld b,d
	djnz get_keycode_loop
get_keycode_no:
	xor a
	ret
get_keycode_search:
	ld b,a
	ld hl,key_table
get_keycode_search_loop:
	ld a,(hl)
	or a
	ret z
	inc hl
	cp c
	jr nz,get_keycode_search_skip
	ld a,(hl)
	cp b
	jr z,get_keycode_search_found
get_keycode_search_skip:
	inc hl
	inc hl
	jr get_keycode_search_loop
get_keycode_search_found:
	inc hl
	ld a,(hl)
	ret


;input: key group in a
;output: scan code in a, key group in c

get_key:
	ld c,a
	out (1),a
	nop
	nop
	in a,(1)
	ret


;input: hl, de are numbers to multiply
;output: hl=hl*de

mult16:
	ld a,d
	or a
	jr z,mult16_de
mult16_hl:
	ld b,l
	ld a,h
	or a
	jp nz,err_argrange
	ex de,hl
	jr mult16_go
mult16_de:
	ld b,e
	ld a,d
	or a
	jp nz,err_argrange
mult16_go:
	ld a,b
	or a
	jr z,mult16_0
	dec b
	ret z
	ld d,h
	ld e,l
mult16_loop:
	add hl,de
	djnz mult16_loop
	ret
mult16_0:
	ld hl,0
	ret


;input: hl is divisor, de is dividend
;output: hl=hl/de (rounded down if fraction)

div16:
	call _cphlde
	jr z,div16_1
	jr c,div16_0
	push hl
	ld hl,1
	call _cphlde
	pop hl
	ret z
	ld c,e
	ld b,d
	ex de,hl
	ld ix,0
	ld hl,0
div16_loop:
	add hl,bc
	jp c,div16_end
	inc ix
	push hl
	or a
	sbc hl,de
	pop hl
	jp c,div16_loop
	jr z,div16_end
	dec ix
div16_end:
	push ix
	pop hl
	ret
div16_1:
	ld hl,1
	ret
div16_0:
	ld hl,0
	ret


;input: none
;output: random number generator seeded

seed_random:
	ld a,r
	ld (_@RandomNumberSeed),a
	ld a,(_apdtimer)
	ld (_@RandomNumberSeed+1),a
	ret


;input: vidplane in a, coords are b,c - d,e
;output: box drawn on screen

box:
	ld (box_patch),a
	ld a,b
	cp d
	call nc,box_swap_bd
	ld a,c
	cp e
	call nc,box_swap_ce
	call FindPixel				;hl=address
	push af
	ld a,h
box_patch	equ	$+1
	add a,0
	ld h,a
	ld a,d
	sub b
	jr z,box_skip				;if width of 0, don't draw
	ld (box_width_patch),a		;a=width
	ld a,e
	sub c
	jr z,box_skip				;also skip if height of 0
	ld b,a						;b=height
	pop af
	ld de,12					;de=newline addition
box_loop:
	push bc
	push hl
	push af
box_width_patch	equ	$+1
	ld b,0						;b=width
box_inner_loop:
	ld c,a
	call pixel_op
	ld a,c
	rrca
	call c,box_inner_loop_inc	;if at end of this byte, move to next
	djnz box_inner_loop
box_inner_loop_end:
	pop af
	pop hl
	pop bc
	add hl,de					;hl to next line
	djnz box_loop
	ret

box_inner_loop_inc:
	inc hl
	ret

box_skip:
	pop hl
	ret

box_swap_bd:
	ld a,b
	ld b,d
	ld d,a
	ret
box_swap_ce:
	ld a,c
	ld c,e
	ld e,a
	ret


;input: zero-terminated strings at (hl), (de)
;output: z if equal, nz if not equal

str_cmp:
	ld a,(de)
	ld b,(hl)
	cp b
	ret nz
	inc hl
	inc de
	or a
	ret z
	jr str_cmp


;misc routines

close_gray:
    im 1
    ld a,$3C
    out (0),a
	ret

clr_buf:
	ld hl,gray1_buf
	ld de,gray1_buf+1
	ld (hl),0
	ld bc,screen_len*2-1
	ldir
	ret

clr_scrn:
	ld hl,scrn1_buf
	call clr_scrn_clear
	ld hl,scrn2_buf
clr_scrn_clear:
	ld bc,1023
	ld d,h
	ld e,l
	inc e
	ld (hl),0
	ldir
	ret

copy_scrn:
	ld hl,gray1_buf
	ld de,scrn1_buf
	call copy_scrn_buf
	ld de,scrn2_buf
copy_scrn_buf:
	ld a,64
copy_scrn_buf_loop:
	inc de
	inc de
	ld bc,12
	ldir
	inc de
	inc de
	dec a
	jr nz,copy_scrn_buf_loop
	ret

open_gray:
;Thanks to Matt Johnson for this code!
 ld hl,vector_table_addr	         ; Source = $8E00 (Start of vector table)
 ld de,vector_table_addr+1		     ; Destination = $8E01
 ld (hl),isr_addr/256                ; Set first byte so it is copied to all of them ($8F8F)
 ld bc,257                           ; 256 bytes (size of vector table)
 ldir                                ; (DE) <- (HL), BC=BC-1, Loop till B=0
 ld hl,int_copy            		     ; Source = grayscale interrupt routine)
 ld de,isr_addr                      ; Destination = $8F8F (as specified in vector table)
 ld bc,int_end-int_start             ; Length of custom interrupt routine
 ldir                                ; (DE) <- (HL), BC=BC-1, Loop till B=0
 ld a,vector_table_addr/256          ; Set up address of vector table
 ld i,a                              ;  MSB = $8E
                                     ;  LSB = (byte supplied by data bus - random)
 im 2                                ; Interrupt Mode 2
 ei                                  ; Enable Interrupts
 ret                                 ; Return to calling program

;====================================================================
; DisplayHL:  (idea from SCaBBy/Mardell)    [Assembly Coder's Zenith]
;  Display HL as a 5-digit decimal number with no leading zeros
;  out: AF, HL, BC, DE destroyed
;====================================================================
;Modified
DisplayHL:
 call FormatHL
 jp _vputs

FormatHL:
 ld c,'0'  				; save ascii value for zero
 ld de,_op1+5			; point to end of the buffer
 xor a				; zero terminate string
 ld (de),a				; set last byte to zero
FormatHLl:
 call _divhlby10		; next digit
 dec de				; next position
 add a,c				; convert to ascii
 ld (de),a				; save digit
 ld a,h				; load upper byte
 or l					; check with lower byte for zero
 jr nz,FormatHLl			; loop
 ex de,hl				; point to buffer
 ret


;===============Very Fast Gray Sprite Routine (8x8) by Donald Straney==============
;
;48 bytes/4609 cycles (worst case)
;Inputs: hl-Pointer to sprite, dark layer first, then light layer,
;		 then mask
;		 bc=x,y position to draw sprite
;
;Outputs: sprite drawn to screen, b=0, c=15
;Destroyed: af,bc,de,hl,ix
;

layer1	equ	gray1_buf
layer2	equ	gray2_buf

VeryFastGraySprite8:
VeryFastGraySprite8_testvclip1:
	ld a,c
	sub 56						;should it be clipped on the bottom of the screen?
	jr c,VeryFastGraySprite8_novclip
	ld e,a
	ld a,8
	sub e						;if so, see how much should be clipped
	ret z
	jr nc,VeryFastGraySprite8_endvclip	;and set the number of rows to draw from that
VeryFastGraySprite8_testvclip2:
	xor a						;if it's clipped on the top,
	sub c
	cp 8
	ret nc
	ld e,a
	add a,l						;offset the sprite pointer
	ld l,a
	ld a,h
	adc a,0
	ld h,a
	ld a,8
	sub e
	ld c,0						;and draw that at the top of the screen
	jr VeryFastGraySprite8_endvclip
VeryFastGraySprite8_novclip:
	ld a,8
VeryFastGraySprite8_endvclip:
	ld (VeryFastGraySprite8_rows),a
	ld a,b
	cp 88						;should it be clipped on the right?
	jr c,VeryFastGraySprite8_nohclip
	cp -8						;should it be clipped on the left?
	jr c,VeryFastGraySprite8_testhclip2
VeryFastGraySprite8_testhclip1:
	ld d,0
	ld e,$77
	ld a,b
	add a,96
	ld b,a
	ld ix,-12
	jr VeryFastGraySprite8_endhclip
VeryFastGraySprite8_testhclip2:
	cp 96
	ret nc
	ld d,$77
	ld e,0
	ld ix,0
	jr VeryFastGraySprite8_endhclip
VeryFastGraySprite8_nohclip:
	ld d,$77
	ld e,$77
	ld ix,0
VeryFastGraySprite8_endhclip:
	ld (VeryFastGraySprite8_vertadjust),ix
	ld a,d
	ld (VeryFastGraySprite8_hl1),a
	ld (VeryFastGraySprite8_hl2),a
	ld a,e
	ld (VeryFastGraySprite8_hr1),a
	ld (VeryFastGraySprite8_hr2),a
	push hl
	ld l,c					;find location on screen
	ld h,0					;hl=y
	add hl,hl				;*2
	add hl,hl				;*4
	ld e,l
	ld d,h
	add hl,hl				;*8
	add hl,de				;*12
	ld a,b					;a=x
	and %11111000
	rra						;/2
	rra						;/4
	rra						;/8
	add a,l					;add that to offset
	ld l,a					;hl=address (offset)
	ld a,h
	adc a,0
	ld h,a
VeryFastGraySprite8_vertadjust	equ	$+1
	ld de,0
	add hl,de
	ld a,b
	and %00000111			;a=shift amount
VeryFastGraySprite8_params:
	ld b,a					;find amount to patch shift code
	ld a,7
	sub b					;a=(7-shift)
	add a,a
	ld c,a
	add a,a
	add a,c					;a=(7-shift)*6
	ld (VeryFastGraySprite_shiftr),a	;jr [right=(7-shift)*6]
	ld a,b
	add a,a
	ld c,a
	add a,a
	add a,c					;a=shift*6
	ld (VeryFastGraySprite_shiftl),a	;jr [left=shift*6]
	pop ix					;restore sprite address
	push hl					;and find addresses of layers
	ld de,layer2
	add hl,de
	ex de,hl				;de=layer2
	pop hl
	ld bc,layer1
	add hl,bc				;hl=layer1
VeryFastGraySprite8_rows	equ	$+1
	ld b,8					;init row counter
VeryFastGraySprite8_loop:
	push bc					;save counter
	ld b,(ix+0)				;and get first layer in b
	ld c,(ix+8)				;second layer in c
	ld a,(ix+16)			;mask in a
	cpl
	push af					;(save them for next byte)
	push bc
	call VeryFastGraySprite_shift_r
	push af
	and (hl)				;and draw to layer 1
	or b
VeryFastGraySprite8_hl1	equ	$
	ld (hl),a
	pop af
	ex de,hl				;then layer 2
	and (hl)
	or c
VeryFastGraySprite8_hl2	equ	$
	ld (hl),a
	inc hl					;move to next byte
	inc de
	pop bc					;and restore the gfx for drawing the next byte
	pop af
	call VeryFastGraySprite_shift_l
	push af
	and (hl)				;again, draw to layer 2
	or c
VeryFastGraySprite8_hr1	equ	$
	ld (hl),a
	pop af
	ex de,hl				;and layer 1
	and (hl)
	or b
VeryFastGraySprite8_hr2	equ	$
	ld (hl),a
	ld bc,11				;now move to next line
	add hl,bc
	ex de,hl
	add hl,bc
	ex de,hl
	inc ix					;point to the next byte in sprite
	pop bc					;restore loop counter
	dec b
	jp nz,VeryFastGraySprite8_loop	;and loop!
	ret


;===============Very Fast Gray Sprite Routine (16x16) by Donald Straney==============
;
;335 bytes/16443 cycles (worst case)
;Inputs: hl-Pointer to sprite, dark layer first, then light layer,
;		 then mask
;		 bc=x,y position to draw sprite
;
;Outputs: sprite drawn to screen, b=0, c=14
;Destroyed: af,bc,de,hl,ix
;

VeryFastGraySprite16:
	ld a,c
	sub 48
	jr c,VeryFastGraySprite16_novclip
	ld e,a
	ld a,16
	sub e
	ret z
	jr nc,VeryFastGraySprite16_endvclip
VeryFastGraySprite16_testvclip2:
	xor a
	sub c
	cp 16
	ret nc
	ld e,a
	add a,a
	add a,l
	ld l,a
	ld a,h
	adc a,0
	ld h,a
	ld a,16
	sub e
	ld c,0
	jr VeryFastGraySprite16_endvclip
VeryFastGraySprite16_novclip:
	ld a,16
VeryFastGraySprite16_endvclip:
	ld (VeryFastGraySprite16_rows),a
	ld a,b
	cp 80
	jr c,VeryFastGraySprite16_nohclip
	cp 88
	jr c,VeryFastGraySprite16_testhclip4		;clip right byte only
	cp -16
	jr c,VeryFastGraySprite16_testhclip3		;clip left byte, cut right
	cp -8
	jr c,VeryFastGraySprite16_testhclip2		;clip right byte, cut left
VeryFastGraySprite16_testhclip1:				;clip left byte only
	ld a,b
	add a,96
	ld b,a
	ld ix,-12
	push bc
	ld de,$0077
	ld bc,$7777
	jr VeryFastGraySprite16_endhclip
VeryFastGraySprite16_testhclip2:
	ld a,b
	add a,96
	ld b,a
	ld ix,-12
	push bc
	ld de,$0000
	ld bc,$0077
	jr VeryFastGraySprite16_endhclip
VeryFastGraySprite16_testhclip3:
	cp 128
	ret nc
	push bc
	ld de,$7700
	ld bc,$0000
	ld ix,0
	jr VeryFastGraySprite16_endhclip
VeryFastGraySprite16_testhclip4:
	push bc
	ld de,$7777
	ld bc,$7700
	ld ix,0
	jr VeryFastGraySprite16_endhclip
VeryFastGraySprite16_nohclip:
	push bc
	ld de,$7777
	ld bc,$7777
	ld ix,0
VeryFastGraySprite16_endhclip:
	ld (VeryFastGraySprite16_vertadjust),ix
	ld a,d					;d=left byte left section
	ld (VeryFastGraySprite16_hl1),a
	ld (VeryFastGraySprite16_hl2),a
	ld a,b					;b=left byte right section
	ld (VeryFastGraySprite16_hl3),a
	ld (VeryFastGraySprite16_hl4),a
	ld a,e					;e=right byte left section
	ld (VeryFastGraySprite16_hr1),a
	ld (VeryFastGraySprite16_hr2),a
	ld a,c					;c=right byte right section
	ld (VeryFastGraySprite16_hr3),a
	ld (VeryFastGraySprite16_hr4),a
	pop bc
	push hl
	ld l,c					;find location on screen
	ld h,0					;hl=y
	add hl,hl				;*2
	add hl,hl				;*4
	ld e,l
	ld d,h
	add hl,hl				;*8
	add hl,de				;*12
	ld a,b					;a=x
	and %11111000
	rra						;/2
	rra						;/4
	rra						;/8
	add a,l					;add that to offset
	ld l,a					;hl=address (offset)
	ld a,h
	adc a,0
	ld h,a
VeryFastGraySprite16_vertadjust	equ	$+1
	ld de,0
	add hl,de
	ld a,b
	and %00000111			;a=shift amount
VeryFastGraySprite16_params:
	ld b,a					;find amount to patch shift code
	ld a,7
	sub b					;a=(7-shift)
	add a,a
	ld c,a
	add a,a
	add a,c					;a=(7-shift)*6
	ld (VeryFastGraySprite_shiftr),a	;jr [right=(7-shift)*6]
	ld a,b
	add a,a
	ld c,a
	add a,a
	add a,c					;a=shift*6
	ld (VeryFastGraySprite_shiftl),a	;jr [left=shift*6]
	pop ix					;restore sprite address
	push hl					;and find addresses of layers
	ld de,layer2
	add hl,de
	ex de,hl				;de=layer2
	pop hl
	ld bc,layer1
	add hl,bc				;hl=layer1
VeryFastGraySprite16_rows	equ	$+1
	ld b,16					;init row counter
VeryFastGraySprite16_loop:
	push bc					;save counter
	ld b,(ix+0)				;and get first layer in b
	ld c,(ix+32)			;second layer in c
	ld a,(ix+64)			;mask in a
	cpl
	push af					;(save them for next byte)
	push bc
	call VeryFastGraySprite_shift_r	;and shift all of them!
	push af
	and (hl)				;and draw to layer 1
	or b
VeryFastGraySprite16_hl1	equ	$
	ld (hl),a
	pop af
	ex de,hl				;then layer 2
	and (hl)
	or c
VeryFastGraySprite16_hl2	equ	$
	ld (hl),a
	inc hl					;move to next byte
	inc de
	pop bc					;and restore the gfx for drawing the next byte
	pop af
	call VeryFastGraySprite_shift_l	;shift them all left this time
	push af
	and (hl)				;again, draw to layer 2
	or c
VeryFastGraySprite16_hr1	equ	$
	ld (hl),a
	pop af
	ex de,hl				;and layer 1
	and (hl)
	or b
VeryFastGraySprite16_hr2	equ	$
	ld (hl),a
	inc ix					;point to the next byte in sprite
	ld b,(ix+0)				;and repeat the process for the right side!
	ld c,(ix+32)
	ld a,(ix+64)
	cpl
	push af
	push bc
	call VeryFastGraySprite_shift_r
	push af
	and (hl)
	or b
VeryFastGraySprite16_hl3	equ	$
	ld (hl),a
	pop af
	ex de,hl
	and (hl)
	or c
VeryFastGraySprite16_hl4	equ	$
	ld (hl),a
	inc hl
	inc de
	pop bc
	pop af
	call VeryFastGraySprite_shift_l
	push af
	and (hl)
	or c
VeryFastGraySprite16_hr3	equ	$
	ld (hl),a
	pop af
	ex de,hl
	and (hl)
	or b
VeryFastGraySprite16_hr4	equ	$
	ld (hl),a
	inc ix
	ld bc,10				;move to next line
	add hl,bc
	ex de,hl
	add hl,bc
	ex de,hl
	pop bc					;restore loop counter
	dec b
	jp nz,VeryFastGraySprite16_loop	;and loop!
	ret


VeryFastGraySprite_shift_r:
VeryFastGraySprite_shiftr	equ	$+1
	jr $
	srl a
	srl b
	srl c
	srl a
	srl b
	srl c
	srl a
	srl b
	srl c
	srl a
	srl b
	srl c
	srl a
	srl b
	srl c
	srl a
	srl b
	srl c
	srl a
	srl b
	srl c
	cpl
	ret

VeryFastGraySprite_shift_l:
VeryFastGraySprite_shiftl	equ	$+1
	jr $
	sla a
	sla b
	sla c
	sla a
	sla b
	sla c
	sla a
	sla b
	sla c
	sla a
	sla b
	sla c
	sla a
	sla b
	sla c
	sla a
	sla b
	sla c
	sla a
	sla b
	sla c
	sla a
	sla b
	sla c
	cpl
	ret


;
; Lite86 II Decompression Routine
; written by Kirk Meyer
;
; NOTE: This must be called as a function
;       unless you modify the "ret z" line.
;
; Inputs:
;   HL -> the compressed data
;   DE -> where to store the output
;
; Outputs:
;   AF, BC, DE, HL trashed
;
Decompress:
 ld a,(hl)
 inc hl
 scf
 adc a,a
 .db $06
DecompressLoop:
 add a,a
 jr z,Decompress
 jr c,DecompressChunk
 ldi
 jr DecompressLoop
DecompressChunk:
 ld b,a
 ld c,$1F
 ld a,(hl)
 cp c
 ret z
 push bc 
 inc hl
 ld b,(hl)
 inc hl
 push hl
 ld l,b
 ld b,a
 or c
 rlca
 rlca
 rlca
 ld h,a
 ld a,b
 and c
 add a,3
 ld c,a
 ld b,0
 add hl,de
 ldir
 pop hl
 pop af
 jr DecompressLoop


;
; FASTLINE 1.0
;
; BRESENHAM'S LINE ALGORITHM
; by Jimmy Mrdell <yarin@acc.umu.se>
;
; Parameters:  BC    = memory area to draw line (usually $FC00)
;              DE,HL = coordinates of line (D,E)-(H,L)
;                      0 < D,H < 128, 0 < E,L < 64
;
; Code size:   145 bytes (excluding FindPixel)
;             
; Registers:   AF destroyed.
;
; Performance: 4 times faster than Stephane Jantzens line routine
;
;              Delta_X > Delta_Y  =>  innerloop =  98.50 T states
;              Delta_Y > Delta_X  =>  innerloop = 106.75 T states
;

FLine:
 push bc
 push de
 push hl

#ifdef FAST_LINES
;Modification
 ld a,h
 cp d
 jp z,FLine_vert
 ld a,l
 cp e
 jp z,FLine_horiz
#endif
 call _cphlde
 jp z,FLine_pixel

 push bc
 ld a,h
 cp d
 jr nc,DXpos
 ex de,hl
DXpos:
 ld b,d
 ld c,e
 pop de
 push hl
 call FindPixel
 add hl,de
 ex (sp),hl
 ld e,c
 ld c,a
 ld a,h
 sub b
 ld b,a
 ld a,l
 sub e
 ld de,12
 jr nc,DYpos
 neg
 ld de,-12
DYpos:
 push af
 sub b
 jr nc,DY_Greater
 add a,a
 neg
 ld (x_ai+1),a
 pop af
 add a,a
 ld (XDNeg+1),a
 sub b
 pop hl
 push af
 ld a,c
;Modification
 call pixel_op
; or (hl)
; ld (hl),a
 pop af
 bit 7,a
 jr nz,XDNeg
XNewRow:
 add hl,de
x_ai:
 sub 0
 push af
 rrc c
 jr nc,XNoWrap1
 inc hl
XNoWrap1:
 ld a,c
;Modification
 call pixel_op
; or (hl)
; ld (hl),a
 pop af
 dec b
 jr z,LineDone
 jr nc,XNewRow
XDNeg:
 add a,0
 push af
 rrc c
 jr nc,XNoWrap2
 inc hl
XNoWrap2:
 ld a,c
;Modification
 call pixel_op
; or (hl)
; ld (hl),a
 pop af
 dec b
 jr z,LineDone
 jr c,XNewRow
 jr XDNeg
DY_Greater:
 neg
 add a,a
 ld (YNoWrap+1),a
 ld a,b
 add a,a
 ld (YDNeg+1),a
 pop hl
 sub h
 ld b,h
 pop hl
 push af
 ld a,c
;Modification
 call pixel_op
; or (hl)
; ld (hl),a
 pop af
YRepeat:
 bit 7,a
 jr nz,YDNeg
 rrc c
 jr nc,YNoWrap
 inc hl
YNoWrap:
 add a,0
 jr YPlot
YDNeg:
 add a,0
YPlot:
 add hl,de
 push af
 ld a,c
;Modification
 call pixel_op
; or (hl)
; ld (hl),a
 pop af
 djnz YRepeat
LineDone:
 pop hl
 pop de
 pop bc
 ret

;More modifications
pixel_op:
pixel_op_patch	equ	$+1
 jp 0
pixel_on:
 or (hl)
 ld (hl),a
 ret
pixel_off:
 cpl
 and (hl)
 ld (hl),a
 ret
pixel_inv:
 xor (hl)
 ld (hl),a
 ret


FLine_pixel:
 push bc
 ld b,h
 ld c,l
 call FindPixel
 pop de
 add hl,de
 call pixel_op
 pop hl
 pop de
 pop bc
 ret


#ifdef FAST_LINES

FLine_vert:
 ld a,l
 cp e
 jr c,FLine_vert_skip
 ex de,hl
FLine_vert_skip:		;hl->de, down
 ld a,e
 sub l
 inc a
 push af
 push bc
 ld b,h
 ld c,l
 call FindPixel
 pop bc
 add hl,bc
 pop bc					;b=height
 ld de,12
FLine_vert_loop:
 ld c,a
 call pixel_op
 ld a,c
 add hl,de
 djnz FLine_vert_loop
 pop hl
 pop de
 pop bc
 ret

FLine_horiz:
 ld a,h
 cp d
 jr c,FLine_horiz_skip
 ex de,hl
FLine_horiz_skip:		;hl->de, right
 ld a,d
 sub h
 inc a
 push af
 push bc
 ld b,h
 ld c,l
 call FindPixel
 pop bc
 add hl,bc
 pop bc					;b=width
FLine_horiz_loop:
 ld c,a
 call pixel_op
 ld a,c
 rrca
 call c,FLine_horiz_inc
 djnz FLine_horiz_loop
 pop hl
 pop de
 pop bc
 ret
FLine_horiz_inc:
 inc hl
 ret

#endif

FindPixel:	;bc is pos?
 push bc
 push de
 ld hl,ExpTable
 ld d,0
 ld a,b
 and $07
 ld e,a
 add hl,de
 ld e,(hl)
 ld h,d
 srl b
 srl b
 srl b
 ld a,c
 add a,a
 add a,a
 ld l,a
 push bc
 ld c,l
 ld b,h
 add hl,hl
 add hl,bc
 pop bc
 ld a,e
 ld e,b
 add hl,de
 pop de
 pop bc
 ret

ExpTable:
 .db $80,$40,$20,$10,$08,$04,$02,$01

;===========================================
; RandomNumber     [Assembly Coder's Zenith]
; out: A = psuedo random number
; modified: AF, BC    (14 bytes/62 t-states)
;===========================================
RandomNumber:
  ld a,r
  ld b,a
  ld a,($1234)
_@RandomNumberSeed =$-2
  ld c,a
  ld a,(bc)
  ld (_@RandomNumberSeed),a
  xor b
  sub b
  add a,c
  ret

;Taken from grayscale in Diamonds, which was taken from Bomberbloke by Matthew Shepcar :-)

int_copy:
.org isr_addr
int_start:
	exx
	ex af,af'
;Modification
	in a,(3)						;if ON has been pressed,
	bit 3,a
	jr z,int_exit					;goto the exit routine
;	in a,(3)
;	bit 3,a
;	jr z,no_down_left_bug
;	and %11111110
;	out (3),a
;no_down_left_bug:
	bit 1,a
	jr z,end_int
	ld hl,gray_cntr
	inc (hl)
	ld a,(hl)
	cp 3
	ld a,$3c	;$3c
	jr c,set_video_port
	ld a,$0a	;$0a
	ld (hl),0
set_video_port:
	out (0),a
end_int:
;Next section from ISR template by Matt Johnson
    in a,(3)                            ; Get byte from Port 3
    rra                                 ; Check Bit 0 (ON interrupt status) by using carry
    ld  a,9
    adc a,0                             ; A = $09 if ON interrupt did not occur, $0A if (carry) so
    out (3),a                           ; Write either $09 or $0A to Port 3
    ld a,$0B                                                        
    out (3),a                           ; Write $0B to Port 3
    ex af,af'
    exx
    ei
    reti

int_exit:
	ld sp,(sp_save)						;reset the sp,
	ld a,$0d							;make sure that the right ROM page is loaded,
	out (5),a
	jp end_prog							;and go back to the shell

gray_cntr: .db 0

int_end:

.end