;MLC (Multi-platform Language for Calculators) by Epic Programming Studios
;TI-86 interpreter
;Written by Donald Straney (burntfuse) - straney@umbi.umd.edu

;Sprite storage/display code (execute)

;sprite data layout
;
;2 - bitmap address
;1 - x
;1 - y
;2 - dx
;2 - dy
;2 - vx
;2 - vy

num_sprites		equ	spr_mem

spr_1_data		equ	spr_mem+1
spr_1_bmp		equ	spr_mem+1
spr_1_x			equ	spr_mem+3
spr_1_y			equ	spr_mem+4
spr_1_dx		equ	spr_mem+5
spr_1_dy		equ	spr_mem+7
spr_1_vx		equ	spr_mem+9
spr_1_vy		equ	spr_mem+11

spr_data_len	equ	12


draw_bmp:					;draw bitmap with address in de, position in b,c
	ex de,hl
	ld a,(hl)
	inc hl
	cp 8					;select the bitmap routine depending on the size
	jp z,VeryFastGraySprite8	;if it's an 8x8, use the simpler, faster one
#ifdef FULL_ERR
	cp 32					;otherwise, it must be a 16x16...*sigh*...
	jp nz,err_int
#endif
	jp VeryFastGraySprite16


get_bmp_size:				;address in de, size (height, width) returned in a
	ld a,(de)
	cp 32
	ret nz
	ld a,16
	ret


set_spr:					;sprite address in de set w/bitmap address in bc
	ex de,hl
;	ld hl,spr_1_bmp-spr_1_data
;	add hl,de
	ld (hl),c
	inc hl
	ld (hl),b
	ret


set_spr_x:					;address in de, x in a
	ld hl,spr_1_x-spr_1_data
	add hl,de
	ld (hl),a
	ret


set_spr_y:					;address in de, y in a
	ld hl,spr_1_y-spr_1_data
	add hl,de
	ld (hl),a
	ret


set_spr_vx:					;address in de, vx in bc
	ld hl,spr_1_vx-spr_1_data
	add hl,de
	ld (hl),c
	inc hl
	ld (hl),b
	ret


set_spr_vy:					;address in de, vy in bc
	ld hl,spr_1_vy-spr_1_data
	add hl,de
	ld (hl),c
	inc hl
	ld (hl),b
	ret


set_spr_dx:					;address in de, dx in bc
	ld hl,spr_1_x-spr_1_data
	add hl,de
	ex de,hl
	ld h,b
	ld l,c
	ld a,(de)
	ld bc,32768			;\
	push hl				;|
	or a				;|--is_hl_neg
	sbc hl,bc			;|
	pop hl				;/
	ld bc,100
	jr nc,set_spr_dx_neg_loop
set_spr_dx_pos_loop:
	or a
	sbc hl,bc
	jr c,set_spr_dx_end_pos
	inc a
	jr set_spr_dx_pos_loop
set_spr_dx_neg_loop:
	add hl,bc
	jr c,set_spr_dx_end_neg
	dec a
	jr set_spr_dx_neg_loop
set_spr_dx_end_neg:
	ld (de),a
	ld a,h
	or l
	ld a,(de)
	jr nz,set_spr_dx_end_neg_nz
	dec a
	jr set_spr_dx_end
set_spr_dx_end_neg_nz:
	or a
	sbc hl,bc
	jr set_spr_dx_end
set_spr_dx_end_pos:
	add hl,bc
set_spr_dx_end:
	ex de,hl
	ld (hl),a
	inc hl
	inc hl
	ld (hl),e
	inc hl
	ld (hl),d
	ret


set_spr_dy:					;address in de, dy in bc
	ld hl,spr_1_y-spr_1_data
	add hl,de
	ex de,hl
	ld h,b
	ld l,c
	ld a,(de)
	ld bc,32768		;\
	push hl			;|
	or a			;|--is_hl_neg
	sbc hl,bc		;|
	pop hl			;/
	ld bc,100
	jr nc,set_spr_dy_neg_loop
set_spr_dy_pos_loop:
	or a
	sbc hl,bc
	jr c,set_spr_dy_end_pos
	inc a
	jr set_spr_dy_pos_loop
set_spr_dy_neg_loop:
	add hl,bc
	jr c,set_spr_dy_end_neg
	dec a
	jr set_spr_dy_neg_loop
set_spr_dy_end_neg:
	ld (de),a
	ld a,h
	or l
	ld a,(de)
	jr nz,set_spr_dy_end_neg_nz
	dec a
	jr set_spr_dy_end
set_spr_dy_end_neg_nz:
	or a
	sbc hl,bc
	jr set_spr_dy_end
set_spr_dy_end_pos:
	add hl,bc
set_spr_dy_end:
	ex de,hl
	ld (hl),a
	inc hl
	inc hl
	inc hl
	ld (hl),e
	inc hl
	ld (hl),d
	ret


get_spr:					;get sprite with address in de, return bitmap address in de, coords in b,c
	ex de,hl
	ld e,(hl)
	inc hl
	ld d,(hl)
	inc hl
	ld b,(hl)
	inc hl
	ld c,(hl)
	ret


get_spr_vx:					;address in de, vx in bc
	ld hl,spr_1_vx-spr_1_data
	add hl,de
	ld c,(hl)
	inc hl
	ld b,(hl)
	ret


get_spr_vy:
	ld hl,spr_1_vy-spr_1_data
	add hl,de
	ld c,(hl)
	inc hl
	ld b,(hl)
	ret


get_spr_dx:					;address in de, dx in bc
	ld hl,spr_1_dx-spr_1_data
	add hl,de
	ld c,(hl)
	inc hl
	ld b,(hl)
	ret


get_spr_dy:
	ld hl,spr_1_dy-spr_1_data
	add hl,de
	ld c,(hl)
	inc hl
	ld b,(hl)
	ret

.end