;MLC (Multi-platform Language for Calculators) by Epic Programming Studios
;TI-86 interpreter
;Written by Donald Straney (burntfuse) - straney@umbi.umd.edu

;Label management code

;label table layout
;
;1 - number of labels
;(separate label definitions)
;1 - name length
;5 - name
;2 - data (address of label)

num_lbls	equ		lbl_table
lbl_1_def	equ		lbl_table+1
lbl_1_nl	equ		lbl_table+1
lbl_1_name	equ		lbl_table+2
lbl_1_data	equ		lbl_table+7

lbl_def_len	equ		8


add_label:					;add label with name at (hl) with length of b, address in de
	push de
	ld a,(num_lbls)			;show that there's one more label than before
	cp max_lbls				;wait...are there too many?
	jp z,err_mem_lbl		;if so, show an error
	inc a
	ld (num_lbls),a
	dec a					;calculate the location of the new label,
	push hl
	ld l,a
	ld h,lbl_def_len
	ld c,b
	call _htimesl
	ld de,lbl_table+1
	add hl,de
	ld (hl),c				;set the name length,
	inc hl
	pop de					;copy the name to the buffer,
	ex de,hl
	ld bc,name_max_len
	ldir
	ex de,hl
	pop de					;write the address,
	ld (hl),e
	inc hl
	ld (hl),d
	ret						;and we're done!


get_label:					;get label with name at (hl) with length of b from table, if doesn't exist c set, otherwise return address in hl
	ld a,b					;first init buffer with label name length/name
	ld (temp_buf),a
	ld de,temp_buf+1
	ld bc,name_max_len
	ldir
	ld a,(num_lbls)			;if there aren't any labels, we don't need to even bother looking!
	or a
	jr z,get_label_no
	ld b,a					;otherwise, init loop counter with number of labels
	ld de,temp_buf			;and init label name pointers
	ld hl,lbl_1_def
get_label_loop:
	push bc
	call _pstrcmp			;compare the label name we're looking for with the current one in the table
	jr z,get_label_end_loop	;if they match, then we've found it!
	ld bc,lbl_def_len		;otherwise, look at the next label in table
	add hl,bc
	pop bc
	djnz get_label_loop		;then loop
get_label_no:				;if we get to here, then all the labels have been tested already - it must not exist
	scf
	ret
get_label_end_loop:			;the label has been found!!! yay!!!!
	pop bc
	ld bc,lbl_1_data-lbl_1_def
	add hl,bc				;point to the address of the label
	call _ldhlind			;and get it
	xor a					;(reset the c flag)
	ret

.end