;MLC (Multi-platform Language for Calculators) by Epic Programming Studios
;TI-86 interpreter
;Written by Donald Straney (burntfuse) - straney@umbi.umd.edu

;Variable storage/management code (tokenize)

;var table layout
;
;1 - number of vars
;(separate var definitions)
;1 - name length
;1 - type char
;5 - name
;2 - data (address of allocated memory)

num_vars	equ		var_table
var_1_def	equ		var_table+1
var_1_nl	equ		var_table+1
var_1_type	equ		var_table+2
var_1_name	equ		var_table+3
var_1_data	equ		var_table+8

var_def_len	equ		9

varname_max_len	equ	name_max_len+1


add_var:					;add variable with name at (hl) with length of b, address returned in de
	ld a,(num_vars)			;show that there's one more var than before
	cp max_vars				;wait...are there too many?
	jp z,err_mem_var		;if so, show an error
	inc a
	ld (num_vars),a
	dec a					;calculate the location of the new var,
	push hl
	ld l,a
	ld h,var_def_len
	ld c,b
	call _htimesl
	ld de,var_1_def
	add hl,de
	ld (hl),c				;set the name length,
	inc hl
	pop de					;copy the name to the buffer,
	ex de,hl
	ld a,(hl)
	ld bc,varname_max_len
	ldir
	cp int_char				;and allocate some memory for this var
	jr z,add_var_2byte
	cp array_char
	jr z,add_var_2byte
	cp map_char
	jr z,add_var_2byte
	cp pointer_char
	jr z,add_var_ptr
	cp sprite_char
	jr z,add_var_spr
	cp bitmap_char
	jr z,add_var_bmp
	cp string_char
	jp nz,err_vartype
add_var_str:
	ld bc,str_max_len
	jr add_var_alloc
add_var_ptr:
	ld bc,ptr_data_len
	jr add_var_alloc
add_var_spr:
	ld hl,num_sprites
	inc (hl)
	ld a,(hl)
	cp max_sprites+1
	jp z,err_mem_spr
	ld bc,spr_data_len
	ld hl,(spr_mem_ptr)
	push hl
	add hl,bc
	ld (spr_mem_ptr),hl
	pop hl
	jr add_var_end
add_var_2byte:
	ld bc,2
add_var_alloc:
	call alloc_mem
add_var_end:
	ex de,hl
	ld (hl),e				;set the allocated mem for the var
	inc hl
	ld (hl),d
	ret
add_var_bmp:
	ex de,hl
	ld (hl),0
	inc hl
	ld (hl),0
	ld de,var_1_data-var_1_def+1
	or a
	sbc hl,de
	ex de,hl
	ret


check_var_index:			;get variable with name at (hl) with length of b from table, if doesn't exist c set, otherwise return address in de
	ld a,b					;first init buffer with var name length/name
	ld (temp_buf),a
	ld de,temp_buf+1
	ld bc,varname_max_len
	ldir
	ld a,(num_vars)			;if there aren't any vars, we don't need to even bother looking!
	or a
	jr z,check_var_index_no
	ld b,a					;otherwise, init loop counter with number of vars
	ld de,temp_buf			;and init var name pointers
	ld hl,var_1_def
check_var_index_loop:
	push bc
	call _pstrcmp			;compare the var name we're looking for with the current one in the table
	jr z,check_var_index_end_loop	;if they match, then we've found it!
	ld bc,var_def_len		;otherwise, look at the next var in table
	add hl,bc
	pop bc
	djnz check_var_index_loop	;then loop
check_var_index_no:			;if we get to here, then all the vars have been tested already - it must not exist
	scf
	ret
check_var_index_end_loop:	;the var has been found!!! yay!!!!
	pop bc
	push hl
	inc hl
	ld a,(hl)
	ld de,var_1_data-var_1_type
	add hl,de
	ld e,(hl)				;so get the address of the memory it's been allocated
	inc hl
	ld d,(hl)
	pop hl
	cp bitmap_char
	jr nz,check_var_index_end_nobmp
check_var_index_end_bmp:
	ex de,hl
	ret
check_var_index_end_nobmp:
	xor a					;(reset the c flag)
	ret


alloc_bmp:					;allocate memory for bitmap with address in de, data at hl, size in a
	push de
	push hl
	push af
	ld c,a					;first get the size (in bytes) of the bitmap
	add a,a
	add a,c
	ld c,a
	ld b,0
	push bc
	inc bc
	ld hl,(bmp_mem_ptr)
	push hl
	add hl,bc				;then check if there's enough memory left
	ld de,end_bmp_mem+1
	call _cphlde
	jp nc,err_mem_bmp
	ld (bmp_mem_ptr),hl		;if there is, update the free memory pointer
	pop de
	pop bc
	pop af
	pop hl
	push de
	ld (de),a
	inc de
	ldir					;and copy the bitmap data
	pop de
	pop hl
	ld bc,var_1_data-var_1_def
	add hl,bc				;then assign the allocated memory to the bitmap table entry
	ld (hl),e
	inc hl
	ld (hl),d
	ret


get_bmp_addr:				;get address of bitmap with address in de, return in de
	ld hl,var_1_data-var_1_def
	add hl,de
	ld e,(hl)
	inc hl
	ld d,(hl)
	ret

.end