;MLC (Multi-platform Language for Calculators) by Epic Programming Studios
;TI-86 sprite editor
;Written by Donald Straney (burntfuse) - straney@umbi.umd.edu

;Routines


;input: none
;output: small cursor moved to beginning of next line

_vnewline:
	ld hl,_penrow
	ld a,(hl)
	add a,6
	ld (hl),a
	dec hl
_vnewline_patch	equ	$+1
	ld (hl),0
	ret


;input: a is byte to convert
;output: b, c are first, second hex digits in ASCII

conv_hex:
	ld b,a
	and $0f
	call conv_hex_digit
	ld c,a
	ld a,b
	call _shracc
	call conv_hex_digit
	ld b,a
	ret
conv_hex_digit:
	cp 10
	jr nc,conv_hex_l
conv_hex_n:
	add a,L0
	ret
conv_hex_l:
	add a,LcapA-10
	ret


;input: (none)
;output: (grayscale handler started)

open_gray:
	res alt_int,(iy+exceptionflg)
	ld hl,int_copy
	ld de,int_ram
	ld bc,int_end-int_start
	ldir
 	ld a,(int_ram)
	ld hl,int_chksum+$28
	ld de,$28
	add a,(hl)
	add hl,de
	add a,(hl)
	add hl,de
	add a,(hl)
	add hl,de
	add a,(hl)
	add hl,de
	add a,(hl)
	ld (int_chksum),a
	set alt_int,(iy+exceptionflg)
	ret


;input: (none)
;output: (grayscale handler stopped)

close_gray:
	res alt_int,(iy+exceptionflg)
	ld a,$3c
	out (0),a
	ret


;input: de is offset from start of screen for clear
;output: (both gfx buffers cleared)

clr_scrn:
	ld hl,scrn1_buf
	add hl,de
	push de
	push bc
	call clr_scrn_clear
	pop bc
	pop de
	ld hl,scrn2_buf
	add hl,de
clr_scrn_clear:
	ld d,h
	ld e,l
	inc e
	ld (hl),0
	ldir
	ret


;=========================================================================
; Sprite routines by David Phillips <david@acz.org>    http://www.acz.org/
;=========================================================================
; All sprites are 8x8 masked sprites
;
; 	PutSprite      -- black/white			      (sprite, mask)
; 	PutSpriteClip  -- black/white with clipping	  (sprite, mask)
; 	GraySprite     -- grayscale				      (dark, light, mask)
; 	GraySpriteClip -- grayscale with clipping	  (dark, light, mask)
;
; in:   hl = pointer to sprite, bc = x,y location
; out:  af, bc, de, hl, ix = trash
; size: 250 bytes
;=========================================================================
GrayDraw1 = scrn1_buf				; [1024] dark grayscale plane
GrayDraw2 = scrn2_buf				; [1024] light grayscale plane
;=========================================================================
; in: hl = 8x8 masked grayscale sprite, bc = x,y location
GraySprite:
;modifications
 ld a,8
 ld (RenderSpriteCV),a
 ld a,c
 sub 57
 jr c,GraySprite_noclip
 ld d,a
 ld a,8
 sub d
 dec a
 ld (RenderSpriteCV),a
GraySprite_noclip:
;end of modifications
 push hl							; push sprite pointer
 pop ix								; pop in ix
 call FindAddress					; calc address and offset
GraySpriteE:
 push hl							; save address offset
 push ix							; save sprite pointer
 ld de,GrayDraw1					; draw to first gray buffer
 add hl,de							; add offset to buffer start
 ld a,16							; grayscale sprites are 16 bytes
 ld (RenderSpriteM),a				; adjust sprite load instruction
 call RenderSprite					; draw the sprite
 ld a,8								; second sprite is 8 bytes from mask
 ld (RenderSpriteM),a				; adjust sprite load instruction back
 pop ix								; restore sprite pointer
 pop hl								; restore address offset
 ld de,GrayDraw2					; now draw to second gray buffer
 add hl,de							; add offset to second buffer start
 ld de,$0008						; sprites are 8 bytes
 add ix,de							; advance sprite pointer

; in: hl = address in buffer, c = offset, ix = sprite
RenderSprite:
RenderSpriteCV =$+1
 ld b,8								; set row counter to 8 bytes
RenderSpriteL:
 push bc							; save row counter and sprite offset
 ld a,(ix+8)						; load mask for this row
RenderSpriteM =$-1
 cpl								; invert the mask
 ld b,a								; save inverted mask
 ld d,(ix)							; load sprite byte for this row
 inc ix								; advance sprite pointer
 ld a,c								; set shift counter to sprite offset
 ld c,$ff							; set all bits of right mask byte
 ld e,0								; clear right byte of sprite
 or a								; is the sprite aligned?
 jr z,RenderSpriteD					; then skip shifting it
RenderSpriteS:
 srl d								; shift left sprite byte
 rr e								; into the right sprite byte
 scf								; set carry
 rr b 								; shift left mask byte
 rr c								; into the right mask byte
 dec a								; decrease shift counter
 jr nz,RenderSpriteS				; loop until counter is 0
RenderSpriteD:
RenderSpriteCL =$
 ld a,b								; load the left sprite byte
 and (hl)							; and it with the screen
 or d								; or it with the mask
 ld (hl),a							; set screen memory to left byte
 inc hl								; increase screen pointer
RenderSpriteCR =$
 ld a,c								; load the right sprite byte
 and (hl)							; and it with the screen
 or e								; or it with the mask
 ld (hl),a							; set screen memory to left byte
 ld de,15							; next row is 15 bytes ahead
 add hl,de							; move screen pointer to next row
 pop bc								; restore loop counter and shift mask
 djnz RenderSpriteL					; loop for all bytes in sprite
 ret								; done drawing the sprite

; in: bc = x,y location
; out: hl = offset of byte, c = offset in video byte
FindAddress:
 ld a,c								; get the y coord
 add a,a							; multiply by 2
 add a,a							; again by 4
 ld l,a								; save as low byte
 ld h,0								; clear high byte
 add hl,hl							; multiply by 8
 add hl,hl							; and again by 16
 ld a,b								; now get the x coord
 and %11111000						; clear lower 3 bits
 rra								; divide by 2
 rra								; this time by 4
 rra								; and finally by 8
 or l								; combine with low byte
 ld l,a								; save it
 ld a,b								; get x coord again
 and %00000111						; clear all but last 3 bits
 ld c,a								; save sprite offset in c
 ret								; return to caller


;Taken from grayscale in Diamonds, which was taken from Bomberbloke by Matthew Shepcar :-)

int_copy:
.org int_ram
int_start:
	in a,(3)
	bit 1,a
	ret z
	ld hl,gray_ctr
	inc (hl)
	ld a,(hl)
	cp 3
	ld a,$3c	;$3c
	jr c,set_video_port
	ld a,$0a	;$0a
	ld (hl),0
set_video_port:
	out (0),a
	ret

gray_ctr: .db 0

int_end:


.end